*********************************************************************************
*FILE:	LINK V3.00								*
*MOTHER FILE:	FANTASM5XX.S							*
*DATE:	311094									*
*MODDED:	160497								*
*PROGRAMMER:	STUART BALL							*
*DESCRIPTION:	LIGHTSOFT LINK V3.00 - SEE DOC FC_281094_01.			*
*INPUT:	ADDRESS OF CONTROL STRING IN A0	- SHOULD BE POINTING AT /L		*
*OUTPUT:	OBJECT FILE IN MEMORY						*
*********************************************************************************

;LINKABLE FILE HEADER v2.xx
;0:		TITLE, COMMENTS,DESCRIPTION ETC
;400:		INPUTS
;724:		OUTPUTS, USEAGE
;1024:		.L		LENGTH OF GLOBAL LABELS/ADDRESSES (32 BYTES=1 LABEL)
;1028:		.L		LENGTH OF EXTERN LABELS/ADDRESSES (32 BYTES=1 LABEL)
;1032:		.L		"cAtH"=id
;1036:				START OF EXTERN LABELS/ADDRESSES
;1036+(1028.L):			START OF GLOBAL LABS/ADDRESSES
;1036+(1028.L)+(1024.L)		START OF CODE

;
;FORMAT OF LABELS/ADDRESSES TABLES
;
;28 BYTES - LABEL
;4 BYTES  - ADDRESS RELATIVE TO 0

**link v2.00 is the library linker.
**first call read_libs.
**read_libs calls calc_lib_size to calculate the size of all library files
**in ul_directory and sl_directory. Library files are type F3LB.
**next it allocates this space on the system heap.
**now it calls read_lib_files to read the files from disk into the 
**system heap.
**V2.11 can handle a creator from BCM

Link_v3:
	reset_locals
	local.l	my_pb_h,1
	local.l	temp_1,1

	entry

	move.l	8(Sp),a1	*parameter block argv
	link	a4,#-1000
	move.l	a1,command_save(A5)
	move.l	(a1),callback(a5)
	move.l	4(a1),callback_toc(a5)
	move.l	16(a1),d0
**check command
	cmpi.w	#A_work_on_file,d0
	beq	add_file_to_list
	cmpi.w	#A_Go,d0
	beq	do_link
	cmpi.w	#A_init,d0
	beq.s	link_init
	cmpi.w	#A_cleanup,d0
	beq	clean_up
	cmpi.w	#A_prefschanged,d0
	bne.s	unknown_command
	bsr	read_link_prefs
	bra.s	unknown_command	
unknown_command:		*just forget it
	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	unlk	a4
	rts_	"Libmake_PPC_v1"
link_init:
	tst.b	initted?(a5)	*we initted already (on boot up for example)
	bne	skip_sign_on	*yes
**Show link title
	lea	link_hello(pc),a0
	bsr	printit
	bsr	read_link_prefs
	move.b	#1,initted?(a5)
	CLR.L	L_CODE_POS(A5)	CODE POS IN BUFFER
	CLR.L	L_NUMBER_OF_FILES(A5)
	clr.l	tot_number_of_files(A5)	*crash now mutha! (3 hours 221195)
	CLR.L	L_HEAD_POS(A5)
	CLR.L	RELOC_TAB_POS(A5)	NO RELOC INFO
	clr.w	NUMBER_OF_LIBRARIES(a5)	*no libys
	clr.b	LINK_E_FLAG(A5)
	clr.l	lib_pos(a5)		*index into libraries buffer
	clr.w	number_of_libraries(a5)
	clr.l	main_entry(a5)	   *have an entry?

**get memory for code buffer
	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
 	MOVE.L 	d0,FREEMEM(A5)	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
 **get 25% of it
 	lsr.l	#1,d0
 	move.l	d0,code_sz(a5)	*so we can check if we run out of memory
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m1
 	pea	mem1(pc)
 	dc.w	_debugstr
m1:
 	move.l	d0,code_buffer_h(a5)
 	dc.w	_hlock
 	move.l	code_buffer_h(a5),a0
 	move.l	(a0),a0
 	add.l	#64,a0		*Some space for our entry stuff
	move.l	a0,CODE_START(a5)
**get memory for headers
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
 **get 25% of it
 	lsr.l	#1,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
  	bne.s	m2
 	pea	mem1(pc)
 	dc.w	_debugstr
m2:

 	move.l	d0,headers_buffer_h(a5)
 	dc.w	_hlock
 	move.l	headers_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,FILE_HEADERS(a5)
**get memory for file table
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 12.5% of it
 	lsr.l	#2,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m3
 	pea	mem1(pc)
 	dc.w	_debugstr
m3:
 	move.l	d0,ftable_buffer_h(a5)
 	dc.w	_hlock
 	move.l	ftable_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,file_table(a5)
**get memory for file names storage
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 6.25% of it
 	lsr.l	#3,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m4
 	pea	mem1(pc)
 	dc.w	_debugstr
m4:
 	move.l	d0,fnames_buffer_h(a5)
 	dc.w	_hlock
 	move.l	fnames_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,FILE_NAMES(a5)

**get memory for libraries storage
	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 12.5% of it
 	lsr.l	#2,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m5
 	pea	mem1(pc)
 	dc.w	_debugstr
m5:
 	move.l	d0,libraries_h(a5)
 	dc.w	_hlock

skip_sign_on:
ibye:	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	unlk	a4
	rts_	"68kLink_init"
crap:	ds.l	1
mem1:	pstring	"Memerr in LibMake PPC"
	align
link_hello:	cstring	"LibMake PPC V1.00 Build #[build$] [hours$]:[minutes$] [ukdate$].",13,"Lightsoft 1994-1997",13,13,0
	align
	
read_link_prefs:
	link	a4,#-100
	clr.l	-(sp)
	move.l	#"myPF",-(sp)
	move.w	#128,-(sp)
	dc.w	_getresource
	move.l	(sp)+,a1
	move.l	a1,d0
	beq.s	fault	*no resource
	move.l	(a1),a2
	move.l	(A2),d0
	move.l	d0,quiet_flag(a5)	*the long word = 1 = alternative chroma
	move.l	a1,-(sp)
	dc.w	_releaseresource	*bye to prefs
fault:
	unlk	a4
	rts_	"LInk68_read_prefs"
	
add_file_to_list:
	move.l	command_save(a5),a3
	move.l	20(a3),source_buff(a5)	*the file
	
	move.l	24(a3),a0	*ptr to name (pascal)
	move.l	28(a3),input_file_size(a5)
	movem.l	a1/a3,-(sp)
	bsr	GET_O_NAME	*file name
	movem.l	(sp)+,a1/a3
	move.l	28(a3),d1	*size of file
	move.l	32(a3),d0	*object=1, lib=2
	cmpi.b	#1,d0
	beq.s	add_obj
	cmpi.b	#2,d0
	beq.s	add_lib
	
	moveq	#-1,d0		*not handled
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	unlk	a4
	rts_	"LibMakePPC_add_file_to_list"
add_obj:
	bsr	link_read_in	*add file
	clr.l	d0
	move.b	LINK_E_FLAG(A5),d0
	extb.l	d0		*return the final error
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	tst.l	d0
	beq.s	no_err51
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog
no_err51:
	unlk	a4
	rts_	"LibMakePPC_add_obj"
add_lib:
**Cant add a library to a library!
	lea	link_no_add_lib(pc),a0
	bsr	printit

	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog
no_err52:

	unlk	a4
	rts_	"LibMakePPC_add_lib"
link_no_add_lib:	cstring	"Can't use libraries with LibMakePPC",13
	align
do_link:
**get memory for library func extraction into source_buff
**get memory for libraries storage

	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 6.25% of it
 	lsr.l	#3,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m6
 	pea	mem1(pc)
 	dc.w	_debugstr
m6:
	move.l	d0,source_buff_h(a5) 
 	dc.w	_hlock
	move.l	source_buff_h(a5),a0
	move.l	(a0),a0
	move.l	a0,source_buff(a5)

;	 tst.l	 main_entry(a5)	*have an entry?
;	 bne.s	 have_entry
;	 move.b	#1,link_e_flag(a5)
;	 LEA	 NO_entry_text(PC),A0
;	 BSR	 PRINTIT
;	 move.b	#1,link_e_flag(a5)
;;	 lea	 linker_errors_string(pc),a0
;;	 bsr	 send_to_error
;	 bra.s	 link_end	 *Stop
;have_entry:

**Shit or bust time really...
	bsr	DO_THE_LINKING
	tst.b	link_e_flag(a5)
	beq.s	link_ok
	LEA	NO_WRITE_TEXT(PC),A0
	BSR	PRINTIT
	move.l	command_save(a5),a0
	move.l	#-1,4(a0)	*no code handle if bad link
	bra.s	link_end
link_ok:
	move.l	l_code_pos(A5),d0
;	add.l	#64,d0		*for our entry code
	bsr	printnum
	lea	bytes_used_text(pc),a0
	bsr	printit
	move.l	code_sz(A5),d0
	bsr	printnum
	lea	cr_link(pc),a0
	bsr	printit
**see if we are getting close to the limit
	move.l	code_sz(a5),d0
	sub.l	l_code_pos(a5),d0
	cmpi.l	#20000,d0
	bgt.s	code_size_ok
	lea	inc_partition(pc),a0
	bsr	printit	
code_size_ok:
link_end:
	lea	link_done(pc),a0
	bsr	printit
	tst.b	link_e_flag(a5)	*warn user with an error window?
	beq.s	link_ok1
	lea	linker_errors_string(pc),a0	*yup
	bsr	send_to_dialog
link_ok1:
	clr.l	d0
**clear mem from source buff
	move.l	source_buff_h(a5),a0
	dc.w	_disposehandle

**bye
	move.b	LINK_E_FLAG(A5),d0
	extb.l	d0		*return the final error
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	bne	no_code1
**put our entry code in
;	 save_all
;	 move.l	code_buffer_h(a5),a2
;	 move.l	(a2),a2
;**insert 
;*	 move.l	#entry,d0
;*	 jmp	 (pc,d0.l)
;	 move.w	#0x4e71,(a2)+	 *nop - I put debug here.
;	 move.w	#0x203c,(A2)+	 *move.l
;	 move.l	main_entry(a5),d0
;	 bclr	 #31,d0		 *Thats how we tell the entry, even at zero
;	 add.l	 #54,d0		 *acct for our entry code - 10 bytes
;	 move.l	d0,(a2)+	 *offset,d0
;	 move.l	#0x4efb0800,(A2)+	 *jmp	 (pc,d0.l)
;	 move.l	#"Gbou",(a2)+	 *Our little note
;	 move.l	#"btn ",(a2)+
;	 move.l	#"6 - ",(A2)+
;	 move.l	#"Mjhi",(a2)+
;	 move.l	#"tpgu",(a2)+
;	 move.l	#" 200",(a2)+
;	 move.l	#"8. (",(A2)+
;	 move.l	#"Gmvg",(A2)+
;	 move.l	#"g xp",(a2)+
;	 move.l	#"a fs",(A2)+
;	 move.l	#"s!).",(a2)+
;	 restore_all
;**Set the size of the code

	move.l	output_index(A5),8(a0)	*size of code
;	add.l	#64,8(a0)		*For our entry code
	move.l	headers_buffer_h(a5),4(a0)	*the outputcode
no_code1:
	unlk	a4
	rts_	"LibMakePPC_do_link"
no_entry_text:	dc.b	"***FATAL***",13,"Code entry point has not been",13
	dc.b	"specified. Use the 'entry' directive",13
	dc.b	"to do this.",13,"Stopping Link.",13,13,0
	align
linker_errors_string:	cstring	"LibMakePPC errors.",13,"Please check the Log window."
	align	     
clean_up:

**bye to our memory
	move.l	headers_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	code_buffer_h(a5),a0
	dc.w	_disposehandle
 	move.l	ftable_buffer_h(a5),a0
	dc.w	_disposehandle
 	move.l	fnames_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	libraries_h(a5),a0
	dc.w	_disposehandle
				
	clr.b	initted?(a5)
	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	unlk	a4
	rts_	"LibMakePPC_clean_up"
****************************************************************
***********************************************************************************
***************
**NOW WE ACTUALLY "DO THE LINK"
**NOTE THAT LIBRARY FUNCTION READS INCREASE TOT_NUMBER_OF_FILES
DO_THE_LINKING:
	move.l	headers_buffer_h(a5),a0
	dc.w	_Hlock

	move.l	headers_buffer_h(a5),a0
	move.l	(a0),a1
	move.l	#128,output_index(a5)
	lea	copyright_text(pc),a0
	moveq	#128/4-1,d0
copy_tit:
	move.l	(a0)+,(A1)+
	dbra	d0,copy_tit
	move.l	headers_buffer_h(a5),a0
	move.l	(a0),a1
	lea	128(a1),a1
	move.l	TOT_NUMBER_OF_FILES(a5),(a1)+
	add.l	#4,output_index(a5)
	clr.l	d7	*file index
**Now copy names of first global in each file to a1+

get_next_func_name:
	MOVE.L	FILE_TABLE(A5),A0
	MOVE.L	0(A0,D7.L*8),d0	*START OF THIS FILE
	move.l	code_start(a5),a2	*pointer to files block
	add.l	d0,a2		*point to this file
	lea	1024(A2),a2	*skip header
	tst.l	(a2)	*size of globals
	bne.s	got_some_globs
**errror - no globals in file
	move.l	d7,link_file_counter(a5)
	bsr	PRINT_FILE_NAME
	lea	no_globs_in_file(pc),a0
	bsr	printit
	move.b	#1,link_e_flag(a5)
	bra	link_finished
no_globs_in_file:	cstring	" ***ERROR*** No globals in this file!",13,13
	align
got_some_globs:
**find globals
	move.l	4(a2),d0	*Size of extern
;	beq.s	no_externs
;**errror - cant have externs in library function
;	 move.l	d7,link_file_counter(a5)
;	 bsr	 PRINT_FILE_NAME
;	 lea	 externs_in_file(pc),a0
;	 bsr	 printit
;	 move.b	#1,link_e_flag(a5)
;	 bra	 link_finished
;	 
;no_externs:
	add.l	d0,a2		*add size of externs
**now add on 
* 8, for global and externs sizes,4 for imports count, 4 for code ptrs count, 4 for data ptrs count, (12)
**4 for entry, 16 for four reserved, 4 for length of code, 4 for length of data (28)
**12 for 3 reserved (See LSA0005 for details)	(12)
**grand total of 60
	add.l	#60,a2		*should now be pointing to globals
	save_all
	lea	function_text(pc),a0
	bsr	printit	*function 
	move.l	a2,a0
	bsr	printit	*name
	lea	from_text(pc),a0	*from file
	bsr	printit
	restore_all
	save_all
	move.l	d7,link_file_counter(a5)
	bsr	PRINT_FILE_NAME	*file name
	lea	cr_text(pc),a0
	bsr	printit
	restore_all
	move.l	(a2)+,(A1)+
	move.l	(a2)+,(A1)+
	move.l	(a2)+,(A1)+
	move.l	(a2)+,(A1)+
	move.l	(a2)+,(A1)+
	move.l	(a2)+,(A1)+
	move.l	(a2)+,(A1)+
	move.l	(a2)+,(A1)+	*copy name
	add.l	#32,output_index(a5)
**Do next file global
	inc.l	d7
	cmp.l	tot_number_of_files(a5),d7
	blt	get_next_func_name
**now save here and move on up number_of_files*4
	clr.l	d7	*file counter
	move.l	a1,a3	*save where offsets to go
	move.l	tot_number_of_files(a5),d6
	lsl.l	#2,d6
	add.l	d6,a1	*point to where files are to go.
	add.l	d6,output_index(a5)	*Add on offsets list size
**now splat files, filling in start addresses in a3 (offsets from START of file)
**which is offset from code+here which is a1

	move.l	a1,a2	*save here
splat_next_file:
	MOVE.L	FILE_TABLE(A5),A0
	move.l	0(a0,d7.l*8),d3	*get offset
	move.l	4(a0,d7.l*8),d4	*size of file
	beq.s	oops
	MOVE.L	CODE_START(A5),a0	*files are here
	add.l	d3,a0		*ptr to file
	move.l	output_index(a5),(a3)+	**Offset to file from start
;	add.l	a2,d3
;	move.l	d3,(a3)+	*offset to file from start
**copy from a0 to a1 d4 bytes
	
ccode:	move.b	(a0)+,(A1)+
	inc.l	output_index(a5)	*Add on this byte
	subq.l	#1,d4
	bne.s	ccode
	
**Check for next
oops:	inc.l	d7
	cmp.l	tot_number_of_files(a5),d7
	blt	splat_next_file


LINK_FINISHED:	RTS_	"Link68LIB_DO_LINK"
	EVEN
copyright_text:	dc.b	"Lightsoft library file. Version 1"
	align
function_text:	cstring	"Function "
	align
from_text:	cstring	" extracted from file "
	align
cr_text:	dc.b	13,0
	align

************
LINK_READ_IN:	*size of file in d1	
**FIRST STORE ITS OFFSET IN FILE_TABLE

	MOVE.L	FILE_TABLE(A5),A2
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	
	MOVE.L	L_CODE_POS(A5),0(A2,D0.L*8) CODE POSITION
	MOVE.L	d1,4(A2,D0.L*8) 	*size of file	

**LENGTH OF FILE IN D1
		
	MOVE.L	SOURCE_BUFF(A5),A1
**For v3, 1032 is the entry point if any
**CHECK IS A LINKABLE FILE (1032="cAtH")
	cmpi.l	#"cAtH",1032(a1)
	beq	FORMAT_FAIL	*it's an old file!
	cmpi.l	#"LOFP",1048(a1)
	bne	not_lppc_file

	MOVE.L	CODE_START(A5),A2	CODE GOES HERE
	ADD.L	L_CODE_POS(A5),A2	POINT TO RIGHT PLACE FOR CODE

	MOVE.L	D1,FILE_CODE_SIZE(A5)
	tst.l	d1
	beq	no_Code			*there is no code in the file!
L_COPY_CODE:	MOVE.B	(A1)+,(A2)+
;	DBRA	D1,L_COPY_CODE
	subq.l	#1,d1
	bne.s	l_copy_code	make 32 bit 
	MOVE.L	FILE_CODE_SIZE(A5),D1
	ADD.L	D1,L_CODE_POS(A5)	READY FOR NEXT FILE
	addq.l	#1,l_number_of_files(A5)
	ADDQ.L	#1,TOT_NUMBER_OF_FILES(A5)
;	move.l	(sp)+,a0
	CLR.L	D0
	RTS_	"Link68_LIB_RIN"
	EVEN

out_of_code_space:
	
	st	LINK_E_FLAG(A5)
	lea	link_err5(pc),a0
	bsr	printit

	moveq	#-1,d0
	rts_	"Link68_Out_of_code_space"
no_code:
	st	LINK_E_FLAG(A5)
	lea	no_code_text(pc),a0
	bsr	printit

	moveq	#-1,d0
	rts_	"Link68_No_code_error"
	
rlf_no_size:	
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper30
	cmpi.b	#"z",d0
	bgt.s	is_upper30
	sub.b	#$20,(a0)	make upper
is_upper30:
	bsr	printit
	LEA	LINK_ERR2(PC),A0
	BSR	PRINTIT

	MOVEQ	#-1,D0
	RTS

READ_LINK_FAILED:
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper3
	cmpi.b	#"z",d0
	bgt.s	is_upper3
	sub.b	#$20,(a0)	make upper
is_upper3:

	bsr	printit
	LEA	LINK_ERR2(PC),A0
	BSR	PRINTIT

	MOVEQ	#-1,D0
	RTS_	"LK_R_FAI"
	EVEN
not_lppc_file:
	st	LINK_E_FLAG(A5)
	lea	bad_file_text(pc),a0	*the label
	bsr	printit

	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	crcrtext(pc),a0
	bsr	printit
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog

	moveq	#-1,d0
	rts_	"LinkPPC_bad_file"
bad_file_text	dc.b	"***FATAL***",13
	dc.b	"File is not a LinkPPC format object file.",13
	dc.b	"File currently being processed:",13,0
	align
FORMAT_FAIL:
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper4F
	cmpi.b	#"z",d0
	bgt.s	is_upper4F
	sub.b	#$20,(a0)	make upper
is_upper4F:

	bsr	printit
	LEA	LINK_ERR7(PC),A0
	BSR	PRINTIT
	MOVEQ	#-1,D0
	RTS_	"Link68_FORMFAIL"
	EVEN
		

COMPARE_NAMES:	LEA	EXTERN_NAME(PC),A4
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CLR.L	D0
	RTS
CN_FAILED:	MOVEQ	#-1,D0
	RTS_	"Link68_CMP_NAME"
	EVEN		

GET_O_NAME:
	MOVE.L	A0,F_NAME_ADDR(A5)	SAVE FILE NAME FROM FRAME
	LEA	LINK_NAME(PC),A1	
	clr.l	d0
	move.b	(a0)+,d0
	move.b	d0,(a1)+
	subq.l	#1,d0
GET_FNAME_L1:	
	MOVE.B	(A0)+,(a1)+	STORE FILE NAME	IN OP_NAME FOR OUR USE.
	dbra	d0,GeT_FNAME_L1
GOT_FNAME_L1:	CLR.B	(A1)+
;	MOVE.L	A0,-(SP)
*8check for quiet mode
	tst.l	quiet_flag(A5)
	BNE.S	BUILD_QUIET_2	YES, QUIET PLEASE
	LEA	LINK_NAME(PC),A0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper2
	cmpi.b	#"z",d0
	bgt.s	is_upper2
	sub.b	#$20,(a0)	make upper
is_upper2:

	BSR.s	PRINTIT
	lea	space_text(pc),a0
	bsr.s	printit
	move.l	input_file_size(a5),d0
	bsr.s	printnum
	lea	bytes_text(pc),a0
	BSR.s	PRINTIT
BUILD_QUIET_2:
;	MOVE.L	(SP)+,A0
;	SUBQ.L	#1,A0	REWIND TO END OF LAST LINE! (DICK HEED)
	MOVE.L	FILE_NAMES(A5),A1
**COPY NAME TO RIGHT PLACE IN NAMES LIST
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	LSL.L	#5,D0	EACH ENTRY 32 BYTES
	ADD.L	D0,A1	POINT TO RIGHT PLACE
	LEA	LINK_NAME(PC),A2
	addq.l	#1,a2	*skip pascal count byte
COPY_LINK_NAME:	MOVE.B	(A2)+,(A1)+
	BNE.S	COPY_LINK_NAME
	RTS		A0-> NEXT LINE
space_text:	cstring	"  "
	align
bytes_text:	cstring	" bytes",13
	align
***********
**CLEAR_OUT MOVES 33K OF DATA FROM CODE_START TO SOURCE_BUFF JUST TO CLEAR
**OUT ANY DATA CACHES...
CLEAR_DATA_CACHE:
	MOVE.L	CODE_START(A5),A0
	MOVE.L	SOURCE_BUFF(A5),A1
	MOVE.L	#33000/4,D0
CO_LOOP:	MOVE.L	(A0)+,D1	GET IT INTO PROCESSOR
	NOP		SYNC PIPELINES
	MOVE.L	D1,(A1)+	
	DBRA	D0,CO_LOOP
	RTS		THAT SHOULD DO IT

************************

************************
**Modified for Anvil
PRINTIT:
	MOVEM.L	D6/A0-A4,-(SP)
	bsr	send_to_log
pi_end:	MOVEM.L	(SP)+,D6/A0-A4
	RTS_	"Link68_printit"
	
**LIMITED TO 1E7
**USES COMMAS FOR SEPERATORS.
PRINTNUM:
**FIND SIZE OF NUMBER TO QUASH LEADING ZEROS
	LEA	NEW_CHAR_BUFF(PC),A0	WHERE CHARS ARE	STORED FOR PRINTING
	clr.l	(a0)
	clr.l	4(a0)
	clr.l	8(a0)
	clr.l	12(a0)	*clear old string
	CMPI.L	#$FFFF,D0
	BGT.S	PRINT_LONG	LONG WORD PRINT
	MOVEQ	#0,D6
	MOVE.W	D0,D6
	CMPI.L	#10,D6
	BLT.S	DO_UNITS
	CMPI.L	#100,D6
	BLT.S	DO_TENS
	CMPI.L	#1000,D6
	BLT.S	DO_HUNS
	CMPI.L	#10000,D6
	BLT.S	DO_THOU
**NOW DIVIDE BY RELEVANT	UNITS
LONG_ENT:	DIVS	#10000,D6	GET TENS OF THOUDSANDS
	BSR.S	CONVERT_N	PRINT NUMBER OF TENS OF THOUSANDS
DO_THOU:	DIVS	#1000,D6	GET THOUSANDS
	BSR.S	CONVERT_N	ETC
	MOVE.B	#",",(A0)+
DO_HUNS:	DIVS	#100,D6
	BSR.S	CONVERT_N
DO_TENS:	DIVS	#10,D6
	BSR.S	CONVERT_N
DO_UNITS:	BSR.S	CONVERT_N
	CLR.B	(A0)
	LEA	NEW_CHAR_BUFF(PC),A0
	BSR	PRINTIT
	RTS
CONVERT_N:	ADDI.B	#48,D6	MAKE CHARACTER	(ASCII)
	MOVE.B	D6,(A0)+
	SWAP	D6
	ANDI.L	#$FFFF,D6
	RTS_	"Link68_PRINTNUM"
**PRINT LONG PRINT DO.L
**LIMITED TO 1E7 (10MEG)
PRINT_LONG:	CMPI.L	#9999999,D0
	BGT.S	NUMB_ERROR
	MOVE.L	D0,D1
	CMPI.L	#99999,D0
	BLT.S	DO_TEN_THOU
	CMPI.L	#999999,D0
	BLT.S	DO_HUND_THOU
	MOVE.L	#1000000,D0	GET MILLIONS
	BSR	LONG_DIV	DIVIDE D1.L BY D0.L WITH REMAINDER IN D0
	ADDI.B	#"0",D1
	MOVE.B	D1,(A0)+
	MOVE.L	D0,D6
	MOVE.B	#",",(A0)+
	BSR	PRINTIT
	MOVE.L	D6,D1
DO_HUND_THOU:	MOVE.L	#100000,D0	GET 100 THOUSANDS
	BSR	LONG_DIV
	ADDI.B	#"0",D1
	MOVE.B	D1,(A0)+
	MOVE.L	D0,D6
	CMPI.L	#$FFFF,D6	<65536?
	BLT	LONG_ENT	YES, PRINT WORD ROUTINE
	MOVE.L	D6,D1	NO,GET 10THOUSANDS
DO_TEN_THOU:	MOVE.L	#10000,D0
	BSR	LONG_DIV
	ADDI.B	#"0",D1
	MOVE.B	D1,(A0)+
	MOVE.L	D0,D6
	ANDI.L	#$FFFF,D6
	BRA	DO_THOU
NUMB_ERROR:	LEA	NUMB_TOO_BIG(PC),A0
	BSR	PRINTIT
	RTS_	"Link68_PRINT_NL"
	EVEN		

**needs C string in a0
;send_to_error:
;	 reset_locals
;	 local.l	my_pb_h2,1
;	 link	 a4,#-100
;	 move.l	#300,d0	*A_report_error
;	 move.l	a0,d1
;	 bsr.s	 send_1
;	 move.l	my_pb_h1(a4),a0
;	 dc.w	 _disposehandle	
;	 unlk	 a4
;	 rts_	 "Send_to_log"	 
	

**needs c string in a0
send_to_log:
;	reset_locals
;	local.l	my_pb_h2,1

;	link	a4,#-100
;	move.l	#320,d0	*A_String_in_log
	move.l	a0,d1
;	 movem.l	d0-d1,-(sp)	 *Save params
;	 move.l	#256,d0		 *message buffer
;	 dc.w	 _newhandleclear
;	 movem.l	(sp)+,d0-d1
;	 move.l	a0,d2
;	 beq.s	 .fault1
;	 movem.l	d0-d1,-(sp)
;	 move.l	d2,my_pb_h2(a4)
;	 dc.w	 _hlock
;	 movem.l	(sp)+,d0-d1
;	 move.l	my_pb_h2(a4),a0
;	 move.l	(a0),a0
	lea	send_buffer(pc),a0
	move.l	#320,(a0)	*command
	move.l	d1,4(A0)	*param1
	move.l	a0,-(sp)	*request parameter block to anvil
	move.l	callback(a5),a0
	jsr	(A0)
	addq.l	#4,sp			*tidy up stack (Ah)
;.fault1:
;	 move.l	my_pb_h2(a4),a0
;	 dc.w	 _disposehandle	
;	unlk	a4
	rts_	"Link68_Send_to_log"
send_buffer:	ds.b	300
	align
	
send_to_dialog:
	align
	reset_locals
	local.b	my_pb10,300
	link	a4,#-400
	move.l	a0,a1	*the string
	lea	my_pb10(a4),a0
	move.l	#500,(a0)			*show string ion dialog
	move.l	a1,4(a0)		*The string
	lea	tool_name(pc),a1
	move.l	a1,8(a0)		*tool name
	move.l	a0,-(sp)		*request parameter block to anvil
	move.l	callback(a5),a0
	jsr	(A0)
	addq.l	#4,sp			*tidy up stack (Ah)
	unlk	a4
	rts_	"LibMakePPC_dialog"
tool_name:	pstring	"Message from LibMake PPC"
	align

*****************************************

**LONG DIV DIVIDES D1.L BY D0.L
LONG_DIV:
 TST.L  D0
 BNE.S DIV
 DIVU.W     D0,D1
 RTS
DIV:
 BPL.S DO_DIV
 NEG.L D0
 BSR.S DO_DIV
 NEG.L D1
 RTS
DO_DIV:
 TST.L D1
 BPL.S DIV_P
 NEG.L D1
 BSR.S DIV_P
 NEG.L D1
 NEG.L D0
 RTS
DIV_P:
 CLR.L D3
 CLR.L D2
DIV2:
 CMP.L D0,D1
 BLS.S DIV1
 ADDQ.L #1,D3
 LSL.L #1,D0
 BMI.S DIV3
 BRA.S DIV2
DIV5:
 LSL.L #1,D2
 LSR.L #1,D0
DIV3:
 CMP.L D0,D1
DIV1:
 BCS.S DIV4
 SUB.L D0,D1
 ADDQ.L #1,D2
DIV4:
 DBF D3,DIV5
 MOVE.L D1,D0
 MOVE.L D2,D1
 RTS_	"Link68_old_LONG_DIV"
 	EVEN
**NEW - Needs:
**a0->FSspec
**A1->Name of file
**Outputs
**Temp_Fsspec(a5)=new fsspec
make_temp_FSspec:
	clr.w	-(sp)
	move.w	0(a0),-(sp)	*Vrefnum
	move.l	2(a0),-(sp)	*Dirid
	move.l	a1,-(sp)	*name
	pea	temp_FSspec(a5)	*where	  
	moveq	#1,d0
	dc.w	$aa52
	move.w	(Sp)+,d0
	ext.l	d0
	rts_	"Link68_Make_temp_Fsspec"
**PRINT_FILENAME PRINTS THE NAME BASED ON file_number
**WHICH IS incremented AS THE LINKER RUNS THROUGH THE FILES
PRINT_FILE_NAME:
	MOVE.L	link_file_counter(A5),D0
	MOVE.L	FILE_NAMES(A5),A0
	LSL.L	#5,D0
	ADD.L	D0,A0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper1
	cmpi.b	#"z",d0
	bgt.s	is_upper1
	sub.b	#$20,(a0)	make upper
is_upper1:

	BSR	PRINTIT
	RTS_	"LINK_print_file_name"
*********************************************************************************
LINK_NAME:	DS.B	40	TEMP FOR .O FILE NAME
EXTERN_NAME:	DS.B	32	LABEL NAME FIND_EXTERN IS LOOKING FOR
*********************************************************************************
link_done:	cstring	"Link finished.",13,13
	align
CR_LINK:	DC.B	13,0
LINK_FOLDER_TEXT:	
	DC.B	13,"Reading files from ",0
bytes_used_text:
	dc.b	" bytes used out of ",0
LINK_ERR2:	DC.B	" - **FATAL** Unable to find this file.",13,0
link_err5:	dc.b	13,"**FATAL** Not enough memory to build the code.",13,0
LINK_ERR3:	DC.B	13," - but the reference doesn't exist.",13,13,0
LINK_ERR4:	DC.B	13," - Offset too large.",13,0
no_code_text:	dc.b	"**FATAL** No code in this file.",13,0
double_global_text1:	dc.b	"**FATAL** The global label '",0
double_global_text2:
	dc.b	"' is",13,"defined more than once in this project",13,"(in another of your files).",13
	dc.b	"File being linked = ",0

double_global_text3:
	dc.b	"' is",13,"defined in a library in this project.",13,"Please use another name.",13
	dc.b	"File being linked = ",0
		

the_offset_text:	dc.b	" wants to",13,"link with ",0
is_too_big:	dc.b	13," but the offset is too big."
	dc.b	" Tip - if the instruction is a branch,",13,"use the long form - i.e. "
	dc.b	"BSR.L label.",13,13,0

LINK_ERR7:	DC.B	" - **FATAL** This file is an old Link (68K) file.",13,"Please remove it from the project.",13,0
Link_err8:	dc.b	" - **FATAL** This file is empty!",13,0
LINK_SPACE_TEXT:	
	DC.B	" -> ",0
error_text:	dc.b	"***ERROR*** ",0
warning_text:	dc.b	"**FATAL ERROR** ",0
is_branching_text:	dc.b	" branches to ",0
is_odd:	dc.b	" which is at an",13,"odd address! This will cause your"
	dc.b	" program to crash.",13,"Tip - put an 'align' directive at the end of",13
	dc.b	"all your assm source files.",0
starting_link:	dc.b	13,"Starting link.",13,0
no_read_libs_text:	 dc.b	 "Could not read libraries - are the paths set up correctly?",13,0

NO_WRITE_TEXT:	DC.B	"**Link errors**",13
	DC.B	0
crcrtext:	dc.b	13,13,0
inc_partition:	dc.b	"The code buffer is nearly out of space.",13,"You may want to"
	dc.b	" increase Anvils memory size.",13,13,0
	EVEN
NUMB_TOO_BIG:	DC.B	"xxxxxxxxxx",0
	align


NEW_CHAR_BUFF:	DS.B	20	SPACE FOR NUMBERS
	align
	EXTERN	SET_DIRECTORY,WRITE_MAC,send_and_wait
	EXTERN	a5_init
	extern	do_alert
	PUBLIC	LINK_V3
	GLOBAL	LINK_NAME,CR_LINK,printit
	