*********************************************************************************
*FILE:	LINK V3.00								*
*MOTHER FILE:	FANTASM5XX.S							*
*DATE:	311094									*
*MODDED:	160497								*
*PROGRAMMER:	STUART BALL							*
*DESCRIPTION:	LIGHTSOFT LINK V3.00 - SEE DOC FC_281094_01.			*
*INPUT:	ADDRESS OF CONTROL STRING IN A0	- SHOULD BE POINTING AT /L		*
*OUTPUT:	OBJECT FILE IN MEMORY						*
*********************************************************************************

;LINKABLE FILE HEADER v2.xx
;0:		TITLE, COMMENTS,DESCRIPTION ETC
;400:		INPUTS
;724:		OUTPUTS, USEAGE
;1024:		.L		LENGTH OF GLOBAL LABELS/ADDRESSES (32 BYTES=1 LABEL)
;1028:		.L		LENGTH OF EXTERN LABELS/ADDRESSES (32 BYTES=1 LABEL)
;1032:		.L		"cAtH"=id
;1036:				START OF EXTERN LABELS/ADDRESSES
;1036+(1028.L):			START OF GLOBAL LABS/ADDRESSES
;1036+(1028.L)+(1024.L)		START OF CODE

;
;FORMAT OF LABELS/ADDRESSES TABLES
;
;28 BYTES - LABEL
;4 BYTES  - ADDRESS RELATIVE TO 0

**link v2.00 is the library linker.
**first call read_libs.
**read_libs calls calc_lib_size to calculate the size of all library files
**in ul_directory and sl_directory. Library files are type F3LB.
**next it allocates this space on the system heap.
**now it calls read_lib_files to read the files from disk into the 
**system heap.
**V2.11 can handle a creator from BCM

Link_v3:
	reset_locals
	local.l	my_pb_h,1
	local.l	temp_1,1

	entry

	move.l	8(Sp),a1	*parameter block argv
	link	a4,#-1000
	move.l	a1,command_save(A5)
	move.l	(a1),callback(a5)
	move.l	4(a1),callback_toc(a5)
	move.l	16(a1),d0
**check command
	cmpi.w	#A_work_on_file,d0
	beq	add_file_to_list
	cmpi.w	#A_Go,d0
	beq	do_link
	cmpi.w	#A_init,d0
	beq.s	link_init
	cmpi.w	#A_cleanup,d0
	beq	clean_up
	cmpi.w	#A_prefschanged,d0
	bne.s	unknown_command
	bsr	read_link_prefs
	bra.s	unknown_command	
unknown_command:		*just forget it
	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	unlk	a4
	rts_	"Link68k_v4"
link_init:
	tst.b	initted?(a5)	*we initted already (on boot up for example)
	bne	skip_sign_on	*yes
	clr.b	no_header_flag(a5)	*1=no header
	clr.b	debugging(a5)
	lea	send_buffer(pc),a0
	move.l	#A_get_current_proj,(a0)	*command
	move.l	a0,-(sp)	*request parameter block to anvil
	move.l	callback(a5),a0
	jsr	(A0)
	addq.l	#4,sp			*tidy up stack (Ah)
	lea	send_buffer(pc),a0
	tst.l	(a0)
	bne.s	no_proj
	move.l	4(a0),a0
	cmp.l	#0,a0
	beq.s	no_proj
**is bit two of p_misc_flags2 set?
	move.l	(a0),a0	*project file
	btst	#2,p_misc_flags2(a0)
	beq.s	no_head3
	move.b	#1,no_header_flag(a5)
no_head3:
**debugging?
	tst.b	p_debug(a0)
	beq.s	not_debug1
	move.b	#1,debugging(a5)
not_debug1:
no_proj:

**Show link title
	lea	link_hello(pc),a0
	bsr	printit
	bsr	read_link_prefs
	move.b	#1,initted?(a5)
	CLR.L	L_CODE_POS(A5)	CODE POS IN BUFFER
	CLR.L	L_NUMBER_OF_FILES(A5)
	clr.l	tot_number_of_files(A5)	*crash now mutha! (3 hours 221195)
	CLR.L	L_HEAD_POS(A5)
	CLR.L	RELOC_TAB_POS(A5)	NO RELOC INFO
	clr.w	NUMBER_OF_LIBRARIES(a5)	*no libys
	clr.b	LINK_E_FLAG(A5)
	clr.l	lib_pos(a5)		*index into libraries buffer
	clr.w	number_of_libraries(a5)
	clr.l	main_entry(a5)	   *have an entry?

**get memory for code buffer
	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
 	MOVE.L 	d0,FREEMEM(A5)	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
 **get 25% of it
 	lsr.l	#1,d0
 	move.l	d0,code_sz(a5)	*so we can check if we run out of memory
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m1
 	pea	mem1(pc)
 	dc.w	_debugstr
m1:
 	move.l	d0,code_buffer_h(a5)
 	dc.w	_hlock
 	move.l	code_buffer_h(a5),a0
 	move.l	(a0),a0
	tst.b	no_header_flag(a5)
	bne.s	no_head4
	add.l	#64,a0		*Some space for our entry stuff
no_head4:
	move.l	a0,CODE_START(a5)
**get memory for headers
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
 **get 25% of it
 	lsr.l	#1,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
  	bne.s	m2
 	pea	mem1(pc)
 	dc.w	_debugstr
m2:

 	move.l	d0,headers_buffer_h(a5)
 	dc.w	_hlock
 	move.l	headers_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,FILE_HEADERS(a5)
**get memory for file table
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 12.5% of it
 	lsr.l	#2,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m3
 	pea	mem1(pc)
 	dc.w	_debugstr
m3:
 	move.l	d0,ftable_buffer_h(a5)
 	dc.w	_hlock
 	move.l	ftable_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,file_table(a5)
**get memory for file names storage
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 6.25% of it
 	lsr.l	#3,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m4
 	pea	mem1(pc)
 	dc.w	_debugstr
m4:
 	move.l	d0,fnames_buffer_h(a5)
 	dc.w	_hlock
 	move.l	fnames_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,FILE_NAMES(a5)

**get memory for libraries storage
	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 12.5% of it
 	lsr.l	#2,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m5
 	pea	mem1(pc)
 	dc.w	_debugstr
m5:
 	move.l	d0,libraries_h(a5)
 	dc.w	_hlock

skip_sign_on:
ibye:	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	unlk	a4
	rts_	"68kLink_init"
crap:	ds.l	1
mem1:	pstring	"Memerr in Link68k"
	align
link_hello:	cstring	"Link68K V4.00 Build #[build$] [time$] [ukdate$].",13,"Lightsoft 1998",13,13,0
	align
	
read_link_prefs:
	link	a4,#-100
	clr.l	-(sp)
	move.l	#"myPF",-(sp)
	move.w	#128,-(sp)
	dc.w	_getresource
	move.l	(sp)+,a1
	move.l	a1,d0
	beq.s	fault	*no resource
	move.l	(a1),a2
	move.l	(A2),d0
	move.l	d0,quiet_flag(a5)	*the long word = 1 = alternative chroma
	move.l	a1,-(sp)
	dc.w	_releaseresource	*bye to prefs
fault:
	unlk	a4
	rts_	"LInk68_read_prefs"
	
add_file_to_list:
	move.l	command_save(a5),a3
	move.l	20(a3),source_buff(a5)	*the file
	
	move.l	24(a3),a0	*ptr to name (pascal)
	move.l	28(a3),input_file_size(a5)
	movem.l	a1/a3,-(sp)
	bsr	GET_O_NAME	*file name
	movem.l	(sp)+,a1/a3
	move.l	28(a3),d1	*size of file
	move.l	32(a3),d0	*object=1, lib=2
	cmpi.b	#1,d0
	beq.s.	add_obj
	cmpi.b	#2,d0
	beq.s	add_lib
	
	moveq	#-1,d0		*not handled
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	unlk	a4
	rts_	"Link68k_add_file_to_list"
add_obj:
	bsr	link_read_in	*add file
	clr.l	d0
	move.b	LINK_E_FLAG(A5),d0
	extb.l	d0		*return the final error
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	tst.l	d0
	beq.s	no_err51
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog
no_err51:
	unlk	a4
	rts_	"Link68k_add_obj"
add_lib:
	bsr	add_library
	clr.l	d0
	move.b	LINK_E_FLAG(A5),d0
	extb.l	d0		*return the final error
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	tst.l	d0
	beq.s	no_err52
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog
no_err52:

	unlk	a4
	rts_	"Link68k_add_lib"

do_link:
**v3.1 - get the project and check if we make a header
**get memory for library func extraction into source_buff
**get memory for libraries storage

	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 6.25% of it
 	lsr.l	#3,d0
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m6
 	pea	mem1(pc)
 	dc.w	_debugstr
m6:
	move.l	d0,source_buff_h(a5) 
 	dc.w	_hlock
	move.l	source_buff_h(a5),a0
	move.l	(a0),a0
	move.l	a0,source_buff(a5)

	tst.l	main_entry(a5)	*have an entry?
	bne.s	have_entry
	tst.b	no_header_flag(a5)
	bne.s	have_entry
	move.b	#1,link_e_flag(a5)
	LEA	NO_entry_text(PC),A0
	BSR	PRINTIT
	move.b	#1,link_e_flag(a5)
;	lea	linker_errors_string(pc),a0
;	bsr	send_to_dialog
	bra.s	link_end	*Stop
have_entry:

**Shit or bust time really...
	bsr	DO_THE_LINKING
	tst.b	link_e_flag(a5)
	beq.s	link_ok
	LEA	NO_WRITE_TEXT(PC),A0
	BSR	PRINTIT
	move.l	command_save(a5),a0
	move.l	#-1,4(a0)	*no code handle if bad link
	bra.s	link_end
link_ok:
	move.l	l_code_pos(A5),d0
	tst.b	no_header_flag(a5)	*no header?
	bne.s	no_head1	*yes
	add.l	#64,d0		*for our entry code
no_head1:
	bsr	printnum
	lea	bytes_used_text(pc),a0
	bsr	printit
	move.l	code_sz(A5),d0
	bsr	printnum
	lea	cr_link(pc),a0
	bsr	printit
**see if we are getting close to the limit
	move.l	code_sz(a5),d0
	sub.l	l_code_pos(a5),d0
	cmpi.l	#20000,d0
	bgt.s	code_size_ok
	lea	inc_partition(pc),a0
	bsr	printit	
code_size_ok:
link_end:
	lea	link_done(pc),a0
	bsr	printit
	tst.b	link_e_flag(a5)	*warn user with an error window?
	beq.s	link_ok1
	lea	linker_errors_string(pc),a0	*yup
	bsr	send_to_dialog
link_ok1:
	clr.l	d0
**clear mem from source buff
	move.l	source_buff_h(a5),a0
	dc.w	_disposehandle

**bye
	move.b	LINK_E_FLAG(A5),d0
	extb.l	d0		*return the final error
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	bne	no_code1
	tst.b	no_header_flag(a5)
	bne	no_head2
**put our entry code in
	save_all
	move.l	code_buffer_h(a5),a2
	move.l	(a2),a2
**insert 
*	move.l	#entry,d0
*	jmp	(pc,d0.l)
	move.w	#0x4e71,(a2)+	*nop - I put debug here. 0
	move.w	#0x203c,(A2)+	*move.l 2
	move.l	main_entry(a5),d0
	bclr	#31,d0		*Thats how we tell the entry, even at zero
	add.l	#54,d0		*acct for our entry code - 10 bytes
	move.l	d0,(a2)+	*offset,d0 4
	move.l	#0x4efb0800,(A2)+	*jmp	(pc,d0.l) 8
	move.l	#"Gbou",(a2)+	*Our little note 12
	move.l	#"btn ",(a2)+	*16
	move.l	#"6 - ",(A2)+	*20
	move.l	#"Mjhi",(a2)+	*24
	move.l	#"tpgu",(a2)+	*28
	move.l	#" 200",(a2)+	*32
	move.l	#"8.  ",(A2)+	*36
	move.w	#0x4071,(a2)+	*40
**if debugging, we need to write a trap d, LS to first 4 bytes and save first 4 bytes at
*848
**copy 8 bytes of code if debugging
;	 tst.b	 debugging(a5)
;	 beq.s	 no_debug2
;	 move.l	code_buffer_h(a5),a2
;	 move.l	(a2),a2
;
;**addr 40
;	 lea	 0(a2),a3	 *the code
;	 lea	 48(a2),a2	 *copied to here
;	 move.l	(a3),(a2)+	 *first 4 bytes of code
;	 move.l	4(a3),(a2)+	 *next 4 bytes of code
;**if debugging insert trap e,"LS"
;	 move.w	#$4e4e,(a3)
;	 move.w	#"LS",2(a3)
;no_debug2:
	restore_all
**Set the size of the code
no_head2:
	move.l	l_code_pos(a5),8(a0)	*size of code
	tst.b	no_header_flag(a5)
	bne.s	no_head5
	add.l	#64,8(a0)		*For our entry code
no_head5:
	move.l	code_buffer_h(a5),4(a0)	*the code
no_code1:
	unlk	a4
	rts_	"Link68k_do_link"
no_entry_text:	dc.b	"***FATAL***",13,"Code entry point has not been",13
	dc.b	"specified. Use the 'entry' directive",13
	dc.b	"to do this.",13,"Stopping Link.",13,13,0
	align
linker_errors_string:	cstring	"Link68K errors.",13,"Please check the Log window."
	align	     
clean_up:

**bye to our memory
	move.l	headers_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	code_buffer_h(a5),a0
	dc.w	_disposehandle
 	move.l	ftable_buffer_h(a5),a0
	dc.w	_disposehandle
 	move.l	fnames_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	libraries_h(a5),a0
	dc.w	_disposehandle
				
	clr.b	initted?(a5)
	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	unlk	a4
	rts_	"LInk68k_clean_up"
****************************************************************
***********************************************************************************
***************
**NOW WE ACTUALLY "DO THE LINK"
**NOTE THAT LIBRARY FUNCTION READS INCREASE TOT_NUMBER_OF_FILES
DO_THE_LINKING:
	MOVE.L	L_NUMBER_OF_FILES(A5),TOT_NUMBER_OF_FILES(A5)
	lea	starting_link(pc),a0
	bsr	printit
	CLR.L	D2	FILE COUNTER
	clr.l	link_file_counter(a5)	for errors

LINK_LOOP:
	MOVE.L	FILE_TABLE(A5),A0
	MOVE.L	0(A0,D2.L*8),D0	CODE START OF THIS FILE
	MOVE.L	4(A0,D2.L*8),D1	HEADER START OF THIS FILE
	MOVE.L	D2,-(SP)
	BSR.S	LINK_FILE	LINK THIS FILE
	MOVE.L	(SP)+,D2
	TST.B	LINK_E_FLAG(A5)
	bne.s	link_finished
	ADDQ.L	#1,D2
	addq.l	#1,link_file_counter(a5)
	CMP.L	TOT_NUMBER_OF_FILES(A5),D2
	BLT.S	LINK_LOOP
LINK_FINISHED:	RTS_	"Link68_DO_LINK "
	EVEN

*****************************************************************
*THIS IS YER ACTUAL LINK AS PER LINK DESIGN NOTE	*
*****************************************************************
**A2 POINTS TO HEADER FOR THIS FILE
**HEADER LAYOUT:
**.L SIZE OF GLOBS
**.L SIZE OF EXTERNS
LINK_FILE:	MOVE.L	D1,HEADER_START(A5)	SAVE FOR LATER
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D1,A0	POINT TO HEADER START FOR THIS FILE
	MOVE.L	4(A0),D7	SIZE OF EXTERNS
	ADDQ.L	#8,A0	POINT TO EXTERNS
	MOVE.L	CODE_START(A5),A1	POINT TO CODE
	LSR.L	#5,D7	GET ACTUAL NUMBER OF EXTERNS
	BEQ	LF_END	NO EXTERNALS TO LINK
	
**NOW COPY EXTERN NAME AND OFFSET
**A0 IS EXTERNAL LABELS, A1 IS CODE.

LINK_FILE_LOOP:

	LEA	EXTERN_NAME(PC),A2
	MOVEQ	#6,D0
CE_LOOP:	MOVE.L	(A0)+,(A2)+	7 LONGS FOR NAME
	DBRA	D0,CE_LOOP
	MOVE.L	(A0)+,D6	GET ABS OFFSET OF LONG THAT NEEDS CHANGING
**NOW SEARCH OTHER FILES AND LIBRARIES FOR EXTERN REF.	
	BSR	FIND_EXTERN	GET ABS OFFSET OF LONG IN D0
	tst.l	d0
	BGE.S	EXTERN_OK	DOESNT EXIST IF -1
	BSR	CHECK_LIBS	CHECK LIBRARIES, AND ISSUE ERROR
	TST.L	D0
	BGE.S	EXTERN_OK	FOUND IT OK
	CLR.L	D0	GOT NOTHING
**HERE D6=LONG THAT NEEDS CHANGING
**     D0=THE ADDRESS THAT D6 NEEDS CHANGING TO
EXTERN_OK:	MOVE.L	CODE_START(A5),A2
	MOVE.L	D6,D5	*IN CASE OF $FFXXXXXX (PC OFFSET)
	ANDI.L	#$00FFFFFF,D6	*MAKE 24 BIT ADDRESS
	ADD.L	D6,A2	POINT TO DATA
**CHECK FOR PC OFFSET
	SWAP	D5	MAKE $FF00
	LSR.W	#8,D5	MAKE $00FF
	TST.B	D5	CHECK FOR -1
	BGE.S	NOT_PC_REL	YEP, SO MAKE PC OFFSET
	SUB.L	D6,D0	MAKE PC REL
	BRA.S	L_TRY_WORD	INSERT THE WORD
**CHECK FOR A BRA OP CODE	
NOT_PC_REL:	MOVE.W	-2(A2),D3	CHECK FOR BRA, OR Bcc
	MOVE.W	D3,D4	COPY IN CASE IT IS
	ANDI.W	#$6000,D3	CHECK ALL TYPE OF BRANCH INSTS
	CMPI.W	#$6000,D3
	BNE.S	L_NOT_BRA	NOT A BRANCH
**THIS IS A BRANCH - CALC OFFSET
	SUB.L	D6,D0	D0=OFFSET
	btst	#0,d0	check for odd address v3.97
	beq.s	tis_even	s'ok
	save_all
	bsr	link_odd_addr	warn of odd address.
	restore_all
	st	LINK_E_FLAG(A5)
	bra.s	lf_end
TIS_EVEN:
**CHECK FOR LONG BRANCH $6XFF
	CMPI.B	#$FF,D4
	BNE.S	L_TRY_WORD
**THIS IS LONG BRANCH
	MOVE.L	D0,(A2)	INSERT LINKED LONG OFFSET (HURRAH)
	BRA.S	LONG_DONE
L_TRY_WORD:	CMPI.L	#32767,D0
	BLT.S	LTW2
	BSR.S	L_OFF_TOO_BIG
	BRA.S	LONG_DONE
	
LTW2:	CMPI.L	#-32768,D0
	BGT.S	LTW3
	BSR.s	L_OFF_TOO_BIG
	BRA.S	LONG_DONE
LTW3:	MOVE.W	D0,(A2)	INSERT LINKED WORD OFFSET
	BRA.S	LONG_DONE
L_NOT_BRA:	MOVE.L	D0,(A2)	JUST INSERT ABS ADDRESS
LONG_DONE:	SUBQ.L	#1,D7
	BNE	LINK_FILE_LOOP	DO NEXT EXTERN
	
LF_END:	RTS_	"Link68_LINK_FIL"
	EVEN
	
L_OFF_TOO_BIG:	st	LINK_E_FLAG(A5)
	SAVE_ALL
	lea	error_text(pc),a0
	bsr	printit
	BSR	PRINT_FILE_NAME	"filename"
	lea	the_offset_text(pc),a0
	bsr	printit	" calls  "
	LEA	EXTERN_NAME(PC),A0
	BSR	PRINTIT	"name"
	lea	is_too_big(pc),a0
	bsr	printit	" - the ofset is too big."

	lea	cr_link(pc),a0
	bsr	printit
	RESTORE_ALL
	RTS
**warn of branch to odd address
link_odd_addr:	st	LINK_E_FLAG(A5)	NO OUTPUT
	lea	warning_text(pc),a0
	bsr	printit
	BSR.s	PRINT_FILE_NAME	"filename"
	lea	is_branching_text(pc),a0
	bsr	printit	" calls  "
	LEA	EXTERN_NAME(PC),A0
	BSR	PRINTIT	"name"
	lea	is_odd(pc),a0
	bsr	printit	" - the ofset is too big."

	lea	cr_link(pc),a0
	bsr	printit
	rts
**CANT FIND THE LABL, SO ERROR
CHECK_LIBS:	MOVEM.L	D6-D7/A0-A1,-(SP)
	TST.W	NUMBER_OF_LIBRARIES(A5)
	BEQ.S	THERE_ARE_NONE	NO LIBS
;	BTST	#1,BUILD_OPTIONS(A5)
;	BNE.S	THERE_ARE_NONE	/N
	BSR	READ_FILE_LIB	TRY LIBRARY FILES
	TST.l	D0		*link V2.12!
	BGT.S	FOUND_IN_LIB	-1 IF NOT FOUND
THERE_ARE_NONE:	st	LINK_E_FLAG(A5)
	lea	error_text(pc),a0
	bsr	printit
	BSR.S	PRINT_FILE_NAME
	lea	the_offset_text(pc),a0
	bsr	printit
	LEA	EXTERN_NAME(PC),A0
	CLR.B	27(A0)
	BSR	PRINTIT
	LEA	LINK_ERR3(PC),A0
	BSR	PRINTIT
FOUND_IN_LIB:	MOVEM.L	(SP)+,D6-D7/A0-A1
	RTS_	"Link68_NO_EXTER"
	EVEN
**PRINT_FILENAME PRINTS THE NAME BASED ON file_number
**WHICH IS incremented AS THE LINKER RUNS THROUGH THE FILES
PRINT_FILE_NAME:
	MOVE.L	link_file_counter(A5),D0
	MOVE.L	FILE_NAMES(A5),A0
	LSL.L	#5,D0
	ADD.L	D0,A0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper1
	cmpi.b	#"z",d0
	bgt.s	is_upper1
	sub.b	#$20,(a0)	make upper
is_upper1:

	BSR	PRINTIT
	RTS_	"LINK_print_file_name"
************
*FIND_EXTERN SEARCHES ALL LOADED FILES HEADERS FOR THE LABEL NAME IN EXTERN_NAME
*IT RETURNS EITHER THE ABS OFFSET OF THE LABEL OR -1 IF NOT FOUND IN D0
*IT SEARCHES THE FILES BACKWARDS (WHY DO I DO THIS?)
find_extern:	MOVEM.L	D6-D7/A0-A1,-(SP)
	
	LEA	EXTERN_NAME(PC),A2
	MOVE.L	FILE_TABLE(A5),A0
	MOVE.L	TOT_NUMBER_OF_FILES(A5),D7
FE_GLOBAL_LOOP:	SUBQ.L	#1,D7	LOOP COUNTER
	BMI.S	FE_NOT_FOUND	DONE ALL FILES
		
	MOVE.L	FILE_TABLE(A5),A0
	MOVE.L	4(A0,D7.L*8),D1	HEADER START OF THIS FILE - error here - d1 is crap!
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D1,A0	POINT TO HEADER START FOR THIS FILE
	MOVE.L	(A0)+,D1	SIZE OF GLOBALS
	MOVE.L	(A0)+,D2	SIZE OF EXTERNS
	ADD.L	D2,A0	SKIP EXTERNALS TO GLOBALS
	LSR.L	#5,D1	MAKE NUMBER OF GLOBALS
	BEQ.S	FE_NO_GLOBS
FE_LOCAL_LOOP
**EACH LABEL IS 7 LONGS
	MOVE.L	A0,A1	COPY START OF LABEL
	MOVEQ	#6,D3	COUNT 7 LONG COMPARES
GLOB_SEARCH:	CMPM.L	(A2)+,(A0)+	I LOVE 68K!
	BNE.S	LAB_NOT_EQUAL
	DBRA	D3,GLOB_SEARCH
	MOVE.L	(A0),D0	GET GLOBAL VALUE OF LABEL
	BRA.S	FE_FOUND
	
LAB_NOT_EQUAL:	MOVE.L	A1,A0	START OF THIS GLOB
	LEA	32(A0),A0	POINT TO NEXT GLOB
	LEA	EXTERN_NAME(PC),A2	RESET LABEL NAME
	SUBQ.L	#1,D1
	BNE.S	FE_LOCAL_LOOP
FE_NO_GLOBS:	BRA	FE_GLOBAL_LOOP	TRY NEXT FILE

FE_NOT_FOUND:	MOVEQ	#-1,D0	NOT FOUND!
FE_FOUND:	MOVEM.L	(SP)+,D6-D7/A0-A1
	tst.l	d0	*double error checking
	RTS_	"Link68_FIND_EXT"
	EVEN
************
LINK_READ_IN:	
**FIRST STORE ITS OFFSET IN FILE_TABLE
	MOVE.L	FILE_TABLE(A5),A2
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	
	MOVE.L	L_CODE_POS(A5),0(A2,D0.L*8) CODE POSITION
	MOVE.L	L_HEAD_POS(A5),4(A2,D0.L*8) HEADER POSITION
	
;	BSR	GET_O_NAME	GET NAME INTO LINK_NAME
	MOVE.L	A0,-(SP)	SAVE NEXT LINE - not really but easier to convert

**LENGTH OF FILE IN D1
		
	MOVE.L	SOURCE_BUFF(A5),A1
**For v3, 1032 is the entry point if any
**CHECK IS A LINKABLE FILE (1032="cAtH")
	cmpi.l	#"cAtH",1032(a1)
	beq	FORMAT_FAIL	*it's an old file!
	cmpi.l	#"LOFP",1048(a1)
	beq	not_68k_file
**Check for entry point
	move.l	1032(a1),d0
	beq.s	no_entry
**Stop double entries (Fnarrr <BG>)
	tst.l	main_entry(a5)
	beq.s	no_entry_yet
	bra	double_entry_error
no_entry_yet:
	add.l	l_code_pos(a5),d0	*relocate entry for abs code position
	move.l	d0,main_entry(a5)
**WE add the following code to the start of the file:
*	move.l	#entry_point,d0
*	jmp	(pc,d0.l)
no_entry:
	MOVE.L	CODE_START(A5),A2	CODE GOES HERE
	ADD.L	L_CODE_POS(A5),A2	POINT TO RIGHT PLACE FOR CODE

	move.l	code_sz(A5),d2	check we havent run out of mem
	move.l	l_code_pos(A5),d3
	cmp.l	d3,d2
	ble	out_of_code_space
	
	MOVE.L	1024(A1),D2	LEN OF GLOBALS was 128
	MOVE.L	1028(A1),D3	LEN OF EXTERNS was 132
;	move.l	1032(a1),d4	len of reloc was 136
	LEA	1036(A1),A1	HEADER+8 FOR 128 AND 132 was 140
	ADD.L	D2,A1	ADD LEN OF GLOBALS
	ADD.L	D3,A1	ADD LEN OF EXTERNS
**A1 NOW -> CODE AND D1=LENGTH OF FILE (PLEASE GOD!!!!) - YES IT DOES - YIPEE
	CLR.L	D0	CODE COUNTER
	SUB.L	#1036,D1	SIZE OF FIXED HEADER was 140
	SUB.L	D2,D1	LENGTH OF GLOBALS
	SUB.L	D3,D1	LENGTH OF EXTERNS
;	SUB.L	D4,D1	LENGTH OF RELOC
	SUBQ.L	#1,D1	QUICK TWEEK.....
	MOVE.L	D1,FILE_CODE_SIZE(A5)
	tst.l	d1
	beq	no_Code			*there is no code in the file!
L_COPY_CODE:	MOVE.B	(A1)+,(A2)+
;	DBRA	D1,L_COPY_CODE
	subq.l	#1,d1
	bne.s	l_copy_code	make 32 bit 
**CODE MERGED IN

**NOW COPY HEADER TO FILE_HEADERS

	MOVE.L	SOURCE_BUFF(A5),A1	POINT TO START
	MOVE.L	FILE_HEADERS(A5),A2	file_buff1 - need 300k
	ADD.L	L_HEAD_POS(A5),A2	POINT TO RIGHT PLACE
	LEA	1024(A1),A1	SKIP TITLE ETC was 128
	MOVE.L	(A1)+,D2	LEN OF GLOBALS
	MOVE.L	(A1)+,D3	LEN OF EXTERNS
	ADDQ.L	#4,A1	SKIP ID
;	MOVE.L	(A1)+,SCRATCH_2(A5)	SIZE OF RELOC
	ADD.L	D2,L_HEAD_POS(A5)	ADD SIZE OF GLOBS
	ADD.L	D3,L_HEAD_POS(A5)	ADD SIZE OF EXTERNS
	ADDq.L	#8,L_HEAD_POS(A5)	ACCOUNT FOR SIZES
	MOVE.L	D2,(A2)+	STORE LEN OF GLOBS
	MOVE.L	D3,(A2)+	STORE LEN OF EXTERNS

	MOVE.L	L_CODE_POS(A5),D4	BASE OFFSET
	LSR.L	#5,D3	LEN OF EXTERN IN 32 BYTES
	BEQ.S	NO_EXTERNS
	SUBQ.L	#1,D3	FOR DBRA

COPY_EXTERNS1:	MOVE.L	(A1)+,(A2)+	COPY EXTERNAL, COMPLETE WITH OFFSET
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	MAKE REAL OFFSET - ADD CODE OFFSET
	MOVE.L	D1,(A2)+	STORE OFFSET
	DBRA	D3,COPY_EXTERNS1
	
NO_EXTERNS:
	LSR.L	#5,D2	LEN OF GLOBS IN 32 BYTES
;	BEQ	NO_GLOBS_err
	beq.s	no_globs
	SUBQ.L	#1,D2
COPY_GLOBS1:
**Check that global hasn't been defined already!
	save_all
	lea	extern_name(pc),a2
	MOVE.L	(A1)+,(A2)+	COPY 32 BYTES
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	clr.l	(a2)+
	bsr	find_extern
	tst.l	d0
	bmi.s	not_exist1
	restore_all
	bra	double_global	*global	defined twice
	
not_exist1:
	BSR	try_LIB		*CHECK LIBRARIES, AND ISSUE ERROR
	tst.w	d0
	bmi.s	not_exist2
	restore_all
	bra	double_global_lib
not_exist2:
	restore_all

	MOVE.L	(A1)+,(A2)+	COPY 32 BYTES
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	ADD CODE OFFSET TO MAKE REAL OFFSET
	MOVE.L	D1,(A2)+	STORE REAL OFFSET
	DBRA	D2,COPY_GLOBS1
NO_GLOBS:
	MOVE.L	FILE_CODE_SIZE(A5),D1
	ADD.L	D1,L_CODE_POS(A5)	READY FOR NEXT FILE
	addq.l	#1,l_number_of_files(A5)
	ADDQ.L	#1,TOT_NUMBER_OF_FILES(A5)
	move.l	(sp)+,a0
	CLR.L	D0
	RTS_	"Link68_LINK_RIN"
	EVEN
not_68k_file:
	st	LINK_E_FLAG(A5)
	lea	bad_file_text(pc),a0	*the label
	bsr	printit

	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	crcrtext(pc),a0
	bsr	printit
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog
	pop	a0
	moveq	#-1,d0
	rts_	"LinkPPC_bad_file"
bad_file_text	dc.b	"***FATAL***",13
	dc.b	"File is not a Link68K format object file.",13
	dc.b	"File currently being processed:",13,0
	align

out_of_code_space:
	
	st	LINK_E_FLAG(A5)
	lea	link_err5(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"Link68_Out_of_code_space"
double_global:
	st	LINK_E_FLAG(A5)
	lea	double_global_text1(pc),a0	*the label
	bsr	printit
	lea	extern_name(pc),a0
	clr.l	28(a0)					*mask out offset long
	bsr	printit				*xyz
	lea	double_global_text2(pc),a0	*is defined more than once...
	bsr	printit
	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	crcrtext(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"Link68_Double_global"
double_global_lib:
	st	LINK_E_FLAG(A5)
	lea	double_global_text1(pc),a0	*the label
	bsr	printit
	lea	extern_name(pc),a0
	clr.l	28(a0)					*mask out offset long
	bsr	printit				*xyz
	lea	double_global_text3(pc),a0	*is defined more than once...
	bsr	printit
	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	crcrtext(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"Link68_Double_global_lib"
double_entry_error:
	st	LINK_E_FLAG(A5)
	lea	double_entry_text(pc),a0	*the label
	bsr	printit
	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	crcrtext(pc),a0
	bsr	printit

	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"Link68_Link_Double_entry"
double_entry_text	dc.b	"***FATAL***",13
	dc.b	"Entry point defined more than once",13
	dc.b	"File currently being processed:",13,0
	align
no_code:
	st	LINK_E_FLAG(A5)
	lea	no_code_text(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"Link68_No_code_error"
	
rlf_no_size:	
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper30
	cmpi.b	#"z",d0
	bgt.s	is_upper30
	sub.b	#$20,(a0)	make upper
is_upper30:
	bsr	printit
	LEA	LINK_ERR2(PC),A0
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS

READ_LINK_FAILED:
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper3
	cmpi.b	#"z",d0
	bgt.s	is_upper3
	sub.b	#$20,(a0)	make upper
is_upper3:

	bsr	printit
	LEA	LINK_ERR2(PC),A0
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS_	"LK_R_FAI"
	EVEN
FORMAT_FAIL:
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper4F
	cmpi.b	#"z",d0
	bgt.s	is_upper4F
	sub.b	#$20,(a0)	make upper
is_upper4F:

	bsr	printit
	LEA	LINK_ERR7(PC),A0
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS_	"Link68_FORMFAIL"
	EVEN
		
no_globs_err:	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper4
	cmpi.b	#"z",d0
	bgt.s	is_upper4
	sub.b	#$20,(a0)	make upper
is_upper4:

	bsr	printit
	LEA	LINK_ERR6(PC),A0	*no globals!
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS_	"Link68_no_globs"
	EVEN

****************
**LINK_READ_LIB READS A FILE FROM THE LIBRARIES ON THE HEAP
**CALLED WHEN FIND EXTERN FAILS
READ_FILE_LIB:	BSR	TRY_LIB	SEARCH LIBRARIES FOR EXTERN
	TST.W	D0
	BGE.S	GOT_LIB
	RTS
**FUNCTION FILE NOW IN SOURCE_BUFF, LENGTH IN D1 
GOT_LIB:
**FIRST STORE ITS OFFSET IN FILE_TABLE

	MOVE.L	FILE_TABLE(A5),A2
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	MOVE.L	L_CODE_POS(A5),0(A2,D0.L*8) CODE POSITION
	MOVE.L	L_HEAD_POS(A5),4(A2,D0.L*8) HEADER POSITION
	
	MOVE.L	SOURCE_BUFF(A5),A1
	MOVE.L	CODE_START(A5),A2	CODE GOES HERE
	ADD.L	L_CODE_POS(A5),A2	POINT TO RIGHT PLACE FOR CODE

	move.l	code_sz(A5),d2	check we havent run out of mem
	move.l	l_code_pos(A5),d3
	cmp.l	d3,d2
	ble	out_of_code_space
	
	MOVE.L	1024(A1),D2	LEN OF GLOBALS was 128
	MOVE.L	1028(A1),D3	LEN OF EXTERNS was 132
;	move.l	1032(a1),d4	len of reloc was 136
	LEA	1036(A1),A1	HEADER+8 FOR 128 AND 132 was 140
	ADD.L	D2,A1	ADD LEN OF GLOBALS
	ADD.L	D3,A1	ADD LEN OF EXTERNS
;	LSL.L	#2,D4	MAKE RELOCS REAL SIZE
;	add.l	d4,a1	add len of reloc

**A1 NOW -> CODE AND D1=LENGTH OF FILE (PLEASE GOD!!!!) - YES IT DOES - YIPEE

	CLR.L	D0	CODE COUNTER
	SUB.L	#1036,D1	SIZE OF FIXED HEADER was 140
	SUB.L	D2,D1	LENGTH OF GLOBALS
	SUB.L	D3,D1	LENGTH OF EXTERNS
;	SUB.L	D4,D1	LENGTH OF RELOC
	SUBQ.L	#1,D1	QUICK TWEEK.....
	
	MOVE.L	D1,FILE_CODE_SIZE(A5)
;	SUBQ.L	#1,D1	FOR DBRA....
L_COPY_CODEL:	MOVE.B	(A1)+,(A2)+
;	DBRA	D1,L_COPY_CODEL
	subq.l	#1,d1
	bne.s	l_copy_codel	32 bit now :-)
**CODE MERGED IN

**NOW COPY HEADER TO FILE_HEADERS

	MOVE.L	SOURCE_BUFF(A5),A1	POINT TO START
	MOVE.L	FILE_HEADERS(A5),A2	file_buff1
	ADD.L	L_HEAD_POS(A5),A2	POINT TO RIGHT PLACE
	LEA	1024(A1),A1	SKIP TITLE ETC was 128
	MOVE.L	(A1)+,D2	LEN OF GLOBALS
	MOVE.L	(A1)+,D3	LEN OF EXTERNS
;	MOVE.L	(A1)+,SCRATCH_2(A5)	SIZE OF RELOC
	ADDQ.L	#4,A1	SKIP ID
	ADD.L	D2,L_HEAD_POS(A5)	ADD SIZE OF GLOBS
	ADD.L	D3,L_HEAD_POS(A5)	ADD SIZE OF EXTERNS
	ADDq.L	#8,L_HEAD_POS(A5)	ACCOUNT FOR SIZES
	MOVE.L	D2,(A2)+	STORE LEN OF GLOBS
	MOVE.L	D3,(A2)+	STORE LEN OF EXTERNS

	MOVE.L	L_CODE_POS(A5),D4	BASE OFFSET

	LSR.L	#5,D3	LEN OF EXTERN IN 32 BYTES
	BEQ.S	NO_EXTERNSL
	SUBQ.L	#1,D3	FOR DBRA

COPY_EXTERNS1L:	MOVE.L	(A1)+,(A2)+	COPY EXTERNAL, COMPLETE WITH OFFSET
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	MAKE REAL OFFSET - ADD CODE OFFSET
	MOVE.L	D1,(A2)+	STORE OFFSET
	DBRA	D3,COPY_EXTERNS1L
	
NO_EXTERNSL:	LSR.L	#5,D2	LEN OF GLOBS IN 32 BYTES
	BEQ.S	NO_GLOBSL
	SUBQ.L	#1,D2
COPY_GLOBS1L:	MOVE.L	(A1)+,(A2)+	COPY 32 BYTES
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	ADD CODE OFFSET TO MAKE REAL OFFSET
	MOVE.L	D1,(A2)+	STORE REAL OFFSET
	DBRA	D2,COPY_GLOBS1L
NO_GLOBSL:

	MOVE.L	FILE_CODE_SIZE(A5),D1
	ADD.L	D1,L_CODE_POS(A5)	READY FOR NEXT FILE
**copy EXTERN_NAME TO FILE NAMES FOR LIBRARY ENTRIES...
	MOVE.L	FILE_NAMES(A5),A0
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	LSL.L	#5,D0
	ADD.L	D0,A0
	LEA	EXTERN_NAME(PC),A1
	MOVE.L	A0,A2	COPY SO WE CAN MODIFY
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+	32 BYTES (LAZY I KNOW)
**NOW SEARCH FOR FIRST ZERO AND INSERT "(lib)"
**accounting for 32 bytes length
	moveq	#30,d0	size of name
find_zero:	tst.b	(a2)+
	beq.s	got_end
	subq.w	#1,d0
	bne.s	find_zero
	bra.s	outta_space
**insert (lib)
got_end:	subq.l	#1,a2	back up over zero
	move.b	#" ",(a2)+	insert space
	subq.w	#1,d0
	beq.s	outta_space	This paste function
	move.b	#"(",(a2)+	should really be outlawed
	subq.w	#1,d0	:-)
	beq.s	outta_space
	move.b	#"L",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
	move.b	#"i",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
	move.b	#"b",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
	move.b	#")",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
outta_space:
	addq.l	#1,l_number_of_files(A5)
	ADDQ.L	#1,TOT_NUMBER_OF_FILES(A5)
	
	CLR.L	D0	NO ERRORS...

	BSR	FIND_EXTERN	NOW DO REAL FIND_EXTERN
	RTS_	"Link68_LINK_RLI"
	EVEN
***********
**TRY_LIB SEARCHES THE LIBRARIES FOR EXTERN_NAME
TRY_LIB:	
	MOVEM.L	A0-A4,-(SP)
	MOVE.L	LIBRARIES_H(A5),A0
	MOVE.L	(A0),A0	START OF LIBRARY FILES

	MOVE.W	NUMBER_OF_LIBRARIES(A5),D7
	beq.s	em_not_found		*There are no libs  - master of coding (really stu! - 2.51!)
	
	LEA	LIBRARY_STARTS(a5),A1	ADDRESS OF LIBS
	SUBQ.W	#1,D7
TL_LOOP:
	MOVE.L	0(A1,D7.W*4),A2	POINT TO LIBRARY
	BSR.S	CHECK_LIB	SEARCH THIS LIB FOR EXTERN NAME
	TST.W	D0
	BGE.S	GOT_IT	FOUND IT!
	DBRA	D7,TL_LOOP
	BRA.S	EM_NOT_FOUND	NOT FOUND
**HERE D7=LIBRARY FILE NUMBER
**     D5=FUNCTION POSITION IN LIBRARY FILE
GOT_IT:
	MOVE.L	0(A1,D7.W*4),A2	POINT TO LIB
	MOVE.L	A2,A3
	MOVE.L	128(A2),D0	NUMBER OF FUNCS IN LIB
	LSL.L	#5,D0	TIMES 32 FOR NAMES
	LEA	132(A3),A3	POINT TO NAMES
	ADD.L	D0,A3	POINT TO OFFSETS
	MOVE.L	0(A3,D5.W*4),D2	GET OFFSET TO START OF FILE
	ADD.L	D2,A2	POINT OT CORRECT FILE!
**NOW WE NEED TO COPY THIS FILE TO SOURCE_BUFF(A5) AND PUT FILE LENGTH IN D1
**FIRST CALCULATE TO LIBRARY FILE SIZE
	MOVE.L	4(A1,D7.W*4),D0
	SUB.L	0(A1,D7.W*4),D0	LIB FILE LENGTH

**NOW SEE IF THIS IS LAST FUNCTION IN THIS LIBRARY
	MOVE.L	0(A1,D7.W*4),A1	POINT TO LIB
	MOVE.L	128(A1),D1	NUMBER OF FUNCTIONS IN LIB
	SUBQ.L	#1,D1
	CMP.L	D5,D1
	BNE.S	NOT_LAST
**LAST FUNCTION - LIBRARY FILE LENGTH IN D0
	SUB.L	D2,D0	NEXT FILE START - OFFST
	MOVE.L	D0,D1	RETURN LENGTH IN D1
	BRA.S	GOT_LENGTH
**NOT LAST FUNCTION, SO GT NEXT OFFSET AND SUBTRACT PREVIOUS
*8D2 IS PREVIOUS OFFSET
NOT_LAST:	MOVE.L	4(A3,D5.W*4),D1
	SUB.L	D2,D1	LENGTH IN D1
**NOW MOVE D1 BYTES FROM FUNCTION TO SOURCE_BUFF(A5)
GOT_LENGTH:	ADD.L	0(A3,D5.W*4),A1	POINT TO FUNCTION FILE
	MOVE.L	D1,D2
	MOVE.L	SOURCE_BUFF(A5),A0
COPY_FUNCTION:	MOVE.B	(A1)+,(A0)+
	SUBQ.L	#1,D2
	BNE.S	COPY_FUNCTION
	MOVEM.L	(SP)+,A0-A4
	CLR.L	D0
	RTS
EM_NOT_FOUND:	MOVEM.L	(SP)+,A0-A4
	MOVEQ	#-1,D0
	RTS_	"Link68_TRY_LIB "
	EVEN
***********	
*8CHECK_LIB SEARCHES LIBRARY IN A2 FOR EXTERN_NAME
*Library file header layout.
*0-127 - title/version etc etc
*128-132.L - Number of functions (.o) files in this libaray
*132->x    - number of functions * 32 byte strings with names of functions
*x->y      - number of functions * 4 byte offsets to start of each function
*The .o files follow consequtively.

CHECK_LIB:	CLR.L	D5	COUNTER
	MOVE.L	128(A2),D6	NUMBER OF FILES
	beq.s	empty_library	V3.95
	LEA	132(A2),A2	POINT TO NAMES
CL_LOOP:	LEA	32(A2),A3	NEXT FUNCTION NAME
	BSR.S	COMPARE_NAMES	COMPARE EXTERN_NAME TO A2 (32 BYTES)
	TST.W	D0	-1=NO MATCH
	BGE.S	GOT_ENAME	FOUND IT
	MOVE.L	A3,A2	POINT TO NEXT NAME
	ADDQ.L	#1,D5
	CMP.L	D5,D6
	BNE.S	CL_LOOP	
empty_library:	MOVEQ	#-1,D0
	RTS
GOT_ENAME:	NOP
	CLR.L	D0
	RTS_	"Link68_CHECK_LIB"
	EVEN
COMPARE_NAMES:	LEA	EXTERN_NAME(PC),A4
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CLR.L	D0
	RTS
CN_FAILED:	MOVEQ	#-1,D0
	RTS_	"Link68_CMP_NAME"
	EVEN		

GET_O_NAME:
	MOVE.L	A0,F_NAME_ADDR(A5)	SAVE FILE NAME FROM FRAME
	LEA	LINK_NAME(PC),A1	
	clr.l	d0
	move.b	(a0)+,d0
	move.b	d0,(a1)+
	subq.l	#1,d0
GET_FNAME_L1:	
	MOVE.B	(A0)+,(a1)+	STORE FILE NAME	IN OP_NAME FOR OUR USE.
	dbra	d0,GeT_FNAME_L1
GOT_FNAME_L1:	CLR.B	(A1)+
;	MOVE.L	A0,-(SP)
*8check for quiet mode
	tst.l	quiet_flag(A5)
	BNE.S	BUILD_QUIET_2	YES, QUIET PLEASE
	LEA	LINK_NAME(PC),A0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper2
	cmpi.b	#"z",d0
	bgt.s	is_upper2
	sub.b	#$20,(a0)	make upper
is_upper2:

	BSR.s	PRINTIT
	lea	space_text(pc),a0
	bsr.s	printit
	move.l	input_file_size(a5),d0
	bsr.s	printnum
	lea	bytes_text(pc),a0
	BSR.s	PRINTIT
BUILD_QUIET_2:
;	MOVE.L	(SP)+,A0
;	SUBQ.L	#1,A0	REWIND TO END OF LAST LINE! (DICK HEED)
	MOVE.L	FILE_NAMES(A5),A1
**COPY NAME TO RIGHT PLACE IN NAMES LIST
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	LSL.L	#5,D0	EACH ENTRY 32 BYTES
	ADD.L	D0,A1	POINT TO RIGHT PLACE
	LEA	LINK_NAME(PC),A2
	addq.l	#1,a2	*skip pascal count byte
COPY_LINK_NAME:	MOVE.B	(A2)+,(A1)+
	BNE.S	COPY_LINK_NAME
	RTS		A0-> NEXT LINE
space_text:	cstring	"  "
	align
bytes_text:	cstring	" bytes",13
	align
***********
**CLEAR_OUT MOVES 33K OF DATA FROM CODE_START TO SOURCE_BUFF JUST TO CLEAR
**OUT ANY DATA CACHES...
CLEAR_DATA_CACHE:
	MOVE.L	CODE_START(A5),A0
	MOVE.L	SOURCE_BUFF(A5),A1
	MOVE.L	#33000/4,D0
CO_LOOP:	MOVE.L	(A0)+,D1	GET IT INTO PROCESSOR
	NOP		SYNC PIPELINES
	MOVE.L	D1,(A1)+	
	DBRA	D0,CO_LOOP
	RTS		THAT SHOULD DO IT

************************

************************
**Modified for Anvil
PRINTIT:
	MOVEM.L	D6/A0-A4,-(SP)
	bsr	send_to_log
pi_end:	MOVEM.L	(SP)+,D6/A0-A4
	RTS_	"Link68_printit"
	
**LIMITED TO 1E7
**USES COMMAS FOR SEPERATORS.
PRINTNUM:
**FIND SIZE OF NUMBER TO QUASH LEADING ZEROS
	LEA	NEW_CHAR_BUFF(PC),A0	WHERE CHARS ARE	STORED FOR PRINTING
	clr.l	(a0)
	clr.l	4(a0)
	clr.l	8(a0)
	clr.l	12(a0)	*clear old string
	CMPI.L	#$FFFF,D0
	BGT.S	PRINT_LONG	LONG WORD PRINT
	MOVEQ	#0,D6
	MOVE.W	D0,D6
	CMPI.L	#10,D6
	BLT.S	DO_UNITS
	CMPI.L	#100,D6
	BLT.S	DO_TENS
	CMPI.L	#1000,D6
	BLT.S	DO_HUNS
	CMPI.L	#10000,D6
	BLT.S	DO_THOU
**NOW DIVIDE BY RELEVANT	UNITS
LONG_ENT:	DIVS	#10000,D6	GET TENS OF THOUDSANDS
	BSR.S	CONVERT_N	PRINT NUMBER OF TENS OF THOUSANDS
DO_THOU:	DIVS	#1000,D6	GET THOUSANDS
	BSR.S	CONVERT_N	ETC
	MOVE.B	#",",(A0)+
DO_HUNS:	DIVS	#100,D6
	BSR.S	CONVERT_N
DO_TENS:	DIVS	#10,D6
	BSR.S	CONVERT_N
DO_UNITS:	BSR.S	CONVERT_N
	CLR.B	(A0)
	LEA	NEW_CHAR_BUFF(PC),A0
	BSR	PRINTIT
	RTS
CONVERT_N:	ADDI.B	#48,D6	MAKE CHARACTER	(ASCII)
	MOVE.B	D6,(A0)+
	SWAP	D6
	ANDI.L	#$FFFF,D6
	RTS_	"Link68_PRINTNUM"
**PRINT LONG PRINT DO.L
**LIMITED TO 1E7 (10MEG)
PRINT_LONG:	CMPI.L	#9999999,D0
	BGT.S	NUMB_ERROR
	MOVE.L	D0,D1
	CMPI.L	#99999,D0
	BLT.S	DO_TEN_THOU
	CMPI.L	#999999,D0
	BLT.S	DO_HUND_THOU
	MOVE.L	#1000000,D0	GET MILLIONS
	BSR	LONG_DIV	DIVIDE D1.L BY D0.L WITH REMAINDER IN D0
	ADDI.B	#"0",D1
	MOVE.B	D1,(A0)+
	MOVE.L	D0,D6
	MOVE.B	#",",(A0)+
	BSR	PRINTIT
	MOVE.L	D6,D1
DO_HUND_THOU:	MOVE.L	#100000,D0	GET 100 THOUSANDS
	BSR	LONG_DIV
	ADDI.B	#"0",D1
	MOVE.B	D1,(A0)+
	MOVE.L	D0,D6
	CMPI.L	#$FFFF,D6	<65536?
	BLT	LONG_ENT	YES, PRINT WORD ROUTINE
	MOVE.L	D6,D1	NO,GET 10THOUSANDS
DO_TEN_THOU:	MOVE.L	#10000,D0
	BSR	LONG_DIV
	ADDI.B	#"0",D1
	MOVE.B	D1,(A0)+
	MOVE.L	D0,D6
	ANDI.L	#$FFFF,D6
	BRA	DO_THOU
NUMB_ERROR:	LEA	NUMB_TOO_BIG(PC),A0
	BSR	PRINTIT
	RTS_	"Link68_PRINT_NL"
	EVEN		
	

**needs c string in a0
send_to_log:
	move.l	a0,d1

	lea	send_buffer(pc),a0
	move.l	#320,(a0)	*command
	move.l	d1,4(A0)	*param1
	move.l	a0,-(sp)	*request parameter block to anvil
	move.l	callback(a5),a0
	jsr	(A0)
	addq.l	#4,sp			*tidy up stack (Ah)
;.fault1:
;	 move.l	my_pb_h2(a4),a0
;	 dc.w	 _disposehandle	
;	unlk	a4
	rts_	"Link68_Send_to_log"
send_buffer:	ds.b	300
	align
send_to_dialog:
	align
	reset_locals
	local.b	my_pb10,300
	link	a4,#-400
	move.l	a0,a1	*the string
	lea	my_pb10(a4),a0
	move.l	#500,(a0)			*show string ion dialog
	move.l	a1,4(a0)		*The string
	lea	tool_name(pc),a1
	move.l	a1,8(a0)		*tool name
	move.l	a0,-(sp)		*request parameter block to anvil
	move.l	callback(a5),a0
	jsr	(A0)
	addq.l	#4,sp			*tidy up stack (Ah)
	unlk	a4
	rts_	"Link68k_dialog"
tool_name:	pstring	"Message from Link68K"
	align
	
*****************************************

**LONG DIV DIVIDES D1.L BY D0.L
LONG_DIV:
 TST.L  D0
 BNE.S DIV
 DIVU.W     D0,D1
 RTS
DIV:
 BPL.S DO_DIV
 NEG.L D0
 BSR.S DO_DIV
 NEG.L D1
 RTS
DO_DIV:
 TST.L D1
 BPL.S DIV_P
 NEG.L D1
 BSR.S DIV_P
 NEG.L D1
 NEG.L D0
 RTS
DIV_P:
 CLR.L D3
 CLR.L D2
DIV2:
 CMP.L D0,D1
 BLS.S DIV1
 ADDQ.L #1,D3
 LSL.L #1,D0
 BMI.S DIV3
 BRA.S DIV2
DIV5:
 LSL.L #1,D2
 LSR.L #1,D0
DIV3:
 CMP.L D0,D1
DIV1:
 BCS.S DIV4
 SUB.L D0,D1
 ADDQ.L #1,D2
DIV4:
 DBF D3,DIV5
 MOVE.L D1,D0
 MOVE.L D2,D1
 RTS_	"Link68_old_LONG_DIV"
 	EVEN
**NEW - Needs:
**a0->FSspec
**A1->Name of file
**Outputs
**Temp_Fsspec(a5)=new fsspec
make_temp_FSspec:
	clr.w	-(sp)
	move.w	0(a0),-(sp)	*Vrefnum
	move.l	2(a0),-(sp)	*Dirid
	move.l	a1,-(sp)	*name
	pea	temp_FSspec(a5)	*where	  
	moveq	#1,d0
	dc.w	$aa52
	move.w	(Sp)+,d0
	ext.l	d0
	rts_	"Link68_Make_temp_Fsspec"
*********************************************************************************
LINK_NAME:	DS.B	40	TEMP FOR .O FILE NAME
EXTERN_NAME:	DS.B	32	LABEL NAME FIND_EXTERN IS LOOKING FOR
*********************************************************************************
link_done:	cstring	"Link68K finished.",13,13
	align
CR_LINK:	DC.B	13,0
LINK_FOLDER_TEXT:	
	DC.B	13,"Reading files from ",0
bytes_used_text:
	dc.b	" bytes used out of ",0
LINK_ERR2:	DC.B	" - **FATAL** Unable to find this file.",13,0
link_err5:	dc.b	13,"**FATAL** Not enough memory to build the code.",13,0
LINK_ERR3:	DC.B	13," - but the reference doesn't exist.",13,13,0
LINK_ERR4:	DC.B	13," - Offset too large.",13,0
no_code_text:	dc.b	"**FATAL** No code in this file.",13,0
double_global_text1:	dc.b	"**FATAL** The global label '",0
double_global_text2:
	dc.b	"' is",13,"defined more than once in this project",13,"(in another of your files).",13
	dc.b	"File being linked = ",0

double_global_text3:
	dc.b	"' is",13,"defined in a library in this project.",13,"Please use another name.",13
	dc.b	"File being linked = ",0
		
Link_err6:	dc.b	" - **FATAL** - No global definitions in this file.",13
	dc.b "A linkable file must have at least one global definition "
	dc.b ", even if it is just a dummy.",13,13,0
the_offset_text:	dc.b	" wants to",13,"link with ",0
is_too_big:	dc.b	13," but the offset is too big."
	dc.b	" Tip - if the instruction is a branch,",13,"use the long form - i.e. "
	dc.b	"BSR.L label.",13,13,0

LINK_ERR7:	DC.B	" - **FATAL** This file is an old Link (68K) file.",13,"Please remove it from the project.",13,0
Link_err8:	dc.b	" - **FATAL** This file is empty!",13,0
LINK_SPACE_TEXT:	
	DC.B	" -> ",0
error_text:	dc.b	"***ERROR*** ",0
warning_text:	dc.b	"**FATAL ERROR** ",0
is_branching_text:	dc.b	" branches to ",0
is_odd:	dc.b	" which is at an",13,"odd address! This will cause your"
	dc.b	" program to crash.",13,"Tip - put an 'align' directive at the end of",13
	dc.b	"all your assm source files.",0
starting_link:	dc.b	13,"Starting link.",13,0
no_read_libs_text:	 dc.b	 "Could not read libraries - are the paths set up correctly?",13,0

NO_WRITE_TEXT:	DC.B	"**Link errors**",13
	DC.B	0
crcrtext:	dc.b	13,13,0
inc_partition:	dc.b	"The code buffer is nearly out of space.",13,"You may want to"
	dc.b	" increase Anvils memory size.",13,13,0
	EVEN
NUMB_TOO_BIG:	DC.B	"xxxxxxxxxx",0
	align


NEW_CHAR_BUFF:	DS.B	20	SPACE FOR NUMBERS
	align
	EXTERN	SET_DIRECTORY,WRITE_MAC,send_and_wait
	EXTERN	a5_init
	extern	read_libs,do_alert,add_library
	PUBLIC	LINK_V3
	GLOBAL	LINK_NAME,CR_LINK,printit
	