*********************************************************************************
*FILE:	LINK PPC V1.00								*
*MOTHER FILE:	FANTASM3XX.S							*
*DATE:	261195	 - 1 year on from link 68k!					*
*MODDED:									*
*PROGRAMMER:	STUART BALL							*
*DESCRIPTION:	LIGHTSOFT LINK PPC V1.00 - SEE DOC FC_281094_01.		*
*INPUT:	ADDRESS OF CONTROL STRING IN A0	- SHOULD BE POINTING AT /L		*
*OUTPUT:	OBJECT FILE IN MEMORY						*
*********************************************************************************

;Linkable file header (PPC)
;0: Title/comment.
;400: Inputs
;724: Outputs
;1024:	 .L	 Length of global labels/addresses (32 bytes=1 label)
;1028:	 .L	 Length of Extern labels (32 bytes=1 label)
;1032:	 .L	 Length of Imports list (32 bytes=label, 4 bytes pc offset of toc offset).
;1036:	 .L	 Length of code pointers (32 bytes label, 4 bytes pc offset of toc offset, 4 bytes def code offset)
;1040:	 .L	 Length of data pointers (32 bytes label, 4 bytes pc offset of toc offset, 4 bytes def data offset).
;1044:	 .L	 "FLPo" - Identifier for checking.
;1048:	 .L	 Reserved for Length of  code exports.
;1052:	 .L	 Reserved for Length of exported data
;1056:	 .L	 Reserved for optional CRC check sum
;1060:	 .L	 Entry point, or -1 if not in this file
;
;1064:	 .L	 Length of code
;1068:	 .L	 Length of data
;1072:	 .L	 Reserved 1
;1076:	 .L	 reserved 2
;1080:	 .L	 reserved 3
;
;1084:	 Start of globals (28 label/4 offset)
;....:	 Start of externs (28/4)
;....:	 Start of Imports (32/4)
;....:	 Start of code pointers (32/4/4)
;....:	 Start of data pointers (32/4/4)
;....:	 Code
;....:	 Data
;EOF

**USes file_buff2 for readfile
**Uses 128k buffer for data
**uses file_buff1 for headers
**Uses file-buff3 for the imports list (sorted)
;FORMAT OF LABELS/ADDRESSES TABLES
;
;28 BYTES - LABEL
;4 BYTES  - ADDRESS RELATIVE TO 0 OF USAGE.


Link_PPC:
	reset_locals
	local.l	my_pb_h,1
	local.b	my_pb,300
	local.l	temp_1,1

	entry

	move.l	8(Sp),a1	*parameter block argv
	link	a4,#-1000
	move.l	a1,command_save(A5)
	move.l	(a1),callback(a5)
	move.l	4(a1),callback_toc(a5)
	move.l	16(a1),d0
**check command
	cmpi.w	#A_work_on_file,d0
	beq.s	add_file_to_list
	cmpi.w	#A_Go,d0
	beq	do_link
	cmpi.w	#A_init,d0
	beq	link_init
	cmpi.w	#A_cleanup,d0
	beq	clean_up
	cmpi.w	#A_prefschanged,d0
	bne.s	unknown_command
	bsr	read_link_prefs
	bra.s	unknown_command	
unknown_command:		*just forget it
	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	unlk	a4
	rts_	"LinkPPC_v4"
add_file_to_list:
	move.l	command_save(a5),a3
	move.l	20(a3),source_buff(a5)	*the file
	cmpi.l	#4,32(a3)	*List of slibs?
	beq.s	add_slibs	
	move.l	24(a3),a0	*ptr to name (pascal)
	move.l	28(a3),input_file_size(a5)
	movem.l	a1/a3,-(sp)
	bsr	GET_O_NAME	*file name
	movem.l	(sp)+,a1/a3
	move.l	28(a3),d1	*size of file
	move.l	32(a3),d0	*object=1, lib=2
	cmpi.b	#1,d0
	beq.s	add_obj
	cmpi.b	#2,d0
	beq	add_lib
	
	moveq	#-1,d0		*not handled
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	unlk	a4
	rts_	"Linkppc_add_file_to_list"
add_Slibs:
	move.l	slibs_handle(a5),a0
	dc.w	_HLock
	move.l	slibs_handle(a5),a1
	move.l	(a1),a1
	move.l	source_buff(a5),a0	*the list
	move.l	#130*127,d0		*max of 127 shared libs
	dc.w	_BlockMove
	move.l	slibs_handle(a5),a0
	dc.w	_HUnLock
	
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	unlk	a4
	rts_	"LinkPPC_add_slibs"	
add_obj:
	bsr	link_read_in	*add file
	clr.l	d0
	move.b	LINK_E_FLAG(A5),d0
	extb.l	d0		*return the final error
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	unlk	a4
	rts_	"Linkppc_add_obj"
add_lib:
	bsr	add_library
	clr.l	d0
	move.b	LINK_E_FLAG(A5),d0
	extb.l	d0		*return the final error
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return
	clr.l	4(a0)		*no return file
	unlk	a4
	rts_	"Linkppc_add_lib"
do_link:
	move.l	20(a1),d0		*Global size from build
	move.l	d0,rs_counter(a5)	*easy
	move.l	24(a1),d0
	move.w	d0,is_ppc(a5)	*needs to be 1 for CXF
	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY

;	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 6.25% of it
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m16
 	pea	mem1(pc)
 	dc.w	_debugstr
m16:
	move.l	d0,source_buff_h(a5) 
 	dc.w	_hlock
	move.l	source_buff_h(a5),a0
	move.l	(a0),a0
	move.l	a0,source_buff(a5)

	tst.l	main_entry(a5)	*have an entry?
	bne.s	have_entry
	move.b	#1,link_e_flag(a5)
	LEA	NO_entry_text(PC),A0
	BSR	PRINTIT
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog

	bra.s	link_end1	*Stop
have_entry:	   

	push	a4
	move.l	a4,our_a4(a5)	*so we can print from cxf etc.
	bsr	ppc_link_go
	pop	a4
link_end1:
	unlk	a4
	rts_	"LINKppc_do_link"
no_entry_text:	dc.b	"***FATAL***",13,"Code entry point has not been",13
	dc.b	"specified. Use the 'entry' directive",13
	dc.b	"to do this.",13,"Stopping Link.",13,13,0
	align
	
read_lin_prefs:
	unlk	a4
	rts_	"LINKppc_read_prefs"
read_link_prefs:
	link	a4,#-100
	clr.l	-(sp)
	move.l	#"myPF",-(sp)
	move.w	#128,-(sp)
	dc.w	_getresource
	move.l	(sp)+,a1
	move.l	a1,d0
	beq.s	fault	*no resource
	move.l	(a1),a2
	move.l	(A2),quiet_flag(a5)	*the long word = 1 = alternative chroma
	move.l	4(a2),cxf_quiet_flag(a5)
	move.l	8(a2),debug_quiet_flag(a5)
**invert pref flags
	move.l	quiet_flag(a5),d0
	bchg	#0,d0
	move.l	d0,quiet_flag(a5)
	move.l	cxf_quiet_flag(a5),d0
	bchg	#0,d0
	move.l	d0,cxf_quiet_flag(a5)
	move.l	debug_quiet_flag(a5),d0
	bchg	#0,d0
	move.l	d0,debug_quiet_flag(a5)

	move.l	a1,-(sp)
	dc.w	_releaseresource	*bye to prefs
**129 - just the one- globals exports
	move.l	#"myPF",-(sp)
	move.w	#129,-(sp)
	dc.w	_getresource
	move.l	(sp)+,a1
	move.l	a1,d0
	beq.s	fault	*no resource
	move.l	(a1),a2
	move.l	(A2),export_all_flag(a5)	*the long word = 1 = alternative chroma
	move.l	a1,-(sp)
	dc.w	_releaseresource	*bye to prefs

fault:
	unlk	a4
	rts_	"LInkPPC_read_prefs"

link_init:
	tst.b	initted?(a5)	*we initted already (on boot up for example)
	bne	skip_sign_on	*yes
**Show link title
	lea	link_hello(pc),a0
	bsr	printit
	bsr	read_link_prefs
	move.b	#1,initted?(a5)
	CLR.L	L_CODE_POS(A5)	CODE POS IN BUFFER
	CLR.L	L_NUMBER_OF_FILES(A5)
	clr.l	tot_number_of_files(A5)	*crash now mutha! (3 hours 221195)
	CLR.L	L_HEAD_POS(A5)
	CLR.L	RELOC_TAB_POS(A5)	NO RELOC INFO
	clr.w	NUMBER_OF_LIBRARIES(a5)	*no libys
	clr.b	LINK_E_FLAG(A5)
	clr.l	lib_pos(a5)		*index into libraries buffer
	clr.w	number_of_libraries(a5)
	clr.l	main_entry(a5)	   *have an entry?
	clr.l	l_data_pos(a5)	data pos in buffer
	clr.l	slibs_handle(a5)
**get memory for code buffer
	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
 	MOVE.L 	d0,FREEMEM(A5)	GET SIZE OF FREE MEMORY
 	divu.l	#18,d0
	move.l	d0,code_sz(a5)	*so we can check if we run out of memory
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m1
 	pea	mem1(pc)
 	dc.w	_debugstr
m1:
 	move.l	d0,code_buffer_h(a5)
 	dc.w	_hlock
 	move.l	code_buffer_h(a5),a0
 	move.l	(a0),a0
 	add.l	#64,a0		*Some space for our entry stuff
	move.l	a0,CODE_START(a5)
**get memory for headers
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 ;	lsr.l	#1,d0	*div 2
 **get 20% of it
 ;	lsr.l	#1,d0
 	divu.l	#12,d0
 	move.l	d0,dbg_headers_size(a5)
	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
  	bne.s	m2
 	pea	mem1(pc)
 	dc.w	_debugstr
m2:

 	move.l	d0,headers_buffer_h(a5)
 	dc.w	_hlock
 	move.l	headers_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,FILE_HEADERS(a5)
**get memory for file table
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
; 	 lsr.l	 #1,d0	 *div 2
**get 10% of it
;	lsr.l	#2,d0
	divu.l	#10,d0 
	move.l	d0,dbg_ftable_size(a5)
	
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m3
 	pea	mem1(pc)
 	dc.w	_debugstr
m3:
 	move.l	d0,ftable_buffer_h(a5)
 	dc.w	_hlock
 	move.l	ftable_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,file_table(a5)
**get memory for file names storage
 	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
 	lsr.l	#1,d0	*div 2
**get 6.25% of it
 	lsr.l	#3,d0
 	move.l	d0,dbg_fnames_size(a5)
 	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m4
 	pea	mem1(pc)
 	dc.w	_debugstr
m4:
 	move.l	d0,fnames_buffer_h(a5)
 	dc.w	_hlock
 	move.l	fnames_buffer_h(a5),a0
 	move.l	(a0),a0
	move.l	a0,FILE_NAMES(a5)

**get memory for libraries storage
	MOVE.L 	FREEMEM(A5),d0	GET SIZE OF FREE MEMORY
;	lsr.l	#1,d0	*div 2
**get 10% of it
;	lsr.l	#2,d0
 	divu.l	#11,d0
 	move.l	d0,dbg_libraries_size(a5)
	dc.w	_newhandleclear	*cant fail
 	move.l	a0,d0
 	bne.s	m5
 	pea	mem1(pc)
 	dc.w	_debugstr
m5:
 	move.l	d0,libraries_h(a5)
 	dc.w	_hlock

**PowerPC stuff
	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
 	MOVE.L 	d0,FREEMEM(A5)	GET SIZE OF FREE MEMORY
 	divu.l	#10,d0
	move.l	d0,data_sz(a5)	*so we can check if we run out of memory

	dc.w	_newhandleclear
 	move.l	a0,d0
 	bne.s	m6
 	pea	mem1(pc)
 	dc.w	_debugstr
m6:
	move.l	a0,ppc_data_buffer_handle(a5)
	dc.w	_hlock
	move.l	ppc_data_buffer_handle(a5),a0
	move.l	(a0),a0
	move.l	a0,ppc_data_buffer_pointer(a5)

	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
 	MOVE.L 	d0,FREEMEM(A5)	GET SIZE OF FREE MEMORY
**Each code def is 4 bytes
	divu.l	#140,d0
	move.l	d0,ppc_number_of_code_defs(a5)
	clr.l	d0
	move.l	ppc_number_of_code_defs(a5),d0
;	ext.l	d0
	lsl.l	#2,d0		*times 4
	move.l	d0,dbg_toc_code_offsets_size(a5)
	dc.w	_newhandlesysclear
 	move.l	a0,d0
 	bne.s	m7
 	pea	mem1(pc)
 	dc.w	_debugstr
m7:
	move.l	a0,toc_code_offsets_h(a5)	so we can delete it
	dc.w	_hlock
	move.l	toc_code_offsets_h(a5),a0	so we can delete it
	move.l	(a0),a0
	move.l	a0,toc_code_offsets(a5)
	clr.l	d0
**each data def is 16 bytes
	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
	divu.l	#120,d0		*how many long labels?
	move.l	d0,ppc_number_of_data_defs(a5)

	move.l	ppc_number_of_data_defs(a5),d0
;	ext.l	d0
	lsl.l	#4,d0	
	move.l	d0,dbg_data_item_offsets_size(a5)
	dc.w	_newhandlesysclear
 	move.l	a0,d0
 	bne.s	m8
 	pea	mem1(pc)
 	dc.w	_debugstr
m8:
	move.l	a0,data_item_offsets_h(a5)	so we can delete it
	dc.w	_hlock
	move.l	data_item_offsets_h(a5),a0	so we can delete it
	move.l	(a0),a0
	move.l	a0,data_item_offsets(a5)

	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
	divu.l	#110,d0
	move.l	d0,ppc_number_of_imports(a5)

**each import is 32 bytes
**get memory for toc_names_table (imports)
	move.l	ppc_number_of_imports(a5),d0
;	ext.l	d0
	lsl.l	#5,d0		*times 32
	move.l	d0,toc_names_table_size(a5)
	dc.w	_newhandleclear
 	move.l	a0,d0
 	bne.s	m9
 	pea	mem1(pc)
 	dc.w	_debugstr
m9:
	move.l	a0,toc_names_table_h(a5)	so we can delete it
	dc.w	_hlock
	move.l	toc_names_table_h(a5),a0
	move.l	(a0),a0
	move.l	a0,toc_names_table(a5)
**get memory for long labels
**each data def is 32 bytes
	move.l	ppc_number_of_data_defs(a5),d0
;	ext.l	d0
	lsl.l	#5,d0
	move.l	d0,dbg_long_labels_size(a5)
	dc.w	_newhandleclear
	move.l	a0,long_labels_h(a5)	so we can delete it
 	move.l	a0,d0
 	bne.s	m10
 	pea	mem1(pc)
 	dc.w	_debugstr
m10:
	dc.w	_hlock
	move.l	long_labels_h(a5),a0	so we can delete it
	move.l	(a0),a0
	move.l	a0,long_labels(a5)
**Get memory for file-buff3
**each data def is 32 bytes
	move.l	ppc_number_of_data_defs(a5),d0
;	ext.l	d0
	lsl.l	#4,d0
	move.l	d0,dbg_file_buff3_size(a5)
	dc.w	_newhandleclear
	move.l	a0,file_buff3_h(a5)	**Used as a sort area
 	move.l	a0,d0
 	bne.s	m11
 	pea	mem1(pc)
 	dc.w	_debugstr
m11:
	dc.w	_hlock
	move.l	file_buff3_h(a5),a0	so we can delete it
	move.l	(a0),a0
	move.l	a0,file_buff3(a5)

	move.l	#130*127,d0
	dc.w	_Newhandleclear	*Space for slibs. Must be a clear handle!
	move.l	a0,slibs_handle(a5)
 	move.l	a0,d0
 	bne.s	m12
 	pea	mem1(pc)
 	dc.w	_debugstr
m12:


fatalm:
skip_sign_on:
ibye:	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return

	unlk	a4
	rts_	"LINKppc_init"

crap:	ds.l	1
mem1:	pstring	"Memerr in LinkPPC. Do NOT continue! Give Anvil more memory."
	align
link_hello:	cstring	"LinkPPC (XCOFF) V4.92 [hours$]:[minutes$] on [day$]/[month$]/[year$].",13,"Build #[build$] Lightsoft 1994-1998.",13,13
	align
	
clean_up:
**bye to our memory
	move.l	file_buff3_h(a5),a0	so we can delete it
	dc.w	_disposehandle
	move.l	code_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	source_buff_h(a5),a0
	dc.w	_disposehandle		
**Lets dump some ancillary handles
	move.l	headers_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	long_labels_h(a5),a0	so we can delete it
	dc.w	_disposehandle
	move.l	toc_names_table_h(a5),a0
	dc.w	_disposehandle
	move.l	data_item_offsets_h(a5),a0
	dc.w	_disposehandle
	move.l	toc_code_offsets_h(a5),a0
	dc.w	_disposehandle
 	move.l	fnames_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	libraries_h(a5),a0
	dc.w	_disposehandle
	move.l	ppc_data_buffer_handle(a5),a0
	dc.w	_disposehandle
 	move.l	ftable_buffer_h(a5),a0
	dc.w	_disposehandle
	move.l	slibs_handle(a5),a0
	dc.w	_DisposeHandle
	clr.b	initted?(a5)
	clr.l	d0
	move.l	command_save(a5),a0
	move.l	d0,(a0)		*error return

	unlk	a4
	rts_	"LINKppc_Cleanaup"
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************
*************************************************************************

ppc_link_go:		
	clr.b	global_err(a5)
	CLR.B	LINK_E_FLAG(A5)
	tst.l	debug_quiet_flag(a5)
	bne.s	skip_debug_info
	bsr	print_stats
skip_debug_info:
	push	a4	*link uses a4
	BSR	DO_THE_LINKING	SEE DOC REF FC_301094_01	
	pop	a4
	tst.b	link_e_FLAG(A5)	any errors?
	bne	link_no_write	yes
**now create a list of globals and their offsets so cxf can export them as needed
	
	bsr	create_globals_for_export
	bsr	check_we_exported_all_in_list	*any export labels not used?
	move.l	exports_list_h(a5),a0	*goodbye to exports list
	move.l	a0,d0
	beq.s	no_exports_to_go
	dc.w	_DisposeHandle
no_exports_to_go:
	tst.b	link_e_FLAG(A5)	any errors?
	bne	link_no_write	yes

**NOW FRIG ABOUT WITH FANTASMS VARIABLES SO WRITE_XCOFF_FRAG GETS THE RIGHT BITS
	CLR.L	LT_POS(A5)	DEF NO LABELS!
	move.l	l_code_pos(A5),pc(A5)	total code size
	add.l	#64,pc(a5)	*Code starts at 64
**First rs_counter - this comes from build as globoff_counter(a5) -ve
;	move.l	globoff_counter(a5),d0
;	neg.l	d0
**next total_imports_num - this is our import_count
	move.l	import_count(a5),d0
	move.w	d0,total_imports_num(a5)
**next data_item_count - this is our data_ptrs_count
	move.l	data_ptrs_count(a5),data_item_count(a5)
**next toc_code_count - our code_ptrs_count
	move.l	code_ptrs_count(a5),toc_code_count(a5)
**next data_buffer_index which is our l_data_pos?
	move.l	l_data_pos(a5),data_buffer_index(a5)
**next entry_point which is entry_point!
**next data_buffer, which is our l_data_start
	move.l	l_data_start(a5),data_buffer(a5)
	move.l	data_buffer(a5),d1
	move.l	data_buffer_index(a5),d0
	add.l	d0,d1		*because write frag subtract index from buffer
	move.l	d1,data_buffer(a5)
**next, toc names table, which is our toc_names_table :-)
**long labels we've already allocated for above
**now we have to clear the data caches....
	move.l	file_buff3(A5),a0
	move.l	a0,source_buff(a5)	*this is where the frag is built
	
	bsr	write_xcoff
	
link_end:
	move.l	ppc_data_buffer_handle(a5),a0
	dc.w	_hunlock

	
	LEA	LINK_DONE_TEXT(PC),A0
	BSr	PRINTIT
**terminate
link_bye:
	clr.l	d0
	move.b	link_e_flag(a5),d0
	or.b	global_err(a5),d0
	move.l	command_save(a5),a0
	move.l	d0,0(a0)	*return any error
**return code frag
	move.l	file_buff3_h(a5),d0
	move.l	d0,4(a0)	*return handle
**return size of frag
	MOVE.L	xcoff_file_size(a5),8(a0)	*LENGTH (ISH)
	rts
LINK_FATAL:

	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog
	bra.s	link_bye
			
	move.l	toc_code_offsets_h(a5),a0	 *toc code entries
	move.l	a0,d0
	beq.s	skip_3
	dc.w	_hunlock
	dc.w	_disposehandle	bye...
skip_3
	move.l	data_item_offsets_h(a5),a0	*toc code entries
	move.l	a0,d0
	beq.s	skip_4
	dc.w	_hunlock
	dc.w	_disposehandle	bye...
skip_4
	move.l	toc_names_table_h(a5),a0	*toc code entries
	move.l	a0,d0
	beq.s	skip_5
	dc.w	_hunlock
	dc.w	_disposehandle	bye...
skip_5
	move.l	long_labels_h(a5),a0	*toc code entries
	move.l	a0,d0
	beq.s	skip_6
	dc.w	_hunlock
	dc.w	_disposehandle	bye...
skip_6
	tst.l	libraries_h(A5)
	beq.s	no_libs_to_go
	move.l	libraries_h(A5),a0
	dc.w	_hunlock
	dc.w	_disposehandle
no_libs_to_go:
	MOVEQ	#-1,D0	RETURN GEN ERROR TO BUILD
	RTS
LINK_NO_WRITE:
	LEA	NO_WRITE_TEXT(PC),A0
	BSR	PRINTIT
	BRA.S	LINK_FATAL
	rts

fatal:
	lea	link_no_mem(pc),a0
	bsr	printit
	bra	link_fatal
print_stats:
	lea	code_size_text(pc),a0
	bsr	printit
	move.l	code_sz(a5),d0
	bsr	printnum
	lea	data_size_text(pc),a0
	bsr	printit
	move.l	data_sz(a5),d0
	bsr	printnum

	lea	code_defs_text(pc),a0
	bsr	printit
	move.l	ppc_number_of_code_defs(a5),d0
	bsr	printnum
	lea	data_defs_text(pc),a0
	bsr	printit	
	move.l	ppc_number_of_data_defs(a5),d0
	bsr	printnum

	lea	imports_text(pc),a0
	bsr	printit	
	move.l	ppc_number_of_imports(a5),d0
	bsr	printnum

	lea	exports_text(pc),a0
	bsr	printit
		
	move.l	dbg_libraries_size(a5),d0
	lsr.l	#5,d0
	bsr	printnum

	lea	dhs(pc),a0
	bsr	printit	
	move.l	dbg_headers_size(a5),d0
	bsr	printnum

	lea	dfs(pc),a0
	bsr	printit	
	move.l	dbg_ftable_size(a5),d0
	bsr	printnum
	
	lea	dtn(pc),a0
	bsr	printit	
	move.l	dbg_fnames_size(a5),d0
	bsr	printnum

	lea	dln(pc),a0
	bsr	printit	
	move.l	dbg_libraries_size(a5),d0
	bsr	printnum

	lea	dtc(pc),a0
	bsr	printit	
	move.l	dbg_toc_code_offsets_size(a5),d0
	bsr	printnum

	lea	dds(pc),a0
	bsr	printit	
	move.l	dbg_data_item_offsets_size(a5),d0
	bsr	printnum

	lea	toc_names(pc),a0
	bsr	printit	
	move.l	toc_names_table_size(a5),d0
	bsr	printnum

	lea	dlong(pc),a0
	bsr	printit	
	move.l	dbg_long_labels_size(a5),d0
	bsr	printnum

	lea	df3(pc),a0
	bsr	printit	
	move.l	dbg_file_buff3_size(a5),d0
	bsr	printnum
	lea	freememt(pc),a0
	bsr	printit	

	lea	freemem(a5),a1
	DC.W	_MAXMEM	GET ALL FREE MEMORY
	bsr	printnum
	rts_	"LinkPPC_prnt_stats"
code_size_text:	dc.b	13,"Code buffer size: ",0
	align
data_size_text:	dc.b	13,"Data buffer size: ",0
	align
code_defs_text:	dc.b	13,"Number of code defs allowed: ",0
	align
data_defs_text:	dc.b	13,"Number of data defs allowed: ",0
	align
imports_text:	dc.b	13,"Number of imports allowed: ",0
	align
exports_text:	dc.b	13,"Number of exports allowed: ",0
	align

dhs:	dc.b	13,13,"headers size             ",0

dfs:	dc.b	13,"file table size          ",0
dtn:	dc.b	13,"file names size          ",0
dln:	dc.b	13,"libraries/exports size   ",0
dtc:	dc.b	13,"toc_code_offsets size    ",0
dds:	dc.b	13,"data_items_offsets size  ",0
dlong: dc.b	13,"long_labels size         ",0
df3:	dc.b	13,"code size                ",0
toc_names:
	dc.b	13,"toc_names size           ",0
freememt:	dc.b	13,"Free memory after init:  ",0
	align

send_2:
	lea	my_pb(a4),a0
	move.l	d0,(a0)	*command
	move.l	d1,4(A0)	*param1
	move.l	d2,8(a0)	*param2
	move.l	a0,-(sp)	*request parameter block to anvil
	move.l	callback(a5),a0
	jsr	(A0)
	addq.l	#4,sp			*tidy up stack (Ah)
fault2:
	rts_	"linkppc_Send_2"
project_exports:	pstring	"project_exports"
	align
**a0=label to search for
**a1=pointer into exports list
**returns label in export_label(a5)
**        d0=-ve=eof, =0=not a label, +ve=label
get_label_from_list:
	move.b	(a1),d0
	beq.s	eof_export_list	*eof
	cmpi.b	#13,d0
	beq.s	blank_line
	cmpi.b	#0xfe,d0
	beq.s	blank_line	*we overwrite cr's to indicate a used label
	cmpi.b	#"*",d0
	beq.s	comment_line
	cmpi.b	#";",d0
	beq.s	comment_line
	lea	export_label(a5),a2
**copy label
get_lab:
	move.b	(a1)+,d0
	beq.s	got_lab_eol
	cmpi.b	#13,d0
	beq.s	got_lab_eol
	cmpi.b	#0xfe,d0
	beq.s	got_lab_eol
	cmpi.b	#" ",d0
	beq.s	got_lab
	cmpi.b	#9,d0
	beq.s	got_lab
	move.b	d0,(a2)+
	bra.s	get_lab
glfl_end:
	rts_	"linkppc_Get_label_from_list"
got_lab:
	clr.b	(a2)
	subq.l	#1,a1
	bsr.s	find_eol
	inc.l	a1	*point to next line
	moveq	#1,d0	*return value
	bra	glfl_end
**here if we've copied a label
got_lab_eol:
	clr.b	(a2)
	moveq	#1,d0
	bra	glfl_end
**here if line starts with ; or *
comment_line:
	bsr.s	find_eol
**here if line is a cr
blank_line:
	clr.l	d0
	add.l	#1,a1	*point to next line
	bra	glfl_end
**here if first char was a zero
eof_export_list:
	moveq	#-1,d0
	bra	glfl_end
****************************************************
find_eol:
	inc.l	a1
check_next_eol:
	move.b	(a1),d0
	cmpi.b	#13,d0
	beq.s	got_eol
	cmpi.b	#0xfe,d0
	beq.s	got_eol
	cmpi.b	#0,d0
	beq.s	got_eol
	inc.l	a1
	bra.s	check_next_eol
got_eol:
	rts_	"linkppc_Find_eol"
***********************************************************************************
create_globals_for_export:
*get the exports list from the project
*a_find_this_file
	clr.l	exports_list_h(a5)
	clr.b	have_exports_list(a5)	*flag
	move.l	#A_find_this_file,d0
	moveq	#6,d1	*search the files defined in the project
	lea	project_exports(pc),a1
	move.l	a1,d2			*The alias
	bsr	send_2
	lea	my_pb(a4),a0
	tst.l	(a0)
	bmi.s	no_exports_file
**copy fsspec to temp

	push	a0
	move.l	4(a0),a0
	lea	temp_fsspec(a5),a1
	move.l	#80,d0
	dc.w	_BlockMove
	pop	a0
*get size of file
	move.l	#A_find_this_file,d0
	moveq	#4,d1	*we have an ffspec now
	move.l	4(a0),d2		*The fsspec
	bsr	send_2
	lea	my_pb(a4),a0
	tst.l	(a0)
	bmi.s	no_exports_file
**size of file in 8(a0)
	move.l	8(a0),d0
	add.l	#32,d0
	dc.w	_NewHandleClear
	move.l	a0,exports_list_h(a5)
	dc.w	_Hlock
**read the file
	move.l	exports_list_h(a5),a0
	move.l	#A_read_file,d0
	move.l	(a0),d1	*ptr to where data to go
	lea	temp_fsspec(a5),a0
	move.l	a0,d2		*ptr to The fsspec
	bsr	send_2	*file now in exports_list_h
	move.b	#1,have_exports_list(a5)
	clr.l	export_all_flag(a5)	*override prefs if we have a list
	move.b	#1,havent_done_first_lab(A5)	*for first lab being All, all, ALL	
;	lea	my_pb(a4),a0
;	tst.l	(a0)
;	bmi	no_exports_file	*error
	
no_exports_file:
**we create this list in libraries_h which is finished with by now
**Size in dbg_libraries_size
**We get the globals from file headers. Top order bit of byte 28 is set for data which we don't export
	move.l	libraries_h(a5),a0
	dc.w	_Hlock
	clr.l	number_of_globs_present(a5)
	move.l	l_number_of_files(a5),d7
	subq.l	#1,d7			*one over!:	
cgfe_loop:
	move.l	file_table(a5),a0
	move.l	d7,d0
	muls	#12,d0
	MOVE.L	4(A0,D0.L),D1	HEADER START OF THIS FILE
	move.l	d1,d0
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D0,A0		*a0->header info
	MOVE.L	(A0)+,D1	SIZE OF GLOBALS
	MOVE.L	(A0)+,D2	SIZE OF EXTERNS
	lea	12(a0),a0	*skip imports,code and data sizes
	ADD.L	D2,A0	SKIP EXTERNALS TO GLOBALS

	tst.l	d1
	beq	no_globs_in_this_file
	lsr.l	#5,d1
	move.l	d1,d6
	subq.l	#1,d6	*globs in this file counter
cop_glob:
	btst	#7,28(A0)
	bne	its_data
	
**Check name against export list if not export all
**name in a0
	move.b	#1,can_export(a5)
	tst.l	export_all_flag(a5)
	bne	no_el	*export all has been set
	clr.b	can_export(a5)	*set to 1 if this label is defined in the list
	tst.b	have_exports_list(a5)
	beq	no_el
	save_all
**check name is in list, if it is, set can export

	move.l	exports_list_h(a5),a1
	move.l	(a1),a1	*ptr to list
	tst.b	havent_done_first_lab(A5)
	beq.s	done_first_lab
**Check if first label in exports list is All, all or ALL
get_first_lab:

	push	a0	*labels were searching for
	bsr	get_label_from_list	*a1 is pointing at next line. d0=-1,eof, 0=not label, +=label
	pop	a0			
	tst.w	d0	*eof
	bmi.s	eofl_list	*yes
	beq.s	get_first_lab	*comment or blank line	
**label in export_label

	push	a0
	lea	export_label(a5),a0
	lea	all1(pc),a1	*all
	bsr	strcmp	*lib
	pop	a0
	tst.w	d0
	beq.s	got_all

	push	a0
	lea	export_label(a5),a0
	lea	all2(pc),a1	*All
	bsr	strcmp	*lib
	pop	a0
	tst.w	d0
	beq.s	got_all

	push	a0
	lea	export_label(a5),a0
	lea	all3(pc),a1	*ALL
	bsr	strcmp	*lib
	pop	a0
	tst.w	d0
	bne.s	not_all
got_all:	*Export all - first label is "All"
	move.l	#1,export_all_flag(a5)
	clr.b	havent_done_first_lab(a5)
	st	can_export(a5)	*set to 1 if this label is defined in the list

	restore_all
	bra.s	no_el
**not all
not_all:
	clr.b	havent_done_first_lab(A5)
eofl_list:
	move.l	exports_list_h(a5),a1
	move.l	(a1),a1	*restore ptr to list

done_first_lab:	
**get this label
get_next_lab:
	push	a0	*labels were searching for
	bsr	get_label_from_list	*a1 is pointing at next line. d0=-1,eof, 0=not label, +=label
	pop	a0			
	tst.w	d0	*eof
	bmi.s	eo_list
	beq.s	get_next_lab	
**label in export_label
	push	a1	*save current pos in list, pointing at next pos in list, cr is previous character
	push	a0
	lea	export_label(a5),a1
	bsr	strcmp	*lib
	pop	a0
	pop	a1	*restore current pos in list
	tst.w	d0
	beq.s	export_it!
	bra.s	get_next_lab
export_it!:
*we need to mark the label as exported now...
*we can do this by seting a1-1 to a special character, say $fe
*yes, we overwrite the cr character
	move.b	#0xfe,-1(a1)	*overwite CR with $fe to mark as used
	move.b	#1,can_export(a5)
eo_list:
	restore_all
no_el:
**now copy name to glob list
	tst.b	can_export(a5)	*defined in list/export all?
	beq.s	its_data	*dont export if not in list or not all
	push	a0
	move.l	libraries_h(a5),a1
	move.l	(a1),a1
	move.l	number_of_globs_present(a5),d2
	lsl.l	#5,d2
	add.l	d2,a1
	moveq	#32,d0
	dc.w	_BlockMove
	pop	a0	*just in case...
	inc.l	number_of_globs_present(a5)
its_data:
	lea	32(a0),a0
	dbra	d6,cop_glob

no_globs_in_this_file:
	subq.l	#1,d7
	bge	cgfe_loop
	move.l	libraries_h(a5),a0
	dc.w	_HUnlock

	rts_	"LinkPPC_Create_export_list"
all1:	cstring	"all"
	align
all2:	cstring	"All"
	align
all3:	cstring	"ALL"
	align
**this routine checks we did actually export all the labels in the export list that we wanted to
check_we_exported_all_in_list:
	tst.l	export_all_flag(a5)
	bne.s	eoc_list	*no check if export all
	move.l	exports_list_h(a5),a1
	move.l	a1,d0
	beq.s	eoc_list	*what exports list?
	move.l	(a1),a1	*ptr to list
check_next_lab:
	bsr	get_label_from_list	*a1 is pointing at next line. d0=-1,eof, 0=not label, +=label
	tst.w	d0	*eof
	bmi.s	eoc_list
	beq.s	check_next_lab	
**label in export_label
	cmpi.b	#0xfe,-1(a1)
	beq.s	check_next_lab
	save_all	*save current pos in list, pointing at next pos in list, cr is previous character
	lea	export_label(a5),a0	*name of label we couldn't export
	bsr	printit
	lea	cant_export_text(pc),a0
	bsr	printit
	st	link_e_FLAG(A5)	*set error


	restore_all
	bra.s	check_next_lab	*carry on number 1
eoc_list:
	rts_	"LinkPPC_check_we_exported"
cant_export_text:	
	dc.b	" - Error ",13
	dc.b	"This label is defined as an export",13
	dc.b	"in project_exports but it is not",13
	dc.b	"defined as a global code label.",13,13,0
	align
linker_errors_string:	cstring	"LinkPPC errors.",13,"Please check the Log window."
	align

***************
DO_THE_LINKING:
	lea	starting_link(pc),a0
	bsr	printit
link_restart:
	clr.w	rebuild_flag(a5)	*come here if lib func imported
**First sort out imports
**here we have to build a "master list" of imports, giving each one a number (4+n*4).
**We can use the normal search algorithms to see if the import is already defined, and
**if so, we insert its number in the import list in each header.
**We'll do it backwards from l_number_of_files down.
**The buffer we'll use is: file_buff3
	clr.l	import_count(a5)	*actual number of entries in import_table
	move.l	l_number_of_files(a5),d7
	subq.l	#1,d7			*one over!
calc_imports_loop:	
	move.l	file_table(a5),a0
	move.l	d7,d0
	muls	#12,d0
	MOVE.L	4(A0,D0.L),D0	HEADER START OF THIS FILE
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D0,A0	POINT TO HEADER START FOR THIS FILE
	move.l	0(a0),d1	*size of globals
	add.l	4(a0),d1	*size of externs
	move.l	8(a0),d2	*number of imports
	beq.s	no_imports
	divs	#36,d2		*actual number of 36 byte imports
	subq.l	#1,d2
	lea	20(a0),a0	*skip sizes
	add.l	d1,a0		*point to imports (36 bytes a piece)
do_import:
	move.l	toc_names_table(a5),a1	*labs go here
**see if import is defined in table
	move.l	a0,-(sp)
	bsr	sort_import	a1=table,a0=import name
	move.l	(sp)+,a0
	tst.w	d0
	bmi	link_finished	*too many imports!
	lea	36(a0),a0	*point to next import
	dbra	d2,do_import	
no_imports:
	subq.l	#1,d7
	bge	calc_imports_loop	*do next file
**Now we have a Master List of imports, we can go through the headers replacing the toc
**offsets with the actual table positions +4 (for the BSS pointer).
**we examine the offset of the toc offset, and modify the code directly!

	move.l	l_number_of_files(a5),d7
	subq.l	#1,d7			*one over!
reloc_imports_loop:	
	move.l	file_table(a5),a0
	move.l	d7,d0
	muls	#12,d0
	MOVE.L	4(A0,D0.L),D1	HEADER START OF THIS FILE
	move.l	0(a0,d0.l),d2	Code start for this file
	MOVE.L	CODE_START(A5),A3	POINT TO CODE
	add.l	d2,a3		*a3->code
	move.l	d1,d0
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D0,A0		*a0->header info
	move.l	0(a0),d1	*size of globals
	add.l	4(a0),d1	*size of externs
	move.l	8(a0),d2	*number of imports
	beq.s	no_imports_to_reloc
	divs	#36,d2		*actual number of 36 byte imports
	lea	20(a0),a0	*skip sizes
	add.l	d1,a0		*point to imports (36 bytes a piece)
**here a3 is code pointer for this file.
**     a0 points to the imports in header
**     d2 is number of imports - 1
**Now we have to search the master list to find the labels position, multiply it by 4
**and add 4 for the BSS pointer (toc[0])
reloc_next_import:
	move.l	a3,-(sp)		*save code ptr	
	bsr	get_master_list_pos	*in d0 as pos*4+4
	move.l	(sp)+,a3		*restore code base ptr
**now mould code
**32(a0) is the pointer
**The toc offset goes in the lower 16 bits
	move.l	32(a0),d1		*get inst offset
	move.w	d0,2(a3,d1.l)		*mix in toc offset to inst	
**now do next import
	lea	36(a0),a0		*point to next import
	dec.l	d2
	bne.s	reloc_next_import		
no_imports_to_reloc:
	subq.l	#1,d7
	bge	reloc_imports_loop	*do next file


***Now the data pointers.
**We really need the data pointers to be global. Fantasm assumes that when extern
**is used, the data is a code pointer - hence we need a new directive called "extern_data"
**that tells fant this label is really a data pointer, and will be found in the 
**linker somewhere.
**So in the headers, an external data item is represented by -1.L at offset 36
	clr.l	data_ptrs_count(a5)	*actual number of entries in data_ptrs_table
	move.l	l_number_of_files(a5),d7
	subq.l	#1,d7			*one over!

calc_dp_loop:
	move.l	file_table(a5),a0
	move.l	d7,d0
	muls	#12,d0
	MOVE.L	4(A0,D0.L),D0	HEADER START OF THIS FILE
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D0,A0	POINT TO HEADER START FOR THIS FILE
	move.l	0(a0),d1	*size of globals
	add.l	4(a0),d1	*size of externs
	add.l	8(a0),d1	*size of imports
	add.l	12(a0),d1	*size of code ptrs
	move.l	16(a0),d2	*number of data ptrs
	beq.s	no_dp
	divu.l	#40,d2		*actual number of 40 byte data ptrs
	lea	20(a0),a0	*skip sizes
	add.l	d1,a0		*point to data ptrs (40 bytes a piece)
**build the master list of code ptrs
dp_loop:
	move.l	file_buff3(a5),a1	*labs go here
**see if dp is defined in table and add offset to toc_data_offsets(a5)
	move.l	a0,-(sp)
	bsr	sort_dps	a1=table,a0=import name
	move.l	(sp)+,a0
	tst.w	d0
	bmi	link_finished	*too many data ptrs!	
	lea	40(a0),a0	*point to next label
	dec.l	d2
	bne.s	dp_loop		
	move.l	file_buff3(a5),a1	*labs go here
	
no_dp:	subq.l	#1,d7
	bge	calc_dp_loop	*do next file
**Now we have a Master List of data ptrs, we can go through the headers replacing the toc
**offsets with the actual table positions +4 (for the BSS pointer)
**we examine the offset of the toc offset, and modify the code directly!
	move.l	l_number_of_files(a5),d7
	subq.l	#1,d7			*one over!
reloc_dp_loop:	
	move.l	file_table(a5),a0
	move.l	d7,d0
	muls	#12,d0
	MOVE.L	4(A0,D0.L),D1	HEADER START OF THIS FILE
	move.l	0(a0,d0.l),d2	Code start for this file
	MOVE.L	CODE_START(A5),A3	*a3->code for this file
	move.l	d1,d0
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D0,A0		*a0->header info
	move.l	0(a0),d1	*size of globals
	add.l	4(a0),d1	*size of externs
	add.l	8(a0),d1	*number of imports
	add.l	12(a0),d1	*size of cps
	move.l	16(a0),d6	*size of dps
	beq	no_dps_to_reloc
	divs	#40,d6		*actual number of 40 byte data ptrs
	lea	20(a0),a0	*skip sizes
	add.l	d1,a0		*point to data ptrs

**here a3 is code pointer for this file.
**     a0 points to the data ptrs in header
**     d2 is number of code ptrs - 1
**Now we have to search the master list to find the labels position, multiply it by 4
**and add 4 for the BSS pointer (toc[0]) and add number of imports
reloc_next_dp:
	move.l	a3,-(sp)		*save code ptr	
	bsr	get_master_dp_list_pos	*in d0 as pos*4+4
	move.l	(sp)+,a3		*restore code base ptr
	tst.w	d0
	bgt.s	dp_found1
**here the extern_data was NOT found, so we try libs
	movem.l	d6-d7/a3,-(sp)
	bsr	search_libs_for_data	*try libs - a0 points to data name
	movem.l	(sp)+,d6-d7/a3

	tst.w	d0
	bge	link_restart		*lib read in, so data should be found next time round
	tst.b	link_e_flag(a5)	*fatal error (v5)
	bne	link_finished
	save_all
	move.l	d7,-(sp)	*for print filename
	bsr	printit		*dp not found! - label name
	lea	dp_not_found(pc),a0
	bsr	printit		*declared as extern_data in '
	move.l	(sp)+,d7	*for print filename
	bsr.l	print_file_name_d7	*filename
	lea	but_not_found(pc),a0	*but has not been defined
	bsr	printit
	btst	#1,build_options(a5)
	bne.s	not_lib1	*no libs
	lea	or_libs(pc),a0	*print "or library function"
	bsr	printit
not_lib1:
	lea	lib_cr(pc),a0	*.crcr
	bsr	printit
	st	LINK_E_FLAG(A5)
	restore_all

	clr.l	d0		*use zero
dp_found1:
	
**now mould code
**32(a0) is the pointer
**The toc offset goes in the lower 16 bits
	move.l	32(a0),d1		*get inst offset
	cmpi.l	#-4,d1
	beq.s	dont_reloc_dummy
	move.w	d0,2(a3,d1.l)		*mix in toc offset to inst	
**now do next import
dont_reloc_dummy:
	lea	40(a0),a0		*point to next dp
	dec.l	d6
	bne	reloc_next_dp		
no_dps_to_reloc:
	subq.l	#1,d7
	bge	reloc_dp_loop	*do next file

**now we have to do the code ptrs. These have a base offset of 
**4+number_of_imports_in_master_list*4+number of data ptrs*4.
**The problem here is what if the user has used two toc-routines with the same name?
**Do we halt the proceedings or do we change the name of one of the names?
**I believe it to be better we alert the user to the fact.
**How do we notify Fant that a toc routine is defined in another file?
**I guess with the extern directive. For this to work Fant has to understand that when it
**gets a label [T]name it must check the label table if not found in the [t] table.
**Eval must then insert the label name into the toc table, and insert its code offset as -1.L
**hence, when we are buiding the master list of of code pointers, we have to take the code
*offset ONLY if it isn't -1.L.
**Then we can go through the headers, looking for -1's and replace the toc offset with the
**right offset for the label, and build the code offset table correctly.
	clr.l	code_ptrs_count(a5)	*actual number of entries in code_ptrs_table
	move.l	l_number_of_files(a5),d7
	subq.l	#1,d7			*one over!

calc_cp_loop:	
	move.l	file_table(a5),a0
	move.l	d7,d0
	muls	#12,d0
	MOVE.L	4(A0,D0.L),D0	HEADER START OF THIS FILE
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D0,A0	POINT TO HEADER START FOR THIS FILE
	move.l	0(a0),d1	*size of globals
	add.l	4(a0),d1	*size of externs
	add.l	8(a0),d1	*size of imports
	move.l	12(a0),d2	*number of code ptrs
	beq.s	no_cp
	divs	#40,d2		*actual number of 40 byte code ptrs
	lea	20(a0),a0	*skip sizes
	add.l	d1,a0		*point to code ptrs (40 bytes a piece)
**build the master list of code ptrs
cp_loop:
	move.l	file_buff3(a5),a1	*labs go here
**see if code is defined in table and add offset to toc_code_offsets(a5)
	move.l	a0,-(sp)
	bsr	sort_cps	a1=table,a0=toc-ptr_name
	move.l	(sp)+,a0
	tst.w	d0
	bmi	link_finished	*too many code ptrs!	
	lea	40(a0),a0	*point to next label
	dec.l	d2
	bne.s	cp_loop		
	move.l	file_buff3(a5),a1	*labs go here

**We have to build the code offset table that write_xcoff_frag needs (toc_code_offsets)
	
no_cp:	subq.l	#1,d7
	bge	calc_cp_loop	*do next file
**Now we have a Master List of code ptrs, we can go through the headers replacing the toc
**offsets with the actual table positions +4 (for the BSS pointer).
**we examine the offset of the toc offset, and modify the code directly!
	move.l	l_number_of_files(a5),d7
	subq.l	#1,d7			*one over!
reloc_cp_loop:	
	move.l	file_table(a5),a0
	move.l	d7,d0
	muls	#12,d0
	MOVE.L	4(A0,D0.L),D1	HEADER START OF THIS FILE
	move.l	0(a0,d0.l),d2	Code start for this file
	MOVE.L	CODE_START(A5),A3	*a3->code for this file
	move.l	d1,d0
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D0,A0		*a0->header info
	move.l	0(a0),d1	*size of globals
	add.l	4(a0),d1	*size of externs
	add.l	8(a0),d1	*number of imports
	move.l	12(a0),d6
	beq.s	no_cps_to_reloc
	divs	#40,d6		*actual number of 40 byte code ptrs
	lea	20(a0),a0	*skip sizes
	add.l	d1,a0		*point to code ptrs
**here a3 is code pointer for this file.
**     a0 points to the code ptrs in header
**     d2 is number of code ptrs - 1
**Now we have to search the master list to find the labels position, multiply it by 4
**and add 4 for the BSS pointer (toc[0]) and add number of imports and data ptrs
reloc_next_cp:
	move.l	a3,-(sp)		*save code ptr	
	bsr	get_master_cp_list_pos	*in d0 as pos*4+4
	move.l	(sp)+,a3		*restore code base ptr
	tst.w	d0
	bgt.s	cp_found1
	tst.b	link_e_flag(a5)	*fatal error (v5)
	bne	link_finished

	save_all
	move.l	d7,-(sp)	*for print filename
	bsr	printit		*dp not found! - label name
	lea	cp_not_found(pc),a0
	bsr	printit		*declared as extern_data in '
	move.l	(sp)+,d7	*for print filename
	bsr.l	print_file_name_d7	*filename
	lea	but_not_found(pc),a0	*but has not been defined
	bsr	printit
	btst	#1,build_options(a5)
	bne.s	not_lib2
	lea	or_libs(pc),a0
	bsr	printit
not_lib2:
	lea	lib_cr(pc),a0
	bsr	printit	
	st	LINK_E_FLAG(A5)
	restore_all
	clr.l	d0		*use zero

cp_found1:	
**now mould code
**32(a0) is the pointer
**The toc offset goes in the lower 16 bits
	move.l	32(a0),d1		*get inst offset
	cmpi.l	#-3,d1
	beq.s	its_a_dummy
	move.w	d0,2(a3,d1.l)		*mix in toc offset to inst	
**now do next import
its_a_dummy:				*Come here if a dummy def is found
	lea	40(a0),a0		*point to next import
	dec.l	d6
	bne.s	reloc_next_cp		
no_cps_to_reloc:
	subq.l	#1,d7
	bge	reloc_cp_loop	*do next file

**NOW WE ACTUALLY "DO THE LINK" - replace externs with real offset
**NOTE THAT LIBRARY FUNCTION READS INCREASE TOT_NUMBER_OF_FILES
**Now link up the globals/externals in the code
the_link:
	MOVE.L	L_NUMBER_OF_FILES(A5),TOT_NUMBER_OF_FILES(A5)
	CLR.L	D2	FILE COUNTER
	clr.l	link_file_counter(a5)	for errors

LINK_LOOP:
	MOVE.L	FILE_TABLE(A5),A0
	move.l	d2,d3
	muls	#12,d3
	MOVE.L	0(A0,D3.l),D0	CODE START OF THIS FILE
	MOVE.L	4(A0,D3.L),D1	HEADER START OF THIS FILE

	MOVE.L	D2,-(SP)
	BSR	LINK_FILE	LINK THIS FILE
	MOVE.L	(SP)+,D2
	tst.b	link_e_flag(a5)
	bne.s	link_finished
	tst.w	rebuild_flag(a5)
	bne	link_restart	*lib func read in, so rebuild
	ADDQ.L	#1,D2
	addq.l	#1,link_file_counter(a5)
	CMP.L	TOT_NUMBER_OF_FILES(A5),D2
	BLT.S	LINK_LOOP
LINK_FINISHED:	RTS_	"LinkPPC_do_the_link"

**a0->name
get_master_list_pos:
	moveq.l	#4,d1			*imports position*4+4
	move.l	toc_names_table(a5),a1	*the master list
	move.l	a1,a3			*save position in master list
find_master_pos:
	move.l	a3,a4		*table
	move.l	a0,a2		*reset name
	bsr	compare_32	*d0=0 found, -1 not found a2,a4
	tst.w	d0
	beq.s	imp_found_in_ML
	lea	32(a3),a3	*point to next entry
	addq.l	#4,d1		*inc position
	bra.s	find_master_pos	
imp_found_in_ML:
	move.l	d1,d0	*position*4+4
	rts_	"LinkPPC_Search_master_list"
**a0->name
get_master_cp_list_pos:
	moveq.l	#4,d1			*imports position*4+4
	move.l	import_count(a5),d0
	lsl.l	#2,d0
	add.l	d0,d1
	move.l	data_ptrs_count(a5),d0
	lsl.l	#2,d0
	add.l	d0,d1
	move.l	file_buff3(a5),a1	*the master list
	move.l	a1,a3			*save position in master list
	move.l	code_ptrs_count(a5),d2	*count of entries in list
find_cp_master_pos:
	move.l	a3,a4		*table
	move.l	a0,a2		*reset name
	bsr	compare_32	*d0=0 found, -1 not found a2,a4
	tst.w	d0
	beq.s	cp_found_in_ML
	lea	32(a3),a3	*point to next entry
	addq.l	#4,d1		*inc position
	dbra	d2,find_cp_master_pos	
	moveq	#-1,d0
	rts			*not found
cp_found_in_ML:
	move.l	d1,d0	*position*4+4+(number of imports*4)
	rts_	"LINKPPC_Srch_mast_CP_list"

**a0->name
get_master_dp_list_pos:
	moveq.l	#4,d1			*imports position*4+4
	move.l	import_count(a5),d0
	lsl.l	#2,d0
	add.l	d0,d1
	move.l	file_buff3(a5),a1	*the master list
	move.l	a1,a3			*save position in master list
	move.l	data_ptrs_count(a5),d2	*count of entries in list
find_dp_master_pos:
	move.l	a3,a4		*table
	move.l	a0,a2		*reset name
	bsr	compare_32	*d0=0 found, -1 not found a2,a4
	tst.w	d0
	beq.s	dp_found_in_ML
	lea	32(a3),a3	*point to next entry
	addq.l	#4,d1		*inc position
	dbra	d2,find_dp_master_pos	
	moveq	#-1,d0
	rts			*not found
dp_found_in_ML:
	move.l	d1,d0	*position*4+4+(number of imports*4)+(number of code ptrs*4)
	rts_	"LINKPPC_Srch_mast_dP_list"
	
**A1=table,a0=import name
sort_import:
	clr.l	d1	*position counter
	move.l	a0,a2
	move.l	a1,a3
	move.l	import_count(a5),d3
	beq.s	none_yet
**search loop
imp_sea_loop:
	move.l	a3,a4		*table
	move.l	a0,a2		*reset name
	bsr	compare_32	*d0=0 found, -1 not found a2,a4
	tst.w	d0
	beq.s	imp_found
	lea	32(a3),a3	*point to next entry
	dec.l	d3
	bne.s	imp_sea_loop	*search
**not found		
none_yet:
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+	*copy import name to table
	addq.l	#1,import_count(a5)
	move.l	ppc_number_of_imports(a5),d0
;	ext.l	d0
	cmp.l	import_count(a5),d0
	blt.s	imports_exceeded
imp_found:	
	clr.l	d0
	rts_	"LINKPPC_sort_import"
imports_exceeded:
	st	LINK_E_FLAG(A5)
	lea	max_imports_text(pc),a0
	bsr	printit
	moveq	#-1,d0		*return error
	rts_	"LINKPPC_Imports_exceeded"

**A1=table,a0=cp name
sort_cps:
	cmpi.l	#-1,32(a0)
	beq.s	not_cp_def
	tst.b	32(a0)
	beq.s	not_cp_def	*its not a def!	
	clr.l	d1	*position counter
	move.l	a0,a2
	move.l	a1,a3
	move.l	code_ptrs_count(a5),d3
	beq.s	none_yetcp
**search loop
cp_sea_loop:
	move.l	a3,a4		*table
	move.l	a0,a2		*reset name
	bsr	compare_32	*d0=0 found, -1 not found a2,a4
	tst.w	d0
	beq.s	cp_found
	lea	32(a3),a3	*point to next entry
	dec.l	d3
	bne.s	cp_sea_loop	*search
**not found		
none_yetcp:

	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+	*copy import name to table
**insert code offset to toc_code_offsets
	move.l	code_ptrs_count(a5),d0
	move.l	toc_code_offsets(a5),a3
	move.l	4(a0),0(a3,d0.l*4)	*store offset of routine
	addq.l	#1,code_ptrs_count(a5)

	move.l	ppc_number_of_code_defs(a5),d0
;	ext.l	d0
	cmp.l	code_ptrs_count(a5),d0
	blt	cps_exceeded
not_cp_def:	
	clr.l	d0
	rts_	"LINKPPC_sort_code_ptrs"
cp_found:

**We have to check if it is a definition
**if the def is found
	save_all
	st	LINK_E_FLAG(A5)
	move.l	a0,-(sp)
	lea	cp_defined_text(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	bsr	printit		*print label name
	lea	cp_defined_text1(pc),a0
	bsr	printit
	lea	file_name_text(pc),a0
	bsr	printit
	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	gcrcrtext(pc),a0
	bsr	printit
	moveq	#-1,d0		*return error
	restore_all
	clr.l	d0
	rts_	"LINKPPC_code_ptrs_double_def"

cps_exceeded:
	st	LINK_E_FLAG(A5)
	lea	max_cps_text(pc),a0
	bsr	printit
	moveq	#-1,d0		*return error
	rts_	"LINKPPC_code_ptrs_exceeded"


**A1=table,a0=dp name
sort_dps:
	cmpi.l	#-4,32(a0)
	bne.s	not_dp_def	*its not a def!	
	clr.l	d1	*position counter
	move.l	a0,a2
	move.l	a1,a3
	move.l	data_ptrs_count(a5),d3
	beq.s	none_yetdp
**search loop
dp_sea_loop:
	move.l	a3,a4		*table
	move.l	a0,a2		*reset name
	bsr	compare_32	*d0=0 found, -1 not found a2,a4
	tst.w	d0
	beq.s	dp_found
	lea	32(a3),a3	*point to next entry
	dec.l	d3
	bne.s	dp_sea_loop	*search
**not found		
none_yetdp:
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+
	move.l	(a0)+,(a3)+	*copy import name to table
**insert code offset to toc_code_offsets
	move.l	data_ptrs_count(a5),d0
	move.l	data_item_offsets(a5),a3
	move.l	4(a0),0(a3,d0.l*4)	*store offset of data
	addq.l	#1,data_ptrs_count(a5)
	move.l	ppc_number_of_data_defs(a5),d0
;	ext.l	d0
	cmp.l	data_ptrs_count(a5),d0
	blt.s	dps_exceeded
not_dp_def:	
	clr.l	d0
	rts_	"LINKPPC_sort_data_ptrs"
dp_found:
	save_all
	st	LINK_E_FLAG(A5)
	move.l	d7,-(sp)
	move.l	a0,-(sp)
	lea	dp_defined_text(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	bsr	printit		*print label name
	lea	file_name_text(pc),a0
	bsr	printit
	move.l	(sp)+,d7
	bsr	print_file_name_d7	
	lea	gcrcrtext(pc),a0
	bsr	printit
	moveq	#-1,d0		*return error
	restore_all
	clr.l	d0
	rts_	"LINKPPC_data_ptrs_double_def"

dps_exceeded:
	st	LINK_E_FLAG(A5)
	lea	max_dps_text(pc),a0
	bsr	printit
	moveq	#-1,d0		*return error
	rts_	"lppc_data_ptrs_exceeded"
		
*****************************************************************
*THIS IS YER ACTUAL LINK AS PER LINK DESIGN NOTE	*
*****************************************************************
**A2 POINTS TO HEADER FOR THIS FILE
**HEADER LAYOUT:
**.L SIZE OF GLOBS
**.L SIZE OF EXTERNS
LINK_FILE:	MOVE.L	D1,HEADER_START(A5)	SAVE FOR LATER
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D1,A0	POINT TO HEADER START FOR THIS FILE
	MOVE.L	4(A0),D7	SIZE OF EXTERNS
	ADD.L	#8+12,A0	POINT TO EXTERNS
	MOVE.L	CODE_START(A5),A1	POINT TO CODE
	LSR.L	#5,D7	GET ACTUAL NUMBER OF EXTERNS
	BEQ	LF_END	NO EXTERNALS TO LINK
	
**NOW COPY EXTERN NAME AND OFFSET
**A0 IS EXTERNAL LABELS, A1 IS CODE.

LINK_FILE_LOOP:
	LEA	EXTERN_NAME(PC),A2
	MOVEQ	#6,D0
CE_LOOP:	MOVE.L	(A0)+,(A2)+	7 LONGS FOR NAME
	DBRA	D0,CE_LOOP
	MOVE.L	(A0)+,D6	GET ABS OFFSET OF LONG THAT NEEDS CHANGING
**NOW SEARCH OTHER FILES AND LIBRARIES FOR EXTERN REF.	
	BSR	FIND_EXTERN	GET ABS OFFSET OF LONG IN D0
	tst.l	d0
	BGE.S	EXTERN_OK	DOESNT EXIST IF -1
	BSR	CHECK_LIBS	CHECK LIBRARIES, AND ISSUE ERROR
	tst.w	rebuild_flag(a5)
	bne	lf_end		*stop link and rebuild master lists
	TST.L	D0
	BGE.S	EXTERN_OK	FOUND IT OK
	CLR.L	D0	GOT NOTHING
**HERE D6=LONG THAT NEEDS CHANGING
**     D0=THE ADDRESS THAT D6 NEEDS CHANGING TO
EXTERN_OK:

	MOVE.L	CODE_START(A5),A2
	MOVE.L	D6,D5	*IN CASE OF $FFXXXXXX (PC OFFSET)
	ANDI.L	#$00FFFFFF,D6	*MAKE 24 BIT ADDRESS
	ADD.L	D6,A2	POINT TO DATA
NOT_PC_REL:
	MOVE.l	-2(A2),D3	CHECK FOR Bx, OR Bccx
	move.l	d3,d4		copy in case it isn't a long branch
	swap	d3		*this is the instruction (b or bcc)
	move.l	d3,diag1(a5)		*f5.1 diagnostics - save for bad inst
	andi.w	#$f800,d3
	cmpI.W	#$4800,D3	CHECK ALL TYPE OF BRANCH INSTS
	BNE.S	L_NOT_BRA	NOT A LONG BRANCH
**THIS IS A BRANCH - CALC OFFSET
	SUB.L	D6,D0	D0=OFFSET
	addq.l	#2,d0
	andi.l	#$3ffffff,d0	*405 mask
	move.l	d0,d1
	andi.w	#$3,d1
	beq.s	tis_even	s'ok
	save_all
	bsr	link_odd_addr	warn of odd address.
	restore_all
TIS_EVEN:
**THIS IS LONG BRANCH
	swap	d3	*long branch
	andi.w	#$3,d3	*preserve lk and aa bits
	or.l	d0,d3	
	MOVE.L	D3,-2(A2)	INSERT LINKED LONG OFFSET (HURRAH)
	BRA.S	LONG_DONE
***Come here if not 48000000 to check cond bra
L_NOT_BRA:
	move.l	d4,d3		copy in case it isn't a long branch
	swap	d4
	andi.w	#$f000,d4
	cmpI.W	#$4000,D4	CHECK ALL TYPE OF BRANCHc INSTS
	BNE.S	L_NOT_sb	NOT A short branch
**THIS IS A Conditional BRANCH - CALC OFFSET
	SUB.L	D6,D0	D0=OFFSET
	addq.l	#2,d0
	
	move.l	d0,d1
	andi.w	#$3,d1
	beq.s	tis_evenw	s'ok
	save_all
	bsr	link_odd_addr	warn of odd address.
	restore_all
TIS_EVENw:
**THIS IS cond BRANCH
;	swap	d3	*long branch
	andi.w	#$3,d3	*preserve lk and aa bit
	CMPI.L	#32767,D0
	BLT.S	LTW2
	BSR.S	L_OFF_TOO_BIG
	BRA.S	LONG_DONE
	
LTW2:	CMPI.L	#-32768,D0
	BGT.S	LTW3
	BSR.s	L_OFF_TOO_BIG
	BRA.S	LONG_DONE
LTW3:
	andi.l	#$ffff,d0	*405 mask
	or.l	d0,d3	
	MOVE.L	D3,-2(A2)	INSERT LINKED LONG OFFSET (HURRAH)

LONG_DONE:	SUBQ.L	#1,D7
	BNE	LINK_FILE_LOOP	DO NEXT EXTERN
	
LF_END:
	RTS_	"LINKPPC_Link_file"
l_not_sb:
	move.l	diag1(a5),d3		*f5.1 diagnostics - save for bad inst
	BSR	print_file_NAME	"filename"
	st	LINK_E_FLAG(A5)	NO OUTPUT
	lea	unkown_inst(pc),a0
	bsr	printit
	rts
				
L_OFF_TOO_BIG:	MOVE.B	#1,LINK_E_FLAG(A5)
	SAVE_ALL
	lea	error_text(pc),a0
	bsr	printit
	BSR	print_file_NAME	"filename"
	lea	the_offset_text(pc),a0
	bsr	printit	" calls  "
	LEA	EXTERN_NAME(PC),A0
	BSR	PRINTIT	"name"
	lea	is_too_big(pc),a0
	bsr	printit	" - the ofset is too big."

	lea	cr_link(pc),a0
	bsr	printit
	RESTORE_ALL

	RTS_	"LINKPPC_Offset_too_big(damn)"
**warn of branch to odd address
link_odd_addr:	MOVE.B	#1,LINK_E_FLAG(A5)	NO OUTPUT
	lea	warning_text(pc),a0
	bsr	printit
	BSR	print_file_NAME	"filename"
	lea	is_branching_text(pc),a0
	bsr	printit	" calls  "
	LEA	EXTERN_NAME(PC),A0
	BSR	PRINTIT	"name"
	lea	is_odd(pc),a0
	bsr	printit	" - the ofset is too big."

	lea	cr_link(pc),a0
	bsr	printit
	rts_	"LINKPPC_Odd_address"
	
**CANT FIND THE LABL, SO ERROR
CHECK_LIBS:
	MOVEM.L	D6-D7/A0-A1,-(SP)
	TST.W	NUMBER_OF_LIBRARIES(A5)
	BEQ.S	THERE_ARE_NONE	NO LIBS
	BSR	READ_FILE_LIB	TRY LIBRARY FILES
	TST.l	D0		*link V2.12!
	BGE.S	FOUND_IN_LIB	-1 IF NOT FOUND
there_are_none:
	MOVE.B	#1,LINK_E_FLAG(A5)
	lea	error_text(pc),a0
	bsr	printit
	BSR.S	print_file_NAME
	lea	the_offset_text(pc),a0
	bsr	printit
	LEA	EXTERN_NAME(PC),A0
	CLR.B	27(A0)
	BSR	PRINTIT
	LEA	LINK_ERR3(PC),A0
	BSR	PRINTIT
FOUND_IN_LIB:	MOVEM.L	(SP)+,D6-D7/A0-A1
	RTS_	"LINKPPC_No_extern_err"
**NAME PRINTS THE NAME BASED ON file_number
**WHICH IS incremented AS THE LINKER RUNS THROUGH THE FILES IN
**REVERSE ORDER
print_file_NAME:
	MOVE.L	link_file_counter(A5),D0
	MOVE.L	FILE_NAMES(A5),A0
	LSL.L	#5,D0
	ADD.L	D0,A0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper1
	cmpi.b	#"z",d0
	bgt.s	is_upper1
	sub.b	#$20,(a0)	make upper
is_upper1:

	BSR	PRINTIT
	RTS_	"LINKPPC_name"
**NAME_D7 PRINTS THE NAME BASED ON file_number in d7
print_file_NAME_D7:
	move.l	d7,d0
	MOVE.L	FILE_NAMES(A5),A0
	LSL.L	#5,D0
	ADD.L	D0,A0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper17
	cmpi.b	#"z",d0
	bgt.s	is_upper17
	sub.b	#$20,(a0)	make upper
is_upper17:

	BSR	PRINTIT
	RTS_	"LINKPPC_name_D7"

************
*FIND_EXTERN SEARCHES ALL LOADED FILES HEADERS FOR THE LABEL NAME IN EXTERN_NAME
*IT RETURNS EITHER THE ABS OFFSET OF THE LABEL OR -1 IF NOT FOUND IN D0
*IT SEARCHES THE FILES BACKWARDS (WHY DO I DO THIS?)
find_extern:	MOVEM.L	D6-D7/A0-A1,-(SP)
	
	LEA	EXTERN_NAME(PC),A2
	MOVE.L	FILE_TABLE(A5),A0
	MOVE.L	TOT_NUMBER_OF_FILES(A5),D7
FE_GLOBAL_LOOP:	SUBQ.L	#1,D7	LOOP COUNTER
	BMI.S	FE_NOT_FOUND	DONE ALL FILES
		
	MOVE.L	FILE_TABLE(A5),A0
	move.l	d7,d6
	muls	#12,d6		*each entry is 3 longs
	MOVE.L	4(A0,D6.L),D1	HEADER START OF THIS FILE - error here - d1 is crap!
	MOVE.L	FILE_HEADERS(A5),A0	POINT TO HEADERS
	ADD.L	D1,A0	POINT TO HEADER START FOR THIS FILE
	MOVE.L	(A0)+,D1	SIZE OF GLOBALS
	MOVE.L	(A0)+,D2	SIZE OF EXTERNS
	lea	12(a0),a0	*skip imports,code and data sizes
	ADD.L	D2,A0	SKIP EXTERNALS TO GLOBALS
	LSR.L	#5,D1	MAKE NUMBER OF GLOBALS
	BEQ.S	FE_NO_GLOBS
FE_LOCAL_LOOP
**EACH LABEL IS 7 LONGS
	MOVE.L	A0,A1	COPY START OF LABEL
	MOVEQ	#6,D3	COUNT 7 LONG COMPARES
GLOB_SEARCH:	CMPM.L	(A2)+,(A0)+	I LOVE 68K!
	BNE.S	LAB_NOT_EQUAL
	DBRA	D3,GLOB_SEARCH
	MOVE.L	(A0),D0	GET GLOBAL VALUE OF LABEL
	BRA.S	FE_FOUND
	
LAB_NOT_EQUAL:	MOVE.L	A1,A0	START OF THIS GLOB
	LEA	32(A0),A0	POINT TO NEXT GLOB
	LEA	EXTERN_NAME(PC),A2	RESET LABEL NAME
	SUBQ.L	#1,D1
	BNE.S	FE_LOCAL_LOOP
FE_NO_GLOBS:	BRA	FE_GLOBAL_LOOP	TRY NEXT FILE

FE_NOT_FOUND:	MOVEQ	#-1,D0	NOT FOUND!
FE_FOUND:	MOVEM.L	(SP)+,D6-D7/A0-A1
	tst.l	d0		*Double error check
	RTS_	"LINKPPC_Find_extern"
************
LINK_READ_IN:	
**FIRST STORE ITS OFFSET IN FILE_TABLE

	MOVE.L	FILE_TABLE(A5),A2
;	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	
;	MOVE.L	L_CODE_POS(A5),0(A2,D0.L*8) CODE POSITION
;	MOVE.L	L_HEAD_POS(A5),4(A2,D0.L*8) HEADER POSITION
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	muls	#12,d0		*we use 3 entries, code, header and data	
	MOVE.L	L_CODE_POS(A5),0(A2,D0.L) CODE POSITION in code chunk
	MOVE.L	L_HEAD_POS(A5),4(A2,D0.L) HEADER POSITION in headers chunk
	move.l	L_Data_pos(a5),8(a2,d0.l) Data position in data chunk

	
;	BSR	GET_O_NAME	GET NAME INTO LINK_NAME
	MOVE.L	A0,-(SP)	SAVE NEXT LINE - not really but easier to convert

**LENGTH OF FILE IN D1
		
	MOVE.L	SOURCE_BUFF(A5),A1
**For v3, 1032 is the entry point if any
**CHECK IS A LINKABLE FILE (1032="cAtH")
	cmpi.l	#"cAtH",1032(a1)
	beq	FORMAT_FAIL	*it's an old file!
**Check for entry point
	move.l	1044(a1),d0
	beq.s	no_entry
**Stop double entries (Fnarrr <BG>)
	tst.l	main_entry(a5)
	beq.s	no_entry_yet
	bra	double_entry_error
no_entry_yet:
**Check ID
	cmpi.l	#"LOFP",1048(a1)
	bne	not_lppc_file
	add.l	l_code_pos(a5),d0	*relocate entry for abs code position
	move.l	d0,main_entry(a5)
**WE add the following code to the start of the file:
*	move.l	#entry_point,d0
*	jmp	(pc,d0.l)
no_entry:
	MOVE.L	CODE_START(A5),A2	CODE GOES HERE (alias of file buff 1)
	ADD.L	L_CODE_POS(A5),A2	POINT TO RIGHT PLACE FOR CODE

	move.l	code_sz(A5),d2	check we havent run out of mem
	move.l	l_code_pos(A5),d3
	cmp.l	d3,d2
	ble	out_of_code_space
	
	MOVE.L	1024(A1),D2	LEN OF GLOBALS was 128
	MOVE.L	1028(A1),D3	LEN OF EXTERNS was 132
	move.l	1032(a1),d4	len of imports
	add.l	1036(a1),d4	len of code ptrs
	add.l	1040(a1),d4	len of data ptrs
;	move.l	1060(a1),d6	entry
;	btst	#31,d6
;	beq.s	no_entry
;	bclr	#31,d6
;	add.l	d3,d6		Modify entry so its global
;	tst.b	entry_set(a5)
;	bne	double_entry
**420
;	add.l	L_CODE_POS(A5),d6	*make global
;	move.l	d6,entry_point(a5)	*the global entry point
;	st	entry_set(a5)
;no_entry:
	move.l	1064(a1),d7	len of code
	move.l	1068(A1),file_data_size(a5)	*store data size
	LEA	1084(A1),A1	skip fixed headers to point to start of globals
	ADD.L	D2,A1	ADD LEN OF GLOBALS
	ADD.L	D3,A1	ADD LEN OF EXTERNS
	add.l	d4,a1	hopefully pointing at code

**Copy the code to code chunk
	move.l	d7,d1	*size of code
	MOVE.L	D1,FILE_CODE_SIZE(A5)	*save code size
;	tst.l	d1
	beq.s	no_Code3			*there is no code in the file!
L_COPY_CODE:	MOVE.B	(A1)+,(A2)+
	subq.l	#1,d1
	bne.s	l_copy_code	make 32 bit 
	bra.s	done_code
no_code3:
**if there is no code, insert a nop just to make linkppc happy
	move.l	#4,file_code_size(a5)
	move.l	#0x60000000,(a2)+	*nop
done_code:
**Copy data to data chunk
	MOVE.L	l_data_START(A5),A2	CODE GOES HERE (source buffer)
	ADD.L	L_data_POS(A5),A2	POINT TO RIGHT PLACE FOR CODE
	move.l	file_data_size(a5),d1	*size of code
;	tst.l	d1
	beq.s	no_data
			*there is no code in the file!
L_COPY_data:	MOVE.B	(A1)+,(A2)+
	subq.l	#1,d1
	bne.s	l_copy_data 

no_data:
**NOW COPY HEADER TO FILE_HEADERS
	MOVE.L	SOURCE_BUFF(A5),A1	 POINT TO START
	MOVE.L	FILE_HEADERS(A5),A2	file_buff1
	ADD.L	L_HEAD_POS(A5),A2	POINT TO RIGHT PLACE
	LEA	1024(A1),A1	SKIP TITLE ETC was 128
	MOVE.L	(A1)+,D2	LEN OF GLOBALS
	MOVE.L	(A1)+,D3	LEN OF EXTERNS
	move.l	(a1)+,d4	len of imports
	move.l	d4,f_len_of_imports(a5)
	move.l	(a1)+,d5	len of code ptrs
	move.l	d5,f_len_of_code_ptrs(a5)
	move.l	(a1)+,d6	len of data ptrs
	move.l	d6,f_len_of_data_ptrs(a5)
	ADD.L	#20,A1	SKIP ID, entry reserved etc
	add.l	#20,a1	skip lengt of code, data and reserved(3) - a1->externs
	
	ADD.L	D2,L_HEAD_POS(A5)	ADD SIZE OF GLOBS
	ADD.L	D3,L_HEAD_POS(A5)	ADD SIZE OF EXTERNS
	add.l	d4,l_head_pos(a5)	add size of imports
	add.l	d5,l_head_pos(a5)	add size of code defs
	add.l	d6,l_head_pos(a5)	add size of data defs
	ADD.L	#20,L_HEAD_POS(A5)	ACCOUNT FOR SIZES

	MOVE.L	D2,(A2)+	STORE LEN OF GLOBS
	MOVE.L	D3,(A2)+	STORE LEN OF EXTERNS
	move.l	d4,(a2)+
	move.l	d5,(a2)+
	move.l	d6,(a2)+	store imps, code and data defs
	
	MOVE.L	L_CODE_POS(A5),D4	BASE OFFSET for globals and externs in this file

	LSR.L	#5,D3	LEN OF EXTERN IN 32 BYTES
	BEQ.S	NO_EXTERNS
	SUBQ.L	#1,D3	FOR DBRA

COPY_EXTERNS1:	MOVE.L	(A1)+,(A2)+	COPY EXTERNAL, COMPLETE WITH OFFSET
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	MAKE REAL OFFSET - ADD CODE OFFSET
	MOVE.L	D1,(A2)+	STORE OFFSET
	DBRA	D3,COPY_EXTERNS1
	
NO_EXTERNS:
	LSR.L	#5,D2	LEN OF GLOBS IN 32 BYTES
;	BEQ	NO_GLOBS_err
	beq.s	no_globs
	SUBQ.L	#1,D2
COPY_GLOBS1:
**Check that global hasn't been defined already!
	save_all
	lea	extern_name(pc),a2
	MOVE.L	(A1)+,(A2)+	COPY 32 BYTES
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	clr.l	(a2)+
	bsr	find_extern
	tst.l	d0
	bmi.s	not_exist1
	restore_all
	bra	double_global	*global	defined twice
	
not_exist1:
	BSR	try_LIB		*CHECK LIBRARIES, AND ISSUE ERROR
	tst.w	d0
	bmi.s	not_exist2
	restore_all
	bra	double_global_lib
not_exist2:
	restore_all
	
	MOVE.L	(A1)+,(A2)+	COPY 32 BYTES
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	ADD CODE OFFSET TO MAKE REAL OFFSET
	MOVE.L	D1,(A2)+	STORE REAL OFFSET
	DBRA	D2,COPY_GLOBS1
NO_GLOBS:
**Now move and relocate all pc's and data offsets according to base code and base data
**Starting with the imports

	move.l	f_len_of_imports(a5),d1
	beq.s	no_file_imports
	divs.l	#36,d1
	subq.l	#1,d1
copy_imports:
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+	36 bytes
	dbra	d1,copy_imports
no_file_imports:

**Copy code ptrs
	move.l	f_len_of_code_ptrs(a5),d1
	beq.s	no_file_code_ptrs
	divs.l	#40,d1
	subq.l	#1,d1
	move.l	l_code_pos(a5),d4
copy_code_ptrs:
	tst.b	32(a1)
	bmi.s	its_a_dummy_def
	add.l	d4,32(a1)	*pc of usage
its_a_dummy_def:
	tst.b	36(a1)
	bmi.s	ext_	*dont reloc external toc routines!
	add.l	d4,36(a1)	*relocate code ptr
ext_:
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+	36 bytes
	move.l	(a1)+,(a2)+	40 bytes
	dbra	d1,copy_code_ptrs
no_file_code_ptrs:
**Copy data ptrs and add data base addr.
	move.l	f_len_of_data_ptrs(a5),d1
	beq.s	no_file_data_ptrs
	divs.l	#40,d1
	subq.l	#1,d1
	move.l	l_data_pos(a5),d4	*size of data
	move.l	l_code_pos(a5),d5	*size of code
copy_data_ptrs:
	cmpi.l	#-4,32(a1)		Is it a def?
	beq.s	data_def
	add.l	d5,32(a1)
data_def:
	cmpi.l	#-1,36(a1)
	beq.s	extern_data_def
	add.l	d4,36(a1)	*relocate data offset	
extern_data_def:
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+	36 bytes
	move.l	(a1)+,(a2)+	40 bytes
	dbra	d1,copy_data_ptrs
no_file_data_ptrs:
	
	MOVE.L	FILE_CODE_SIZE(A5),D1
	ADD.L	D1,L_CODE_POS(A5)	READY FOR NEXT FILE
	move.l	file_data_size(a5),d1	*size of data
	add.L	D1,l_data_pos(A5)	*new start of data

	addq.l	#1,l_number_of_files(A5)
	ADDQ.L	#1,TOT_NUMBER_OF_FILES(A5)
	MOVE.L	(SP)+,A0	A0->NEXT LINE
	CLR.L	D0
	RTS_	"LINKPPC_read_in"
	EVEN

out_of_code_space:
	st	LINK_E_FLAG(A5)
	lea	link_err5(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"LINKPPC_Out_of_code_space"
double_entry:

	st	LINK_E_FLAG(A5)
	lea	entry2_text(pc),a0
	bsr	printit

	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"LinkPPC_Double_entry"

double_global:
	st	LINK_E_FLAG(A5)
	lea	double_global_text1(pc),a0	*the label
	bsr	printit
	lea	extern_name(pc),a0
	clr.l	28(a0)					*mask out offset long
	bsr	printit				*xyz
	lea	double_global_text2(pc),a0	*is defined more than once...
	bsr	printit
	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	gcrcrtext(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"LINKPPC_Double_global"

double_global_lib:
	st	LINK_E_FLAG(A5)
	lea	double_global_text1(pc),a0	*the label
	bsr	printit
	lea	extern_name(pc),a0
	clr.l	28(a0)					*mask out offset long
	bsr	printit				*xyz
	lea	double_global_text3(pc),a0	*is defined more than once...
	bsr	printit
	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	gcrcrtext(pc),a0
	bsr	printit

	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"LINKPPC_Double_global_lib"

double_entry_error:
	st	LINK_E_FLAG(A5)
	lea	double_entry_text(pc),a0	*the label
	bsr	printit

	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	crcrtext(pc),a0
	bsr	printit
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog

	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"LinkPPC_Double_entry"

double_entry_text	dc.b	"***FATAL***",13
	dc.b	"Entry point defined more than once",13
	dc.b	"File currently being processed:",13,0
	align
not_lppc_file:
	st	LINK_E_FLAG(A5)
	lea	bad_file_text(pc),a0	*the label
	bsr	printit

	lea	link_name(pc),a0
	addq.l	#1,a0
	bsr	printit
	lea	crcrtext(pc),a0
	bsr	printit
	lea	linker_errors_string(pc),a0
	bsr	send_to_dialog

	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"LinkPPC_bad_file"
bad_file_text	dc.b	"***FATAL***",13
	dc.b	"File is not a LinkPPC format object file.",13
	dc.b	"File currently being processed:",13,0
	align	     
no_code:
	st	LINK_E_FLAG(A5)
	lea	no_code_text(pc),a0
	bsr	printit
	move.l	(sp)+,a0
	moveq	#-1,d0
	rts_	"LINKPPC_No_code_error"
	
rlf_no_size:	
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper30
	cmpi.b	#"z",d0
	bgt.s	is_upper30
	sub.b	#$20,(a0)	make upper
is_upper30:
	bsr	printit
	LEA	LINK_ERR2(PC),A0
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS_	"PPC_Lk_rlfnosize"

READ_LINK_FAILED:
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper3
	cmpi.b	#"z",d0
	bgt.s	is_upper3
	sub.b	#$20,(a0)	make upper
is_upper3:

	bsr	printit
	LEA	LINK_ERR2(PC),A0
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS_	"PPC_Lk_read_link_failed"
	EVEN
FORMAT_FAIL:
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper4F
	cmpi.b	#"z",d0
	bgt.s	is_upper4F
	sub.b	#$20,(a0)	make upper
is_upper4F:

	bsr	printit

	LEA	LINK_ERR7(PC),A0
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS_	"PPC_LK_Format_fail"
	
LIB_FORMAT_FAIL:
	section
	st	LINK_E_FLAG(A5)
	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	.is_upper4F
	cmpi.b	#"z",d0
	bgt.s	.is_upper4F
	sub.b	#$20,(a0)	make upper
.is_upper4F:
	bsr	printit
	LEA	LINK_ERR17(PC),A0
	BSR	PRINTIT
	MOVEQ	#-1,D0
	RTS_	"PPC_LK_LIB_Format_fail"		
		
no_globs_err:	lea	link_name(pc),a0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper4
	cmpi.b	#"z",d0
	bgt.s	is_upper4
	sub.b	#$20,(a0)	make upper
is_upper4:

	bsr	printit
	LEA	LINK_ERR6(PC),A0	*no globals!
	BSR	PRINTIT
	MOVE.L	(SP)+,A0
	MOVEQ	#-1,D0
	RTS_	"PPC_Lk_No_globs_error"

****************
**LINK_READ_LIB READS A FILE FROM THE LIBRARIES ON THE HEAP
**CALLED WHEN FIND EXTERN FAILS
READ_FILE_LIB:

	BSR	TRY_LIB	SEARCH LIBRARIES FOR EXTERN
	TST.W	D0
	BGE.S	GOT_LIB
	RTS_	"PPC_Lk_read_libs_link"
**FUNCTION FILE NOW IN file_buff2, LENGTH IN D1 
GOT_LIB:
**FIRST STORE ITS OFFSET IN FILE_TABLE
	MOVE.L	FILE_TABLE(A5),A2
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	muls	#12,d0		*we use 3 entries, code, header and data	
	MOVE.L	L_CODE_POS(A5),0(A2,D0.L) CODE POSITION in code chunk
	MOVE.L	L_HEAD_POS(A5),4(A2,D0.L) HEADER POSITION in headers chunk
	move.l	L_Data_pos(a5),8(a2,d0.l) Data position in data chunk

	
	MOVE.L	source_buff(A5),A1	*a1->the function
***insert this function in files et al
**CHECK IS A LINKABLE FILE (1044="FLPo")
	cmpi.l	#"LOFP",1048(a1)
	bne	LIB_FORMAT_FAIL

	MOVE.L	CODE_START(A5),A2	CODE GOES HERE (alias of file buff 1)
	ADD.L	L_CODE_POS(A5),A2	POINT TO RIGHT PLACE FOR CODE

	move.l	code_sz(A5),d2	check we havent run out of mem
	move.l	l_code_pos(A5),d3
	cmp.l	d3,d2
	ble	out_of_code_space
	
	MOVE.L	1024(A1),D2	LEN OF GLOBALS was 128
	MOVE.L	1028(A1),D3	LEN OF EXTERNS was 132
	move.l	1032(a1),d4	len of imports
	add.l	1036(a1),d4	len of code ptrs
	add.l	1040(a1),d4	len of data ptrs
	move.l	1060(a1),d6	entry
	btst	#31,d6
	beq.s	no_entryl
	bclr	#31,d6
;	add.l	d3,d6		Modify entry so its global
	tst.l	entry_point(a5)
	bne	double_entry
	move.l	d6,entry_point(a5)	*the global entry point
		
no_entryl:
	move.l	1064(a1),d7	len of code
	move.l	1068(A1),file_data_size(a5)	*store data size
	LEA	1084(A1),A1	skip fixed headers to point to start of globals
	ADD.L	D2,A1	ADD LEN OF GLOBALS
	ADD.L	D3,A1	ADD LEN OF EXTERNS
	add.l	d4,a1	hopefully pointing at code

**Copy the code to code chunk
	move.l	d7,d1	*size of code
	MOVE.L	D1,FILE_CODE_SIZE(A5)	*save code size
;	tst.l	d1
	beq	no_Code			*there is no code in the file!
L_COPY_CODEl:	MOVE.B	(A1)+,(A2)+
	subq.l	#1,d1
	bne.s	l_copy_codel	make 32 bit 
**Copy data to data chunk
	MOVE.L	l_data_START(A5),A2	CODE GOES HERE (source buffer)
	ADD.L	L_data_POS(A5),A2	POINT TO RIGHT PLACE FOR CODE
	move.l	file_data_size(a5),d1	*size of code
;	tst.l	d1
	beq.s	no_datal			*there is no code in the file!
L_COPY_datal:	MOVE.B	(A1)+,(A2)+
	subq.l	#1,d1
	bne.s	l_copy_datal 
no_datal:
**NOW COPY HEADER TO FILE_HEADERS
	MOVE.L	source_buff(A5),A1	POINT TO START
	MOVE.L	FILE_HEADERS(A5),A2	file_buff1
	ADD.L	L_HEAD_POS(A5),A2	POINT TO RIGHT PLACE
	LEA	1024(A1),A1	SKIP TITLE ETC was 128
	MOVE.L	(A1)+,D2	LEN OF GLOBALS
	MOVE.L	(A1)+,D3	LEN OF EXTERNS
	move.l	(a1)+,d4	len of imports
	move.l	d4,f_len_of_imports(a5)
	move.l	(a1)+,d5	len of code ptrs
	move.l	d5,f_len_of_code_ptrs(a5)
	move.l	(a1)+,d6	len of data ptrs
	move.l	d6,f_len_of_data_ptrs(a5)
	ADD.L	#20,A1	SKIP ID, entry reserved etc
	add.l	#20,a1	skip lengt of code, data and reserved(3) - a1->externs
	
	ADD.L	D2,L_HEAD_POS(A5)	ADD SIZE OF GLOBS
	ADD.L	D3,L_HEAD_POS(A5)	ADD SIZE OF EXTERNS
	add.l	d4,l_head_pos(a5)	add size of imports
	add.l	d5,l_head_pos(a5)	add size of code defs
	add.l	d6,l_head_pos(a5)	add size of data defs
	ADD.L	#20,L_HEAD_POS(A5)	ACCOUNT FOR SIZES

	MOVE.L	D2,(A2)+	STORE LEN OF GLOBS
	MOVE.L	D3,(A2)+	STORE LEN OF EXTERNS
	move.l	d4,(a2)+
	move.l	d5,(a2)+
	move.l	d6,(a2)+	store imps, code and data defs
	
	MOVE.L	L_CODE_POS(A5),D4	BASE OFFSET for globals and externs in this file

	LSR.L	#5,D3	LEN OF EXTERN IN 32 BYTES
	BEQ.S	NO_EXTERNSl
	SUBQ.L	#1,D3	FOR DBRA

COPY_EXTERNS1l:	MOVE.L	(A1)+,(A2)+	COPY EXTERNAL, COMPLETE WITH OFFSET
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	MAKE REAL OFFSET - ADD CODE OFFSET
	MOVE.L	D1,(A2)+	STORE OFFSET
	DBRA	D3,COPY_EXTERNS1l
	
NO_EXTERNSl:
	LSR.L	#5,D2	LEN OF GLOBS IN 32 BYTES
	BEQ	NO_GLOBS_err
	SUBQ.L	#1,D2
COPY_GLOBS1l:	
	MOVE.L	(A1)+,(A2)+	COPY 32 BYTES
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,(A2)+
	MOVE.L	(A1)+,D1
	ADD.L	D4,D1	ADD CODE OFFSET TO MAKE REAL OFFSET
	MOVE.L	D1,(A2)+	STORE REAL OFFSET
	DBRA	D2,COPY_GLOBS1l
no_globsl:
**Now move and relocate all pc's and data offsets according to base code and base data
**Starting with the imports

	move.l	f_len_of_imports(a5),d1
	beq.s	no_file_importsl
	divs.l	#36,d1
	subq.l	#1,d1
copy_importsl:
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+	36 bytes
	dbra	d1,copy_importsl
no_file_importsl:

**Copy code ptrs
	move.l	f_len_of_code_ptrs(a5),d1
	beq.s	no_file_code_ptrsl
	divs.l	#40,d1
	subq.l	#1,d1
	move.l	l_code_pos(a5),d4
copy_code_ptrsl:
	tst.b	32(a1)
	bmi.s	its_a_dummy_defl
	add.l	d4,32(a1)	*pc of usage
its_a_dummy_defl:
	tst.b	36(a1)
	bmi.s	ext_l	*dont reloc external toc routines!
	add.l	d4,36(a1)	*relocate code ptr
ext_l:
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+	36 bytes
	move.l	(a1)+,(a2)+	40 bytes
	dbra	d1,copy_code_ptrsl
no_file_code_ptrsl:
**Copy data ptrs and add data base addr.
	move.l	f_len_of_data_ptrs(a5),d1
	beq.s	no_file_data_ptrsl
	divs.l	#40,d1
	subq.l	#1,d1
	move.l	l_data_pos(a5),d4	*size of data
	move.l	l_code_pos(a5),d5	*size of code
copy_data_ptrsl:
	cmpi.l	#-4,32(a1)		Is it a def?
	beq.s	data_defl
	add.l	d5,32(a1)
data_defl:
	cmpi.l	#-1,36(a1)
	beq.s	extern_data_defl
	add.l	d4,36(a1)	*relocate data offset	
extern_data_defl:
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+
	move.l	(a1)+,(a2)+	36 bytes
	move.l	(a1)+,(a2)+	40 bytes
	dbra	d1,copy_data_ptrsl
no_file_data_ptrsl:
	
	MOVE.L	FILE_CODE_SIZE(A5),D1
	ADD.L	D1,L_CODE_POS(A5)	READY FOR NEXT FILE
	move.l	file_data_size(a5),d1	*size of data
	add.L	D1,l_data_pos(A5)	*new start of data

;	addq.l	#1,l_number_of_files(A5)
;	ADDQ.L	#1,TOT_NUMBER_OF_FILES(A5)


**copy EXTERN_NAME TO FILE NAMES FOR LIBRARY ENTRIES...
	MOVE.L	FILE_NAMES(A5),A0
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	LSL.L	#5,D0
	ADD.L	D0,A0
	LEA	EXTERN_NAME(PC),A1
	MOVE.L	A0,A2	COPY SO WE CAN MODIFY
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+
	MOVE.L	(A1)+,(A0)+	32 BYTES (LAZY I KNOW)

**NOW SEARCH FOR FIRST ZERO AND INSERT "(lib)"
**accounting for 32 bytes length
	moveq	#30,d0	size of name
find_zero:	tst.b	(a2)+
	beq.s	got_end
	subq.w	#1,d0
	bne.s	find_zero
	bra.s	outta_space
**insert (lib)
got_end:	subq.l	#1,a2	back up over zero
	move.b	#" ",(a2)+	insert space
	subq.w	#1,d0
	beq.s	outta_space	This paste function
	move.b	#"(",(a2)+	should really be outlawed
	subq.w	#1,d0	:-)
	beq.s	outta_space
	move.b	#"L",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
	move.b	#"i",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
	move.b	#"b",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
	move.b	#")",(a2)+
	subq.w	#1,d0
	beq.s	outta_space
outta_space:
	addq.l	#1,l_number_of_files(A5)
	ADDQ.L	#1,TOT_NUMBER_OF_FILES(A5)
	
	CLR.L	D0	NO ERRORS...
	move.w	#1,rebuild_flag(a5)	*tell link to restart with new import list etc!
;	BSR	FIND_EXTERN	NOW DO REAL FIND_EXTERN
	RTS_	"PPC_Lk_read_libs_in"

***********
**TRY_LIB SEARCHES THE LIBRARIES FOR EXTERN_NAME
TRY_LIB:	
	MOVEM.L	A0-A4,-(SP)
	MOVE.L	LIBRARIES_H(A5),A0
	MOVE.L	(A0),A0	START OF LIBRARY FILES
	MOVE.W	NUMBER_OF_LIBRARIES(A5),D7
	beq.s	em_not_found		*There are no libs  - master of coding (really stu! - 2.51!)
	
	LEA	LIBRARY_STARTS(a5),A1	ADDRESS OF LIBS
	SUBQ.W	#1,D7
TL_LOOP:	MOVE.L	0(A1,D7.W*4),A2	POINT TO LIBRARY
	BSR	CHECK_LIB	SEARCH THIS LIB FOR EXTERN NAME
	TST.W	D0
	BGE.S	GOT_IT	FOUND IT!
	DBRA	D7,TL_LOOP
	BRA.S	EM_NOT_FOUND	NOT FOUND
**HERE D7=LIBRARY FILE NUMBER
**     D5=FUNCTION POSITION IN LIBRARY FILE
GOT_IT:	
	MOVE.L	0(A1,D7.W*4),A2	POINT TO LIB
	MOVE.L	A2,A3
	MOVE.L	128(A2),D0	NUMBER OF FUNCS IN LIB
	LSL.L	#5,D0	TIMES 32 FOR NAMES
	LEA	132(A3),A3	POINT TO NAMES
	ADD.L	D0,A3	POINT TO OFFSETS
	MOVE.L	0(A3,D5.W*4),D2	GET OFFSET TO START OF FILE
	ADD.L	D2,A2	POINT OT CORRECT FILE!
**NOW WE NEED TO COPY THIS FILE TO file_buff2(A5) AND PUT FILE LENGTH IN D1
**FIRST CALCULATE TO LIBRARY FILE SIZE
	MOVE.L	4(A1,D7.W*4),D0
	SUB.L	0(A1,D7.W*4),D0	LIB FILE LENGTH

**NOW SEE IF THIS IS LAST FUNCTION IN THIS LIBRARY
	MOVE.L	0(A1,D7.W*4),A1	POINT TO LIB
	MOVE.L	128(A1),D1	NUMBER OF FUNCTIONS IN LIB
	SUBQ.L	#1,D1
	CMP.L	D5,D1
	BNE.S	NOT_LAST
**LAST FUNCTION - LIBRARY FILE LENGTH IN D0
	SUB.L	D2,D0	NEXT FILE START - OFFST
	MOVE.L	D0,D1	RETURN LENGTH IN D1
	BRA.S	GOT_LENGTH
**NOT LAST FUNCTION, SO GT NEXT OFFSET AND SUBTRACT PREVIOUS
*8D2 IS PREVIOUS OFFSET
NOT_LAST:	MOVE.L	4(A3,D5.W*4),D1
	SUB.L	D2,D1	LENGTH IN D1
**NOW MOVE D1 BYTES FROM FUNCTION TO file_buff2(A5)
GOT_LENGTH:	ADD.L	0(A3,D5.W*4),A1	POINT TO FUNCTION FILE
	MOVE.L	D1,D2

	MOVE.L	source_buff(A5),A0
COPY_FUNCTION:	MOVE.B	(A1)+,(A0)+
	SUBQ.L	#1,D2
	BNE.S	COPY_FUNCTION

	MOVEM.L	(SP)+,A0-A4
	CLR.L	D0
	RTS
EM_NOT_FOUND:	MOVEM.L	(SP)+,A0-A4
	MOVEQ	#-1,D0
	RTS_	"PPC_Lk_try_lib"
	
****************
**search_lib_for_data SEARCHES THE LIBRARIES FOR EXTERN_NAME
search_libs_for_data:	
	MOVEM.L	A0-A4,-(SP)
	lea	extern_name(pc),a1
	move.l	(a0)+,(a1)+
	move.l	(a0)+,(a1)+
	move.l	(a0)+,(a1)+
	move.l	(a0)+,(a1)+
	move.l	(a0)+,(a1)+
	move.l	(a0)+,(a1)+
	move.l	(a0)+,(a1)+
	move.l	(a0)+,(a1)+	*copy data name to extern_name

	MOVE.L	LIBRARIES_H(A5),A0
	MOVE.L	(A0),A0	START OF LIBRARY FILES
	MOVE.W	NUMBER_OF_LIBRARIES(A5),D7
	beq	em_not_found		*There are no libs  - master of coding (really stu! - 2.51!)

	LEA	LIBRARY_STARTS(a5),A1	ADDRESS OF LIBS
	SUBQ.W	#1,D7
TLd_LOOP:	MOVE.L	0(A1,D7.W*4),A2	POINT TO LIBRARY
	BSR	CHECK_LIBd	SEARCH THIS LIB FOR EXTERN NAME
	TST.W	D0
	BGE.s	GOT_IT_d	FOUND IT!
	DBRA	D7,TLd_LOOP
	BRA.S	EM_NOT_FOUND	NOT FOUND
got_it_d:
;	movem.l	(sp)+,a1-a4	*restore regs
;	movem.l	a1-a4,-(sp)
**HERE D7=LIBRARY FILE NUMBER
**     D5=FUNCTION POSITION IN LIBRARY FILE	
	MOVE.L	0(A1,D7.W*4),A2	POINT TO LIB
	MOVE.L	A2,A3
	MOVE.L	128(A2),D0	NUMBER OF FUNCS IN LIB
	LSL.L	#5,D0	TIMES 32 FOR NAMES
	LEA	132(A3),A3	POINT TO NAMES
	ADD.L	D0,A3	POINT TO OFFSETS
	MOVE.L	0(A3,D5.W*4),D2	GET OFFSET TO START OF FILE
	ADD.L	D2,A2	POINT OT CORRECT FILE!
**NOW WE NEED TO COPY THIS FILE TO file_buff2(A5) AND PUT FILE LENGTH IN D1
**FIRST CALCULATE TO LIBRARY FILE SIZE
	MOVE.L	4(A1,D7.W*4),D0
	SUB.L	0(A1,D7.W*4),D0	LIB FILE LENGTH

**NOW SEE IF THIS IS LAST FUNCTION IN THIS LIBRARY
	MOVE.L	0(A1,D7.W*4),A1	POINT TO LIB
	MOVE.L	128(A1),D1	NUMBER OF FUNCTIONS IN LIB
	SUBQ.L	#1,D1
	CMP.L	D5,D1
	BNE.S	NOT_LASTd
**LAST FUNCTION - LIBRARY FILE LENGTH IN D0
	SUB.L	D2,D0	NEXT FILE START - OFFST
	MOVE.L	D0,D1	RETURN LENGTH IN D1
	BRA.S	GOT_LENGTHd
**NOT LAST FUNCTION, SO GT NEXT OFFSET AND SUBTRACT PREVIOUS
*8D2 IS PREVIOUS OFFSET
NOT_LASTd:	MOVE.L	4(A3,D5.W*4),D1
	SUB.L	D2,D1	LENGTH IN D1
**NOW MOVE D1 BYTES FROM FUNCTION TO file_buff2(A5)
GOT_LENGTHd:	ADD.L	0(A3,D5.W*4),A1	POINT TO FUNCTION FILE
	MOVE.L	D1,D2

	MOVE.L	source_buff(A5),A0
COPY_FUNCTIONd:	MOVE.B	(A1)+,(A0)+
	SUBQ.L	#1,D2
	BNE.S	COPY_FUNCTIONd

	MOVEM.L	(SP)+,A0-A4
	CLR.L	D0
	bra	got_lib

***********	
*8CHECK_LIB SEARCHES LIBRARY IN A2 FOR EXTERN_NAME
*Library file header layout.
*0-127 - title/version etc etc
*128-132.L - Number of functions (.o) files in this libaray
*132->x    - number of functions * 32 byte strings with names of functions
*x->y      - number of functions * 4 byte offsets to start of each function
*The .o files follow consequtively.

CHECK_LIB:	CLR.L	D5	COUNTER
	MOVE.L	128(A2),D6	NUMBER OF FILES
	beq.s	empty_library	V3.95
	LEA	132(A2),A2	POINT TO NAMES
CL_LOOP:	LEA	32(A2),A3	NEXT FUNCTION NAME
	BSR	COMPARE_NAMES	COMPARE EXTERN_NAME TO A2 (32 BYTES)
	TST.W	D0	-1=NO MATCH
	BGE.S	GOT_ENAME	FOUND IT
	MOVE.L	A3,A2	POINT TO NEXT NAME
	ADDQ.L	#1,D5
	CMP.L	D5,D6
	BNE.S	CL_LOOP	
empty_library:	MOVEQ	#-1,D0
	RTS
GOT_ENAME:
;	NOP
	CLR.L	D0
	RTS_	"PPC_LK_Try_lib"

***********	
*CHECK_LIBd SEARCHES LIBRARY IN A2 FOR EXTERN_NAME as a data def!
*Library file header layout.
*0-127 - title/version etc etc
*128-132.L - Number of functions (.o) files in this libaray
*132->x    - number of functions * 32 byte strings with names of functions
*x->y      - number of functions * 4 byte offsets to start of each function
*The .o files follow consequtively.

CHECK_LIBd:
	move.l	a1,-(sp)
	CLR.L	D5	COUNTER
	MOVE.L	128(A2),D6	NUMBER OF FUNCTIONS
	move.l	d6,d4		*copy number of functions
	beq.s	empty_libraryd	V3.95
	lsl.l	#5,d4		*times 32 for name
	move.l	a2,a3		*base of lib	
	add.l	d4,A3	POINT TO offsets
	lea	132(a3),a4	Err, now we do...
	move.l	a2,a3		*base of lib again
**a4->offset list
**a2 and a3 ->lib base	
CL_LOOPd:
	move.l	a2,A1	point to file
	add.l	(a4)+,a1	*Start of file
	move.l	1040(a1),d0
	beq.s	empty_libraryd	*no data pointers
	move.l	1024(a1),d0	*globals
	add.l	1028(a1),d0	*externs
	add.l	1032(a1),d0	*imports (36)
	add.l	1036(a1),d0	*code pointers (40)
	move.l	1040(a1),d1	*data pointers (40)
	divs	#40,d1		*actual number of data pointers
	subq.w	#1,d1

	add.l	#1084,d0	*start of globals
	add.l	d0,a1		*a1->data pointers?
check_data_lp:
	move.l	a1,a2			
	BSR.S	COMPARE_NAMES	COMPARE EXTERN_NAME TO A2 (32 BYTES)
	lea	4(a2),a2	*make sure its not an extern_data type
	cmp.l	#-1,(a2)
	beq.s	its_an_extern
	TST.W	D0	-1=NO MATCH
	BGE.S	GOT_ENAMEd	FOUND IT
its_an_extern:
	
	lea	40(a1),a1
	dbra	d1,check_data_lp

	MOVE.L	A3,A2	POINT TO NEXT NAME

	ADDQ.L	#1,D5
	CMP.L	D5,D6
	BNE.S	CL_LOOPd	
empty_libraryd:	
	move.l	(sp)+,a1	
	MOVEQ	#-1,D0
	RTS
GOT_ENAMEd:	move.l	(sp)+,a1
	CLR.L	D0
	RTS_	"PPC_LK_Try_lib"
	
COMPARE_NAMES:
	move.l	a4,-(sp)
	LEA	EXTERN_NAME(PC),A4
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN_FAILED
	move.l	(sp)+,a4
	CLR.L	D0
	RTS
CN_FAILED:
	move.l	(sp)+,a4
	MOVEQ	#-1,D0
	RTS_	"PPC_lk_Compare_name"

**Needs a2->name32
**      a4->other name32
**Destroys a2 and a4
COMPARE_32:
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CMPM.L	(A2)+,(A4)+
	BNE.S	CN32_FAILED
	CLR.L	D0
	RTS
CN32_FAILED:	MOVEQ	#-1,D0
	RTS_	"PPC_lk_Compare_names32"
	
;GET_O_NAME:	 ADDQ.L	#3,A0	 SKIP /LSPACE
;	 MOVE.L	A0,F_NAME_ADDR(A5)	 SAVE FILE NAME FROM FRAME
;	 LEA	 LINK_NAME(PC),A1	 	 	 	 IN	 WRITE	 ST
;	 addq.l	#1,a1
;	 moveq	 #-1,d1
;GET_FNAME_L1:
;	 addq.l	#1,d1	  
;	 MOVE.B	(A0)+,D0	 STORE FILE NAME	IN OP_NAME FOR OUR USE.
;	 BEQ.S	 GOT_FNAME_L1
;	 CMPI.B	#32,D0	NAME TERMINATED?
;	 BEQ.S	 GOT_FNAME_L1	 YES
;	 CMPI.B	#13,D0
;	 BEQ.S	 GOT_FNAME_L1
;	 MOVE.B	D0,(A1)+
;	 BRA.S	 GET_FNAME_L1	 COPY NEXT CHARACTER
;GOT_FNAME_L1:	 CLR.B	 (A1)+
;	 lea	 link_name(pc),a1
;	 move.b	d1,(a1)	*320 goes pascal!
;	 MOVE.L	A0,-(SP)
;*8check for quiet mode
;	 BTST	 #0,BUILD_OPTIONS(A5)
;	 BNE.S	 BUILD_QUIET_2	 YES, QUIET PLEASE
;	 LEA	 LINK_NAME(PC),A0
;	 addq.l	#1,a0
;	 move.b	(a0),d0
;	 cmpi.b	#"a",d0
;	 blt.s	 is_upper2
;	 cmpi.b	#"z",d0
;	 bgt.s	 is_upper2
;	 sub.b	 #$20,(a0)	 make upper
;is_upper2:
;
;	 BSR.l	 PRINTIT
;	 LEA	 CR_LINK(PC),A0
;	 BSR.L	 PRINTIT
;BUILD_QUIET_2:	MOVE.L	(SP)+,A0
;	 SUBQ.L	#1,A0	 REWIND TO END OF LAST LINE! (DICK HEED)
;L_FIND_EOL:	 MOVE.B	(A0)+,D0
;	 BEQ.S	 GOT_NL1	EOF
;	 CMPI.B	#13,D0
;	 BEQ.S	 GOT_NL1
;	 BRA.S	 L_FIND_EOL
;GOT_NL1:	 MOVE.L	FILE_NAMES(A5),A1
;**COPY NAME TO RIGHT PLACE IN NAMES LIST
;	 MOVE.L	L_NUMBER_OF_FILES(A5),D0
;	 LSL.L	 #5,D0	 EACH ENTRY 32 BYTES
;	 ADD.L	 D0,A1	 POINT TO RIGHT PLACE
;	 LEA	 LINK_NAME(PC),A2
;	 addq.l	#1,a2	 *skip pascal count byte
;COPY_LINK_NAME:	MOVE.B	(A2)+,(A1)+
;	 BNE.S	 COPY_LINK_NAME
;	 RTS_	 "PPC_LK_Get_O_name"	 	 A0-> NEXT LINE


GET_O_NAME:
	MOVE.L	A0,F_NAME_ADDR(A5)	SAVE FILE NAME FROM FRAME
	LEA	LINK_NAME(PC),A1	
	clr.l	d0
	move.b	(a0)+,d0
	move.b	d0,(a1)+
	subq.l	#1,d0
GET_FNAME_L1:	
	MOVE.B	(A0)+,(a1)+	STORE FILE NAME	IN OP_NAME FOR OUR USE.
	dbra	d0,GeT_FNAME_L1
GOT_FNAME_L1:	CLR.B	(A1)+
;	MOVE.L	A0,-(SP)
*8check for quiet mode
	tst.l	quiet_flag(A5)
	BNE.S	BUILD_QUIET_2	YES, QUIET PLEASE
	LEA	LINK_NAME(PC),A0
	addq.l	#1,a0
	move.b	(a0),d0
	cmpi.b	#"a",d0
	blt.s	is_upper2
	cmpi.b	#"z",d0
	bgt.s	is_upper2
	sub.b	#$20,(a0)	make upper
is_upper2:

	BSR	PRINTIT
	lea	space_text(pc),a0
	bsr	printit
	move.l	input_file_size(a5),d0
	bsr	printnum
	lea	bytes_text(pc),a0
	BSR	PRINTIT
BUILD_QUIET_2:
;	MOVE.L	(SP)+,A0
;	SUBQ.L	#1,A0	REWIND TO END OF LAST LINE! (DICK HEED)
	MOVE.L	FILE_NAMES(A5),A1
**COPY NAME TO RIGHT PLACE IN NAMES LIST
	MOVE.L	L_NUMBER_OF_FILES(A5),D0
	LSL.L	#5,D0	EACH ENTRY 32 BYTES
	ADD.L	D0,A1	POINT TO RIGHT PLACE
	LEA	LINK_NAME(PC),A2
	addq.l	#1,a2	*skip pascal count byte
COPY_LINK_NAME:	MOVE.B	(A2)+,(A1)+
	BNE.S	COPY_LINK_NAME
	RTS_	"PPC_link_getoname"		A0-> NEXT LINE
space_text:	cstring	"  "
	align
bytes_text:	cstring	" bytes",13
	align

***********
**CLEAR_OUT MOVES 33K OF DATA FROM CODE_START TO file_buff2 JUST TO CLEAR
**OUT ANY DATA CACHES...
CLEAR_DATA_CACHE:
	MOVE.L	CODE_START(A5),A0
	MOVE.L	source_buff(A5),A1
	MOVE.L	#33000/4,D0
CO_LOOP:	MOVE.L	(A0)+,D1	GET IT INTO PROCESSOR
	NOP		SYNC PIPELINES
	MOVE.L	D1,(A1)+	
	DBRA	D0,CO_LOOP
	RTS_	"PPC_LK_Clear_data_cache_ARGGHH!"		THAT SHOULD DO IT
***********
SET_OP_NAME:	MOVE.L	A0,A2	SAVE POS
	ADDQ.L	#3,A0	SKIP /LSPACE	
	MOVE.L	A0,F_NAME_ADDR(A5)	SAVE FILE NAME FROM FRAME
	LEA	OP_NAME(a5),A1				IN	WRITE	ST
	addq.l	#1,a1
	moveq	#-1,d1
GET_FNAME_L2:
	addq.l	#1,d1	 
	MOVE.B	(A0)+,D0	STORE FILE NAME	IN OP_NAME FOR OUR USE.
	BEQ.S	GOT_FNAME_L2
	CMPI.B	#".",D0	NAME TERMINATED?
	BEQ.S	GOT_FNAME_L2	YES
	MOVE.B	D0,(A1)+
	BRA.S	GET_FNAME_L2	COPY NEXT CHARACTER
GOT_FNAME_L2:
*	MOVE.L	#".APP",(A1)+	NO MORE APPS PPPPLEASSE!
	CLR.B	(A1)+
	lea	op_name(a5),a1
	move.b	d1,(a1)	*320 pascal strings
	MOVE.L	A2,A0
	RTS_	"PPC_LK_Set_op_name"
		
LINK_FIND_EOL:	MOVE.B	(A0)+,D0
	BEQ.S	LFE_END	EOF
	CMPI.B	#13,D0
	BEQ.S	LFE_END
	BRA.S	LINK_FIND_EOL
LFE_END:	RTS_	"PPC_LK_Find_eol"
	
*********************************************************************************
LINK_NAME:	DS.B	40	TEMP FOR .O FILE NAME
EXTERN_NAME:	DS.B	32	LABEL NAME FIND_EXTERN IS LOOKING FOR
*********************************************************************************

	
CR_LINK:	DC.B	13,0
LINK_FOLDER_TEXT:	
	DC.B	13,"Reading files from '",0
dot_cr:	dc.b	"'...",13,0

bytes_used_text:
	dc.b	" bytes used out of ",0
LINK_DONE_TEXT:	DC.B	13,"LINK finished.",13,0
LINK_ERR2:	DC.B	" - **FATAL** Unable to find this file in the Build output folder.",13,0
link_err5:	dc.b	13,"**FATAL** Not enough memory to build the code.",13,0
LINK_ERR3:	DC.B	" - but the reference doesn't exist.",13,13,0
LINK_ERR4:	DC.B	13," - Offset too large.",13,0
no_code_text:	dc.b	"**FATAL** No code in this file.",13,0
double_global_text1:	dc.b	"**FATAL** The global label '",0
double_global_text2:
	dc.b	"' is defined",13,"more than once in this project",13,"(in another of your files).",13
	dc.b	"File being linked = '",0

double_global_text3:
	dc.b	"' is defined in a library in this",13,"project. Please use another name.",13
	dc.b	"File being linked = '",0
		
Link_err6:	dc.b	" - **FATAL** - No global definitions in this file.",13
	dc.b "A linkable file must have at least one global definition",13
	dc.b ", even if it is just a dummy.",13,13,0
the_offset_text:	dc.b	" wants to link with ",13,0
is_too_big:	dc.b	13,"Unfortunately, the offset is too",13,"far for a conditional branch,",13,"which is limited to +/-32k bytes.",13,13,0

LINK_ERR7:	DC.B	" - **FATAL** This file is an old Link (PowerPC)",13,"file. ID at 1044 is FLPo.",13,0
LINK_ERR17:	DC.B	" - **FATAL** This file is an old Link (PowerPC)",13
	dc.b	"library file. ID at 1048 is not LOFP.",13
	dc.b	"Use version 5 libraries.",13,13,0
	align
Link_err8:	dc.b	" - **FATAL** This file is empty?",13,0
	rts_	"PPC_Link_errors_text"
LINK_SPACE_TEXT:	
	DC.B	" -> ",0
error_text:	dc.b	"***ERROR*** ",0
warning_text:	dc.b	"**FATAL ERROR** ",0
is_branching_text:	dc.b	" branches to ",0
is_odd:	dc.b	13," which is at an odd address (i.e. not  multiple of 4)!",13,"This will cause your"
	dc.b	13," program to run slowly."
	dc.b	0
starting_link:	dc.b	13,"Starting link.",13,0
NO_WRITE_TEXT:	DC.B	"**B A D  L I N K**.",13
	DC.B	0
entry2_text:	dc.b	"***FATAL*** Code entry point has been",13,"defined more than once.",13,0
	align
unkown_inst:	dc.b	13,"Linker fail in the above file - unknown instruction.",13,"You're not using LXT and passing"
	dc.b	13,"a routine address via LEA are you?",13,"This is a no-no.",13,0
	align
max_imports_text:	dc.b	"***FATAL*** Exceeded TOC limit for a",13,"fragment during building master "
	dc.b	"list ",13,"of imports! Increase Anvil's memory allowance.",13
	dc.b	13,0
	align
max_cps_text:	dc.b	"***FATAL*** Exceeded TOC limit for a fragment",13,"during building master "
	dc.b	"list of TOC code pointers!",13,"Increase Anvil's memory allowance.",13
	dc.b	13,13,0
	align
cp_defined_text:
	dc.b	"***FATAL*** ",0
cp_defined_text1:	dc.b	" - this code label",13,"has already been",13,"defined as a TOC_ROUTINE label.",0
	align
max_dps_text:	dc.b	"***FATAL*** Exceeded TOC limit for a",13,"fragment during building master "
	dc.b	"list of TOC data pointers! Increase Anvil's memory allowance.",13
	dc.b	13,13,0
dp_defined_text:
	dc.b	"***FATAL*** This label has already",13,"been defined as a data label: ",0

crcrtext:	dc.b	13,13,0
gcrcrtext:	dc.b	"'",13,13,0
link_no_mem:	dc.b	"***FATAL*** Out of memory during link.",13,0
cp_not_found:	dc.b	" - **ERROR** This code label has been declared as an EXTERN in '",0
dp_not_found:	dc.b	" - **ERROR** This data label has been declared as EXTERN_DATA in '",0
but_not_found:	dc.b	"'.",13,"The label has not been defined in any file",0
or_libs:	dc.b	" or library function",0
lib_cr:	dc.b	".",13,13,0
file_name_text:	dc.b	13,"File being processed = '",0

	rts_	"PPC_Link_more_errors"
	align

	EXTERN	printit,SET_DIRECTORY,WRITE_MAC,size_code,send_and_wait
	EXTERN	READ_FILE,GET_NAME,WRITE_SEG,write_xcoff,send_to_dialog
	extern	printnum,do_alert,make_temp_FSspec,read_libs_ppc
	extern	add_library,export_label,strcmp
	PUBLIC	Link_PPC
	
	